###########################################################################
# Contribution to the Chevie Package
#
#  This file contains some supplementary programs for working with braids
#    (C) Jean MICHEL 1995-2007
###########################################################################
##
#F  orbitcircperm(w [,F]) computes the equivalence class of w in the braid
##    monoid for the equivalence relation where x * F(y) and y * x are
##    equivalent
##
##   (this is sometimes the whole F-conjugacy class of w)
##
orbitcircperm:=function(arg)local res,current,v,s,c,F,b,i,M;
  b:=arg[1];if Length(arg)=2 then F:=arg[2];else F:=x->x;fi;
  res:=[b]; current:=1;M:=b.monoid;
  while current<=Length(res) do
    c:=res[current];
    for i in [1..M.nrAtoms] do
      if M.IsLeftDescending(GarsideAlpha(c),i) then
	s:=M.Elt([M.atoms[i]]);
	v:=s^-1*c*F(s);
	if not v in res then Add(res,v);fi;
      fi;
    od;
    current:=current+1;
  od;
  return res;
end;

whichcircperm:=function(arg)local res,current,v,s,c,F,b,b1,i,M,from,l;
  b:=arg[1];b1:=arg[2];
  if Length(arg)=3 then F:=arg[3];else F:=x->x;fi;
  M:=b.monoid;res:=[b]; from:=[[M.Elt([]),]];current:=1;
  if b1=b then return M.Elt([]);fi;
  while current<=Length(res) do
    c:=res[current];
    for s in Concatenation(SimpleLeftDivisors(M,GarsideAlpha(c))) do
      i:=M.Elt([s]); v:=i^-1*c*F(i);
      if not v in res then 
        Add(res,v);Add(from,[i,current]); 
        if v=b1 then
          l:=[i];
          while IsBound(from[current][2]) do
            Add(l,from[current][1]);current:=from[current][2];
          od;
          return Product(Reversed(l));
        fi;
      fi;
    od;
    current:=current+1;
  od;
  return false;
end;
#############################################################################
##
#F  BraidConjugacyClass(w) computes the intersection of the conjugacy class 
##  in the braid group of w with the braid monoid
##
BraidConjugacyClass:=function(b)local B,W,class,a,x,axa;
  W:=CoxeterGroup(b);B:=b.monoid.B; class:=[b];
  for x in class do
    for a in Elements(W) do
      axa:=x^B(a);
      if axa.pd>=0 and not axa in class then Add(class,axa);fi;
    od;
  od;
  return class;
end;

#############################################################################
##
#F  LeftDivisors(b)   returns all divisors of b
##
LeftDivisors:=function(b)local s,M,w;
  M:=b.monoid;if b=M.Elt([]) then return [b];fi;
  w:=GarsideAlpha(b);
  s:=Filtered([1..M.nrAtoms],i->M.IsLeftDescending(w,i));
  s:=List(s,x->M.Elt([M.atoms[x]]));
  return Union([M.Elt([])],s,Union(List(s,x->x*LeftDivisors(x^-1*b))));
end;

RightDivisors:=b->List(LeftDivisors(b.monoid.reverse(b)),b.monoid.reverse);

########################################################################
##
#F  puiss(n,x,F) let x be an object (monoid element) on which F acts.
##               this function computes  (xF)^n 
##
puiss:=function(n,x,F)  
  if n=0 then return x^0;
  else return x* F(puiss(n-1,x,F));
  fi;
end;

########################################################################
##
#F  PiRoots(wF,d) this function returns all d-th roots of pi in the
##              Coxeter coset WF using a very bestial algorithm
##
PiRoots:=function(WF,d)local W,e,pi,F,p,phiorder;
  phiorder:=function(w,p)
    return First(DivisorsInt(OrderPerm(w*p)),i->(w*p)^i=p^i);
  end;
  if IsBound(WF.phi) then W:=CoxeterGroup(WF);p:=WF.phi;F:=Frobenius(WF);
		     else W:=WF;p:=();F:=x->x;
  fi;
  e:=CoxeterElements(W,2*W.N/d);
  e:=List(Filtered(e,x->phiorder(x,p)=d),Braid(W));
  pi:=Braid(W)(LongestCoxeterElement(W))^2;
  e:=Filtered(e,x->puiss(d,x,F)=pi);
  return e;
end;

if false then
test:=function(WF)local F,r,d,rr,n,pi,b,r,W,red,bad,rbad;
  if IsBound(WF.phi) then W:=CoxeterGroup(WF);F:=Frobenius(WF);
  else W:=WF;F:=x->x;fi;
  r:=RegularNumbers(WF);
  r:=r{[2..Length(r)]};
  pi:=Braid(W)(LongestCoxeterElement(W))^2;
  for d in r do
    rr:=PiRoots(WF,d);
    Print("d=",d," reduced roots:",Length(rr));
    rr:=orbitcircperm(rr[1],F);
    Print(" orbitcirc:", Length(rr));
    bad:=0;
    rbad:=0;
    red:=0;
    for b in rr do
      if Length(b.elm)<2 then red:=red+1;fi;
      if puiss(d,b,F)<>pi then 
          if Length(b.elm)<2 then rbad:=rbad+1;fi;
	  bad:=bad+1;
      fi;
    od;
    Print(" reduced:",red," bad:",bad," rbad:",rbad,"\n");
  od;
end;
fi;

AlphaI:=function(b,I)local M,res,i,s;
  M:=b.monoid;
  res:=M.Elt([]);
  i:=1;
  while i<=Length(I)  do
    if M.IsLeftDescending(GarsideAlpha(b),I[i]) then
      s:=M.Elt([M.atoms[I[i]]]);res:=res*s; b:=s^-1*b; i:=1;
    else i:=i+1;
    fi;
  od;
  return res;
end;

# all words representing b
AllWords:=function(b)local M,s;
  M:=b.monoid;
  if b=M.Elt([]) then return [[]];fi;
  s:=GarsideAlpha(b);
  return Concatenation(List(Filtered([1..M.nrAtoms],i->M.IsLeftDescending(s,i)),
    i->List(AllWords(M.Elt([M.LeftQuotient(s,i)])*
    GarsideOmega(b)),d->Concatenation([i],d))));
end;

#############################################################################
# Returns the set of all elements which appear as an item in one of
# the lists in the Hurwitz orbit of the list l.
# to find the dual monoid apply it to an expression for a Coxeter element
HurwitzOrbitItems:=function(l)local orbit,new,totest,r,refs;
  orbit:=[]; totest:=[l];r:=Length(l);
  refs:=Set(l);
  repeat
    new:=Set(Concatenation(List([1..r-1],
      i->List(totest,function(e)local l;
      l:=ShallowCopy(e);
      l{[i,i+1]}:=[l[i+1]^(l[i]^-1),l[i]];
      return l;
      end))));
    UniteSet(orbit,totest);totest:=Difference(new,orbit);
    UniteSet(refs,Set(Concatenation(totest)));
    InfoChevie("#I Hurwitz orbit: ",Length(orbit)," items=",Length(refs),"\n");
  until Length(totest)=0;
  return refs;
end;

# for each atom of dual monoid M, give expression in usual monoid
DualToOrdinary:=function(M)local W,B,h,w;
  W:=M.group;
  if M.deltaword=Concatenation(BipartiteDecomposition(W)) then
     h:=OrderPerm(M.delta);
     if h mod 2=0 then w:=Concatenation(List([1..h/2],i->M.deltaword));
     else w:=Concatenation(List([1..QuoInt(h,2)],i->M.deltaword));
       Append(w,M.deltaword{[1..Length(M.deltaword)/2]});
     fi;
     w:=List([1..Length(w)],i->Concatenation(w{[1..i]},-w{[i-1,i-2..1]}));
     B:=Braid(W);
     w:=List(w,x->ApplyFunc(B,x));
  else w:=HurwitzOrbitItems(List(M.deltaword,Braid(W)));
  fi;
  SortParallel(List(w,x->Position(Reflections(W),EltBraid(x))),w);
  return w;
end;

#############################################################################
##
#F  DualRelations
##  The relations of the dual braid monoid M
##
DualRelations:=function(M)local res,i,j,w,p,l,rels;
  res:=[];
  for i in [1..M.nrAtoms] do
    w:=M.LeftQuotient(M.delta,i);
    for j in [1..M.nrAtoms] do
      if M.IsLeftDescending(w,j) then Add(res,[i,j]);fi;
    od;
  od;
  p:=List(res,x->Product(M.atoms{x}));
  l:=Set(p);
  rels:=List(l,x->[]);
  for i in [1..Length(res)] do Add(rels[Position(l,p[i])],res[i]);od;
  return rels;
end;
