################################################################################
##                                                                            ##
#A  conjbraid.g             CHEVIE library         Nuno Franco, Jean Michel.  ##
##                                                                            ##
#Y  Copyright (C) 2002 - 2007  The CHEVIE Team                                ##
##                                                                             #
##  This file contains the  Franco & Gonzlez-Meneses  and Gebhardt Algorithms #
##                                                                             #
## I: http://arxiv.org/ps/math.GT/0112310 (Conjugacy problem)                  #
## II:http://arxiv.org/ps/math.GT/0201243 (Centralizers)                       #
##                                                                             #
################################################################################

################################################################################
# MinimalConjugatingSimple(a,x) a is a garside monoid element and x a simple.
# Returns the minimal multiple of x such that a^x has an infimum not less than
# a. This multiple is simple. See algorithm 2 in Franco-Gonzales I.
################################################################################

MinimalConjugatingSimple:=function(a,x)local min,s,l,M;
  min:=x;M:=a.monoid;
  if a.pd mod M.orderDelta>0 then return x;fi;
  repeat
    x:=min;for s in a.elm do x:=M.RightLcmSimples(x,s)[3];od;
    l:=M.RightLcmSimples(x,min); min:=l[1];
  until l[3]=M.identity;
  return min;
end;

Cycling:=function(b) local M,w,l,res;
  M:=b.monoid;l:=Length(b.elm);
  if l=0 then return [b,M.identity];fi;
  w:=M.DeltaAction(b.elm[1],-b.pd);
  if Length(b.elm)=1 then return [M.Elt([w],b.pd),w];fi;
  return
    [GarsideEltOps.Normalize(M.Elt(M.AddToNormal(b.elm{[2..l]},w),b.pd)),w];
end;

DeCycling:=function(b) local M,w;
  M:=b.monoid;
  if Length(b.elm)=0 then return [b,M.Elt([])];fi;
  w:=M.Elt([b.elm[Length(b.elm)]])^-1;
  return [b^w,w];
end;

################################################################################
#
# RepresentativeSSS(a): This returns [b,h] such that b is an element
#   of the summit class of a, and h such that  b=a^h)
#
################################################################################

RepresentativeSSS:=function(a)local b,h,Inf,Canlength,counter,M,l;
  if not IsBound(a.pd) then Error("Only for Garside monoids");fi;
  M:=a.monoid;
  h:=M.Elt([]);
  l:=Length(AsWord(M.Elt([M.delta])));
  counter:=0;
  Inf:=a.pd;
  while counter<l do
    b:=Cycling(a); a:=b[1];h:=h*M.Elt([b[2]]);
    if Inf<a.pd then counter:=0; Inf:=a.pd;
    else counter:=counter+1;
    fi;
  od;
  counter:=0;
  Canlength:=Length(a.elm);
  while counter<l do
    b:=DeCycling(a); a:=b[1];h:=h*b[2];
    if Canlength<Length(a.elm) then counter:=0; Canlength:=Length(a.elm);
    else counter:=counter+1;
    fi;
  od;
  return [a,h];
end;

###############################################################################
#
# RepresentativeUSS(a):
#    returns [b,h] such that b is in USS(a), and h such that  b=a^h
#
###############################################################################

RepresentativeUSS:=function(b)local l,c,p,M;
  M:=b.monoid;b:=RepresentativeSSS(b);
  l:=[b[1]];c:=[b[2]];b:=b[1];
  while true do
    b:=Cycling(b);
    p:=Position(l,b[1]);
    if p<>false then return [b[1],Product(c{[1..p]})];
    else Add(l,b[1]); Add(c,M.Elt([b[2]]));b:=b[1];
    fi;
  od;
end;

# Assumes b in USS(b), otherwise may loop infinitely
CyclingOrbit:=function(b)local l;
  l:=[b];
  while true do b:=Cycling(b)[1];
    if b=l[1] then return l;else Add(l,b);fi;
  od;
end;

# Transport(x,s)
# conditions:  x braid in its SSS, s simple  such that x^s in SSS(x). Let c(x)
# be  the element which  conjugates x to  Cycling(x). Then Transport returns
# c(x)^-1*s*c(x^s)
Transport0:=function(x,s)local M,res;
  M:=x.monoid;
  res:=GarsideAlpha(M.Elt([Cycling(x)[2]])^-1*M.Elt([s])*M.Elt([Cycling(x^M.Elt([s]))[2]]));
  if res=M.identity then return M.delta;else return res;fi;
end;

Transport:=function(x,s) local M,R,b;
  M:=x.monoid;
  if Length(x.elm)=0 then return s;fi;
  b:=M.alpha2(M.RightComplementToDelta(x.elm[1]),M.DeltaAction(s,1))[1];
  R:=M.LeftGcdSimples(M.AddToNormal(x.elm{[2..Length(x.elm)]},s)[1],b)[1];
  if R=M.identity then return M.delta;
  else return M.DeltaAction(R,-x.pd);
  fi;
end;

# returns the iterated transports of s along the cycling orbit of b
# (returns only the looping part)
TransportOrbit:=function(b,s)local t,tl,b0,p;
  t:=s;tl:=[s];b0:=b;
  while true do
    repeat
#     if Transport0(b,t)<>Transport(b,t) then Error();fi;
      t:=Transport(b,t);
      b:=Cycling(b)[1]; until b=b0;
    p:=Position(tl,t);
    if p<>false then return tl{[p..Length(tl)]};
    else Add(tl,t);
    fi;
  od;
end;

################################################################################
#                                                                              #
# MinimalSSSConjugatingSimple(a,x)                                          #
#   This  program assumes that a is in its SSS. It finds  the minimal (for     #
#   divisibility) multiple of the simple x such that  to that  a^x is in       #
#   SSS(a). This multiple is  a simple. See algorithm 5 in Franco-Gonzales I.  #
#                                                                              #
################################################################################

MinimalSSSConjugatingSimple:=function(a,x)local min,M,RM,w;
  min:=MinimalConjugatingSimple(a,x);M:=a.monoid;
  repeat
    x:=ReversedWord(a^M.Elt([min]));
    RM:=x.monoid; # is a different monoid in the dual case
    if Length(x.elm)<>Length(a.elm) then
      if x.pd+Length(x.elm)>Length(a.elm)+a.pd then
	w:=M.Elt([min])*ReversedWord(RM.Elt([x.elm[Length(x.elm)]]));
	if w.pd mod M.orderDelta>0 or Length(w.elm)=0 then min:=M.delta;
	else min:=w.elm[1];
        fi;
      else
    min:=min*M.RightComplementToDelta(M.DeltaAction(GarsideAlpha(
                   ReversedWord(RM.Elt([x.elm[1]]))),1-a.pd));
      fi;
    fi;
  until Length(x.elm)=Length(a.elm) or min=M.delta;
  return min;
end;

# pullback is an almost-inverse of the transport
Pullback:=function(y,s)local M,b0,br,i,b;
  M:=y.monoid;
  if Length(y.elm)=0 then return s;fi;
  b0:=M.DeltaAction(M.RightLcmSimples(M.DeltaAction(
     M.RightComplementToDelta(y.elm[1]),-y.pd),s)[2],-1);
  br:=M.DeltaAction(s,y.pd);
  for i in [2..Length(y.elm)] do
    br:=M.RightLcmSimples(br,y.elm[i])[3];
  od;
  b:=M.RightLcmSimples(b0,br)[1];
  return MinimalSSSConjugatingSimple(y,b);
end;

MainPullback:=function(y,s)local co,po,cycle,i,l,k;
  co:=CyclingOrbit(y);
  po:=[];
  cycle:=function(s)
    for i in [Length(co),Length(co)-1..1] do s:=Pullback(co[i],s);od;
    return s;
  end;
  repeat Add(po,s); s:=cycle(s); until s in po;
  k:=Position(po,s)-1;l:=Length(po)-k;
  return po[1+k+((l-k) mod l)];
end;

MinimalUSSConjugatingSimple:=function(y,b)local a,to,p,M;
  a:=MinimalSSSConjugatingSimple(y,b);
  to:=TransportOrbit(y,a);
  M:=y.monoid;
  p:=PositionProperty(to,x->M.RightLcmSimples(b,x)[3]=M.identity);
  if p<>false then return to[p];fi;
  a:=MainPullback(y,b);
  to:=TransportOrbit(y,a);
  p:=PositionProperty(to,x->M.RightLcmSimples(b,x)[3]=M.identity);
  if p<>false then return to[p];
  else Error("theory");
  fi;
end;

###############################################################################
# XConjugacyClass(b)
###############################################################################

XConjugacyClass:=function(b,minconjugating)local a,computed,MinH,w,i,class,M;
  if not IsBound(b.pd) then Error("Only for Garside monoids");fi;
  M:=b.monoid;
  computed:=0;  class:=[b];
  while Length(class)<>computed do
    a:=class[computed+1];
    for i in M.atoms do
      MinH:=minconjugating(a,i);
      w:=a^M.Elt([MinH]);
      if not w in class then Add(class,w); fi;
    od;
    computed:=computed+1;
  od;
  return class;
end;

PositiveConjugacyClass:=b->XConjugacyClass(b,MinimalConjugatingSimple);
SSS:=b->XConjugacyClass(RepresentativeSSS(b)[1],MinimalSSSConjugatingSimple);
USS:=b->XConjugacyClass(RepresentativeUSS(b)[1],MinimalUSSConjugatingSimple);

################################################################################
#                                                                              #
# AreConjugate: Given two positive elements in a Garside group gives as an     #
#                output true or false if they are conjugated or not.           #
#                                                                              #
################################################################################

AreConjugate:=function(a,b)
  a:=RepresentativeUSS(a)[1];b:=RepresentativeUSS(b)[1];
  if a.pd<>b.pd or Length(a.elm)<>Length(b.elm) then return false;
  else return b in USS(a);
  fi;
end;

################################################################################
#                                                                              #
# Pi_1_Conjugacy_Graph(T) for T a "conjugacy table", return generators of the  #
#   Pi_1 of the corresponding graph.                                           #
#                                                                              #
################################################################################

Pi_1_Conjugacy_Graph:=function(CT)local M,i,j,CT,minseq,minel,result,posel,t;
  M:=CT.class[1].monoid; minseq:=[[]]; minel:=[M.Elt([])];
  for i in [1..Length(CT.class)] do
     t:=CT.table[i];
     for j in [1..Length(t)] do
       if t[j].p>Length(minseq) then # t[j].p=Length(minseq)+1 !!!
                                     # this uses the way CT was constructed
         Add(minseq,Concatenation(minseq[i],[[i,j]]));
         Add(minel,minel[i]*t[j].e);
       fi;
     od;
  od;
  # here minseq[p] describes a path to get to p
  result:=[];
  for i in [1..Length(CT.class)] do
     t:=CT.table[i];
     for j in [1..Length(t)] do
       if Concatenation(minseq[i],[[i,j]])<>minseq[t[j].p] then
     posel:=minel[i]*t[j].e;
     if posel<>minel[t[j].p] then
        AddSet(result,posel*minel[t[j].p]^-1);
     fi;
       fi;
     od;
  od;
  return result;
end;

###############################################################################
# XTable: Construct an  'x' table for an element b in W which belongs
# to its 'x' class. 'minconjugating(b,s)' should give the minimal multiple
# of the atom s which conjugates b inside 'x'.
# Returns a record with fields:
#   class: 'x' class of b
#   table: a list of same length such that table[i] is a list of records
#          (e,p) such that class[i]^e is equal to class[p]
###############################################################################
XTable:=function(b,minconjugating)local computed,MinH,w,i,M,p,res;
  if not IsBound(b.pd) then Error("Only for Garside monoids");fi;
# b:=Representativex(b)[1];
  M:=b.monoid; res:=rec(table:=[[]],class:=[b]); computed:=0;
  while Length(res.class)<>computed do
    for i in [1..M.nrAtoms] do
      b:=res.class[computed+1];
      MinH:=minconjugating(b,M.atoms[i]);
      if not ForAny([i+1..M.nrAtoms],k->M.IsLeftDescending(MinH,k)) then
        w:=b^M.Elt([MinH]); p:=Position(res.class,w);
        if p=false then Add(res.class,w); Add(res.table,[]);
          p:=Length(res.class);fi;
        Add(res.table[computed+1],rec(e:=M.Elt([MinH]),p:=p));
      fi;
    od;
    computed:=computed+1;
  od;
  return res;
end;

PositiveConjugacyTable:=b->XTable(b,MinimalConjugatingSimple);
SSSTable:=b->XTable(b,MinimalSSSConjugatingSimple);
USSTable:=b->XTable(b,MinimalUSSConjugatingSimple);

# initially on my laptop with profile the following takes 14.15 sec for USSTable
# B(1,3,4,5,4,3,6,5,1,2,3,2,4,5,4,6,5,4,3,2,1,3,5,6,5,4,3,4,6,5) in B7

################################################################################
# CentralizerGenerators(a): Given a in a Garside group, builds a generating    #
#    set for the centralizer of a. This is done by constructing a generating   #
#    set of the Pi_1 of the USS Table of a.                                    #
################################################################################
CentralizerGenerators:=function(a)local CT;
  a:=RepresentativeUSS(a);CT:=USSTable(a[1]);
  return List(Pi_1_Conjugacy_Graph(CT),x->x^(a[2]^-1));
end;

################################################################################
# RepresentativeConjugation(a,b) Given a,b in a Garside group returns false    #
#   if a and b are not conjugate, and c if a^c=b.                              #
################################################################################
RepresentativeConjugation:=function(b1,c1)
  local a,computed,USS,MinH,w,i,M,b,c,minel,e;
  if not IsBound(b1.pd) then Error("Only for Garside monoids");fi;
  b1:=RepresentativeUSS(b1);  c1:=RepresentativeUSS(c1);
  b:=b1[1]; c:=c1[1];
  if b.pd<>c.pd or Length(b.elm)<>Length(c.elm) then  return false;fi;
  if b=c then return b1[2]*c1[2]^-1;
  else
    M:=b.monoid; minel:=[M.Elt([])]; computed:=0; USS:=[b];
    while Length(USS)<>computed do
      a:=USS[computed+1];
      for i in [1..M.nrAtoms] do
	MinH:=MinimalUSSConjugatingSimple(a,M.atoms[i]);
	if not ForAny([i+1..M.nrAtoms],k->M.IsLeftDescending(MinH,k)) then
	  w:=a^M.Elt([MinH]);
	  if not w in USS then
	    e:=minel[Position(USS,a)]*M.Elt([MinH]);
	    if w=c then return b1[2]*e*c1[2]^-1;fi;
	    Add(USS,w); Add(minel,e);
	  fi;
	fi;
      od;
      computed:=computed+1;
    od;
    return false;
  fi;
end;
