#############################################################################
##
#A  init.g                 CHEVIE library         Meinolf Geck, Frank Luebeck, 
##                              Gunter Malle, Jean Michel and G\"otz Pfeiffer.
##
#Y  Copyright (C) 1992 - 2002 Lehrstuhl D f\"ur Mathematik, RWTH Aachen,
#Y  IWR der Universit\"at Heidelberg,  Germany and University Paris VII,
#Y  France.
##
##  This is the init file of the CHEVIE package.
## 
if not IsBound(CHEVIEPATH) then 
  CHEVIEPATH:=LOADED_PACKAGES.chevie;
fi;

ReadChv:= function(name)
  if not ReadPath(CHEVIEPATH, name, ".g", "ReadChv") then
     Error("CHEVIE library file '", name, "' must exist and be readable");
  fi;
end;

#############################################################################
##     AUTO section
##    
AUTO( ReadChv( "lib/classinv" ),
  CentreMultFunction,
  ClassInvariants,
  ConjugacyTestFunction,
  CycleStructurePermOrbits,
  FingerPrintFunction,
  PointsRepOrb,
  PositionClass,
  PositionSet,
  ShortClassListFunction);

AUTO(ReadChv("lib/complex"),
  Complex,
  ComplexConjugate,
  ComplexOps,
  IsComplex);

AUTO( ReadChv( "lib/cycpol" ),
  AsRootOfUnity,
  CycPol,
  CycPolOps,
  Valuation);

AUTO(ReadChv("lib/decimal"),
  DecimalOps,
  Exp,
  IsDecimal,
  Pi,
  SetDecimalPrecision,
  evalf);

AUTO( ReadChv( "lib/symbol" ),
  CycPolFakeDegreeSymbol,
  CycPolGenericDegreeSymbol,
  DefectSymbol,
  DifferencePartitions,
  HighestPowerGenericDegreeSymbol,
  LessSymbols,
  LowestPowerGenericDegreeSymbol,
  PartBeta,
  PartitionTupleToString,
  RankSymbol,
  ShiftBeta,
  StringSymbol,
  SymbolPartitionTuple,
  Symbols,
  SymbolsDefect,
  Tableaux);

AUTO( ReadChv( "lib/util" ),
  BraidRelation,
  CartesianPosition,
  CharRepresentationWords,
  CheckBraidRelation,
  ChevieIndeterminate,
  EvalPolRoot,
  FastValue,
  GetRoot,
  InductionTable,
  InductionTableOps,
  Inherit,
  InverseListsMap,
  Join,
  PointsAndRepresentativesOrbits,
  PositionSublist,
  Replace,
  Rotations,
  Split,
  SublistUnbnd,
  SymmetricDifference,
  TwoTree,
  Zip);

AUTO( ReadChv( "lib/matrix" ),
  DecomposedMat,
  DiagonalMat,
  DistHelpedRepresentativeOperation,
  EigenvaluesMat,
  ExteriorPower,
  IndependentLines,
  IsDiagonalMat,
  IsLowerTriangularMat,
  IsNormalizing,
  KroneckerMatrix,
  MatStab,
  PermMatMat,
  ProportionalityCoefficient,
  OnMatrices,
  RepresentativeDiagonalConjugation,
  SchurFunctor,
  SymmetricPower);

AUTO( ReadChv( "lib/format" ),
  Format,
  FormatTable,
  BracketIfNeeded,
  IntListToString,
  LaTeXTable,
  PrintToString,
  SPrint,
  StringCoefficient,
  StringCyclotomic,
  StringMonomial,
  StringPolynomial,
  StringQuadratic,
  StringQuotient,
  StringTable,
  TeXTable);

AUTO( ReadChv( "prg/abscox" ),
  AbsCoxOps,
  Bruhat,
  BruhatSmaller,
  CartanMatFromCoxeterMatrix,
  CoxeterHeckeAlgebraOps,
  CoxeterMatrix,
  CoxeterMatrixFromCartanMat,
  ForEachElement,
  IsCoxeterGroup,
  IsCoxeterWord,
  LongestCoxeterElement,
  LongestCoxeterWord,
  PermCosetsSubgroup ,
  ReducedCoxeterWord ,
  ReducedRightCosetRepresentatives,
  RightDescentSet);

AUTO( ReadChv( "prg/abshecke" ), 
  AbsHeckeOps,
  CheckHeckeDefiningRelations,
  HeckeReflectionRepresentation,
  IsHeckeAlgebra); 

AUTO( ReadChv( "prg/affine" ), 
  Affine,
  AffineCoxeterGroupOps,
  AffineRootAction);

AUTO(ReadChv("prg/compatib"),
  BraidWords,
  CartanName,
  CartanType,
  CharHeckeRepresentation,
  CoxeterConjugacyClasses,
  CoxeterElementsLength,
  CoxeterWordReflections,
  DirectSumCartanMat,
  DoublePartitionToString,
  DoublePartitions,
  HeckeCharTable,
  HeckeCharTableDirectProduct,
  HeckeFusion,
  HeckePowermap,
  HeckeReducedChar,
  HeckeScalarProducts,
  KLCoefficient,
  LongestWeylWord,
  ParametersCentralizers,
  PermBraid,
  PermCoxeterWord,
  PermRepresentationRoots,
  PermWeylWord,
  PrintDynkinDiagram,
  ReducedInCoxeterCoset,
  ReducedWeylWord,
  Rootsystem,
  SimpleReflexionMatrices,
  Weyl,
  WeylClassPolynomials,
  WeylConjugacyClasses,
  WeylCosetPermRepresentation,
  WeylElements,
  WeylLengthPerm,
  WeylMueMat,
  WeylReflections,
  WeylRightCosetRepresentatives,
  WeylWordPerm,
  WordBraid);

AUTO( ReadChv( "prg/complexr" ),
  ComplexGroupOps,
  ComplexReflectionGroup,
  CycHeckeOps,
  HeckeCentralMonomials);

AUTO( ReadChv( "prg/coset" ),
  CoxeterCoset,
  CoxeterCosetOps,
  CoxeterSubCoset,
  IsCoxeterCoset,
  Frobenius);

AUTO( ReadChv( "prg/coxeter" ),
  AddComponentsCoxeterGroup ,
  CartanMat,
  CoxeterGroup,
  CoxeterGroupOps,
  CoxeterTypeFromCartanMat,
  HighestShortRoot,
  IsomorphismType,
  MatYPerm,
  OmegaSubgroup,
  PermMatXCoxeterElement,
  PermMatY,
  ReflectionType,
  RootsAndMatgens ,
  RootsIntCartan,
  SimpleRootsSubsystem,
  WordsClassRepresentatives
);

AUTO( ReadChv( "prg/dispatch" ),
  AlmostCharNames,
  AlphaInvolution, 
  AltInvolution, 
  BetaInvolution,
  BraidRelations,
  BrieskornNormalForm,
  CharName,
  CharNames,
  CharParams,
  ChevieAttributeDispatcher,
  ChevieCharInfo,
  ChevieClassInfo,
  ChevieDispatcher,
  Coefficient,
  CoxeterElements,
  CoxeterLength,
  CoxeterWord,
  CoxeterWords,
  EltWord,
  FakeDegrees,
  FirstLeftDescending,
  GenericOrder,
  Hecke,
  HeckeCharValues,
  HeckeClassPolynomials,
  HeckeSubAlgebra,
  HighestPowerFakeDegrees,
  HighestPowerGenericDegrees,
  Invariants,
  IsLeftDescending,
  LeftDescentSet,
  LowestPowerFakeDegrees,
  LowestPowerGenericDegrees,
  MatXPerm,
  NrConjugacyClasses,
  PermMatX,
  PositionId,
  PositionSgn,
  PrintDiagram,
  ReducedInRightCoset,
  ReflectionCharValue,
  ReflectionCoDegrees,
  ReflectionDegrees,
  ReflectionEigenvalues,
  ReflectionName,
  ReflectionSubgroup,
  Reflections,
  RelativeCoset,
  RelativeGroup,
  Representations,
  SchurElement,
  SchurElements,
  Specialization,
  TorusOrder,
  Valuation
);

AUTO( ReadChv( "prg/garside" ), 
  AsFraction,
  AsWord, 
  BipartiteDecomposition,
  Braid, 
  BraidMonoid,
  BraidOps,
  CompleteGarsideRecord,
  DualBraid,
  DualBraidMonoid,
  EltBraid, 
  GarsideAlpha,
  GarsideEltOps,
  GarsideOmega,
  GarsideWords,
  GoodCoxeterWord,
  LeftGcd,
  LeftLcm,
  Presentation,
  ReflectionLength,
  ReflectionWord,
  ReversedWord,
  RightGcd,
  RightLcm,
  ShrinkGarsideGeneratingSet,
  SimpleLeftDivisors);

AUTO( ReadChv( "prg/gencox" ),
  CoxeterGroupByCartanMatrix,
  CoxeterGroupByCoxeterMatrix,
  CoxeterGroupHyperoctaedralGroup,
  CoxeterGroupSymmetricGroup,
  GenCoxOps);

AUTO( ReadChv( "prg/hastype" ),
  FakeDegree,
  HasTypeOps,
  ReflectionGroup);

AUTO( ReadChv( "prg/hecke" ),
  CoxeterGroupOpsHecke,
  HeckeAlgebraOps,
  HeckeCharValuesGood, 
  HeckeScalarProduct,
  JInductionTable,
  PoincarePolynomial);

AUTO( ReadChv( "prg/heckeelt" ), 
  CreateHeckeBasis, 
  HeckeEltOps); 

AUTO( ReadChv( "prg/hekcoset" ),
  CoxeterCosetOpsHecke,
  HeckeCosetOps);

AUTO( ReadChv( "prg/kl" ),
  CriticalPair,
  DualWGraph,
  KLMueMat, 
  KazhdanLusztigCoefficient, 
  KazhdanLusztigMue, 
  KazhdanLusztigPolynomial, 
  LeftCellRepresentation,
  LeftCells, 
  RootParameter,
  WGraphToRepresentation);

AUTO( ReadChv( "prg/permroot" ),
  AsReflection,
  CyclicRegularNumbers,
  GetRelativeAction,
  GetRelativeRoot,
  HyperplaneOrbits,
  IsWordFor,
  PermRootGroup,
  PermRootOps,
  PositionRegularClass,
  Reflection,
  ReflectionCharacter,
  RegularNumbers,
  jInductionTable);

AUTO( ReadChv( "prg/refltype" ), 
  ReflTypeOps);

AUTO( ReadChv( "prg/spets" ),
  GenericSign,
  IsSpets,
  PhiFactors,
  PhiOnDiscriminant,
  Spets,
  SpetsOps,
  SubSpets,
  TorusSign);

AUTO( ReadChv( "prg/wclsinv" ),
  ComponentWordsPerm,
  ComponentWordsPermCoset,
  CoxeterClassParamCheckFunction,
  CoxeterCosetClassParamCheckFunction,
  CoxeterCosetOpsClassInvariants,
  CoxeterGroupOpsClassInvariants);

AUTO(ReadChv("tbl/cmplximp"),
  ImprimitiveCuspidalName);

AUTO(ReadChv("tbl/exceptio"),
  G4_22Helper,G4_22Checker,G4_22Test,VcycSchurElement);

AUTO(ReadChv("contr/conjbrai"),
  Cycling,
  DeCycling,
  AreConjugate,
  CentralizerGenerators,
  RepresentativeConjugation,
  RepresentativeSSS,
  RepresentativeUSS);

AUTO(ReadChv("lib/vk"),Mvp,IsMvp,MvpOps,ScalMvp); # features requiring vkcurve

# initialize global control variables
CHEVIE:=rec(version:="4.devel (Mar 2007)",
  Cache:=rec(
    ReflectionSubgroups:=true,
    CoxeterCosets:=true,
    Spetss:=true,
    SubCosets:=true,
    HeckeAlgebras:=true,
    HeckeCosets:=true,
    IrredInfo:=false),
  AUTO:=rec()
);

if BANNER then
Print("\n  WELCOME  to  the  CHEVIE  package,  Version ",CHEVIE.version,"\n",
"    http://www.math.rwth-aachen.de/~CHEVIE for the stable version 3.1\n",
"    http://www.math.jussieu.fr/~jmichel/chevie for this version\n\n",
      "     Meinolf Geck,  Frank Luebeck,   Gerhard Hiss, \n",
      "     Gunter Malle,  Jean Michel,     Goetz Pfeiffer\n",
      "         Lehrstuhl D fuer Mathematik, RWTH Aachen\n",
      "         Universit'e Paris VII\n",
      "         AG Computational Mathematics Universit\"at Kassel\n",
      "         Galway University\n",
      "\n   This replaces the former weyl package.",
      " For first help type\n",
      "          ?CHEVIE Version 4 -- a short introduction\n\n");
fi;

if not IsBound(InfoChevie) then InfoChevie:=Ignore;fi;
if not IsBound(InfoChevie2) then InfoChevie2:=Ignore;fi;
CHEVIE.Diagnose:=Print; # Diagnose informs about non-fatal error conditions

CHEVIE.RawData:=function(f,field)
  if IsBound(CHEVIE.(field)) and IsBound(CHEVIE.(field).(f))
  then return CHEVIE.(field).(f);
  fi;
  if not IsBound(CHEVIE.AUTO.(field)) then
    Error(field," not in CHEVIE tables");
  fi;
  if not IsBound(CHEVIE.AUTO.(field).(f)) then
    InfoChevie("# no stored data for ",field,".",f,"\n");
    return false;
  fi;
  ReadChv(CHEVIE.AUTO.(field).(f));
  if not IsBound(CHEVIE.AUTO.(field)) or not IsBound(CHEVIE.AUTO.(field).(f))
  then Error("error in CHEVIE auto file");
  fi;
  return CHEVIE.(field).(f);
end;

#############################################################################
##
#F  The following function prepares the way for CHEVIE.Data   It takes
#F  arguments which are an element of the result of ReflectionType for
##  CoxeterGroups, ComplexReflectionGroups or CoxeterCosets
##  
##  gap> CHEVIE.Load("CharName",rec(series:="A",indices:=[1],rank:=1));
##    rec( arg := [ 1 ],  fct := function ( arg ) ... end )

CHEVIE.Load:=function(f,t)local res,p,field,orderphi,phi;
  res:=rec();
  if IsBound(t.orbit) then 
    phi:=t.twist;orderphi:=OrderPerm(t.twist);t:=t.orbit[1];
  else orderphi:=1;
  fi;
  field:=t.series;
  if field="I" then res.arg:=[t.bond];
    if orderphi=2 then field:=SPrint(orderphi,field);fi;
  elif field="ST" then 
    if IsBound(t.ST) then 
      if t.ST in [4..7] then field:="G4_7";res.arg:=[t.ST];
      elif t.ST in [8..15] then field:="G8_15";res.arg:=[t.ST];
      elif t.ST in [16..22] then field:="G16_22";res.arg:=[t.ST];
      else field:=SPrint("G",t.ST);res.arg:=[];
      fi;
    elif orderphi<>1 then field:="timp";res.arg:=[t.p,t.q,t.rank,phi];
    else field:="imp";res.arg:=[t.p,t.q,t.rank];
    fi;
  elif orderphi=1 then 
    if field in ["A","B","D"] then res.arg:=[t.rank];
    elif field in ["H","E","F","G"] then
      field:=SPrint(field,t.rank);res.arg:=[];
    fi;
  else field:=SPrint(orderphi,field);
    if field in ["2A","2D"] then res.arg:=[t.rank];
    elif field="2B" then field:="2I";res.arg:=[4];
    elif field="2G" then field:="2I";res.arg:=[6];
    else field:=SPrint(field,t.rank);res.arg:=[];
    fi;
  fi;
  res.fct:=CHEVIE.RawData(f,field);
  return res;
end;

############################################################################
##
#F  CHEVIE.Data( <fctname>, <type>,<extraargs>)   . . . general function
#F   to access information for irreducible types from libraries
##  
##  see CHEVIE.Load for accepted forms of <type>
##  
##  <fctname> must give the function whose value is required, e.g. 
##  "CharParams" or "HeckeCharTable"
##  
##  example: 
##  gap> CHEVIE.Data("CharParams",rec(series:="B",rank:=2));
##  [[[1,1],[ ]], [[1],[1]], [[],[1,1]], [[2],[]],[[],[2]]]
##
CHEVIE.Data:=function(arg)local res,fct,type,extraargs;
  fct:=arg[1];type:=arg[2];extraargs:=arg{[3..Length(arg)]};
  res:=CHEVIE.Load(fct,type); 
  if res=false then return false;fi;
  if not IsFunc(res.fct) then return res.fct; # a variable
  else return ApplyFunc(res.fct,Concatenation(res.arg,extraargs));
  fi;
end;

CHEVIE.AddData:=function(name,types,data)local type;
  if not IsString(types[1]) then types:=[types];fi;
  for type in types do
    if not IsBound(CHEVIE.(type)) then
      InfoChevie2("# cheviedata: creating type ",type,"\n");
      CHEVIE.(type):=rec();
    fi;
    InfoChevie2("# adding CHEVIE.",type,".",name,"\n");
    CHEVIE.(type).(name):=data;
  od;
end;

CHEVIE.IndirectAddData:=function(f,types,fun)local t;
  for t in types do CHEVIE.AddData(f,t,fun(t));od;
end;

CHEVIE.AutoLoad:=function(flist,types,file)local f,t;
  for t in types do 
    if not IsBound(CHEVIE.AUTO.(t)) then CHEVIE.AUTO.(t):=rec();fi;
    for f in flist do CHEVIE.AUTO.(t).(f):=file;od;
  od;
end;

############################################################################
##
##  Function to maintain a cache
##
#F  CHEVIE.GetCached(parent,name,obj,equaltest)
##  
##  GetCached tests if an object x in parent.(name) satisfies 
##   equaltest(x,obj) and if so returns it else adds obj to parent.(name)
##   and returns obj.
##
CHEVIE.GetCached:=function(parent,name,obj,equaltest)local i;
  if not IsBound(CHEVIE.Cache.(name)) or not CHEVIE.Cache.(name) 
  then return obj;
  fi;
  if not IsBound(parent.(name)) then parent.(name):=[];fi;
  i:=PositionProperty(parent.(name),x->equaltest(x,obj));
  if i=false then Add(parent.(name),obj);return obj;
  else return parent.(name)[i];
  fi;
end;

# Things we want always available 
#
ReadChv("lib/polycyc");
  # Corrects  CyclotomicsOps.\= and adds CyclotomicPolynomialOps
ReadChv("lib/patch");  
  # Corrects  MatGroupOps.in and FieldOps.\=
ReadChv("lib/classinv");
  # Adds PermGroupOps.PositionClass and PermGroupOps.FusionConjugacyClasses
ReadChv("auto");
  # Tells in which files to find table data
ReadChv("tbl/compat3");
ReadChv("unip/init");
