#############################################################################
##
#A  format.g              CHEVIE library                          Jean Michel
##
#Y  Copyright (C) 1999 - 2006  University Paris VII.
##
##  This file contains routines to format various objects in various ways.
##  In particular in provides general support for outputting objects in
##  TeX or LaTeX.
##

############################################################################
#  PrintToString(s,x1,...,xn)
#  Prints objects x1,...,xn to string s if they have a String method
#
SPrint:=function(arg)return String(Concatenation(List(arg,String)));end;

PrintToString:=function(arg)Append(arg[1],
  ApplyFunc(SPrint,arg{[2..Length(arg)]}));end;

#############################################################################
##
#F  IntListToString( <part>, [<brackets>] ) . . . . . . . . as the name says
##  
##  <part> must be a list of integers. If all of them are in [0..9] then
##  a  string  of  digits  corresponding  to  the  entries  of <part> is
##  returned. Otherwise the elements of <part> are converted to strings,
##  concatenated  with separating  commas and  the result  surrounded by
##  brackets (default brackets are parentheses):
##  
##     IntListToString([4,2,2,1,1])      ="42211"
##     IntListToString([14,2,2,1,1])     ="(14,2,2,1,1)"
##     IntListToString([14,2,2,1,1],"[]")="[14,2,2,1,1]"
##  
IntListToString:=function(arg) local digits,brackets,p;
  p:=arg[1];
  if Length(arg)=2 then brackets:=arg[2];else brackets:="()";fi;
  if ForAll(p,x->x<10 and x>=0) then digits:="0123456789"; return String(digits{1+p});
  else return Concatenation([brackets[1]],Join(p),[brackets[2]]);
  fi;
end;

############################################################################
##
#F  StringTable( <t> [,ncol] ) . . . . . . . . . . . displays nicely a table
##  
##  The first line of <t> is taken as titles for the columns.
##  The first ncol columns (default 1) are taken as titles for the rows.
##  The table is split in as many vertical slices as necessary each
##  slice fits on the screen. The first ncol columns are repeated as labels
##  for each of these slices.
##
##  Used in displaying InductionTables, and UnipotentCharacters.
##

StringTable:=function(arg)local t,ncol,max,mbr,br,li,k,StringArr,res;
  t:=arg[1];if Length(arg)=1 then ncol:=1; else ncol:=arg[2];fi;
  t:=List(t,x->List(x,String)); #make sure all entries are strings
  max:=List(TransposedMat(t),x->Maximum(List(x,Length)));
  t:=List(t,l->List([1..Length(max)],function(i)
   if i<=ncol then return String(l[i],-max[i]);
   else            return String(l[i],max[i]);fi;end));
  mbr:=SizeScreen()[1]-2;
  br:=Sum(max{[1..ncol+1]},x->x+1); li:=[ncol+1];  
  res:=[];

  StringArr:=function(li)local l;
    for l in [1 .. Length(t)]  do
      if l = 2  then PrintToString(res,List([1..br],i->'_'),"\n");fi;
      PrintToString(res,Join(t[l]{[1..ncol]}," ")," |",Join(t[l]{li}," "),"\n");
    od;
  end;
  
  for  k in [ncol+2..Length(max)] do
    if br+max[k]+1<mbr then br:=br+max[k]+1; Add(li,k);
    else StringArr(li);Append(res,"\n"); 
      br:=Sum(max{[1..ncol]},x->x+1)+max[k]+1;li:=[k];
    fi;
  od;
  StringArr(li);
  return res;
end;

TeXTable:=function(t)local res,pad;
  res:="{\\tabskip 0mm";pad:="\\kern 1mm";
  PrintToString(res,"\\halign{\\strut\\vrule",pad,"\\hfill$#$",pad,"\\vrule&");
  PrintToString(res,
    Join(List([2..Length(t[1])],i->SPrint(pad,"\\hfill$#$",pad)),"&"),
    "\\vrule\\cr\n");
  t:=List(t,x->List(x,y->Format(y,rec(TeX:=1))));#make all entries TeX
  t:=List(t,l->SPrint(Join(l,"&"),"\\cr\n"));
  PrintToString(res,"\\noalign{\\hrule}",t[1],"\\noalign{\\hrule}");
  Append(res,Concatenation(t{[2..Length(t)]}));
  Append(res,"\\noalign{\\hrule}}}");
  return res;
end;

LaTeXTable:=function(t)local res,n;
  n:=Length(t[1]);
  t:=List(t,x->List(x,y->Format(y,rec(TeX:=1,LaTeX:=1))));
  res:=SPrint("\\tablehead{\\hline ",SPrint(Join(t[1],"&"),"\\\\\n"),"\\hline}\n");
  PrintToString(res,"\\tabletail{\\hline\n}");
  PrintToString(res,"\\begin{supertabular}{|R|",List([2..n],i->'R'),"|}\n");
  PrintToString(res,"\\shrinkheight{30pt}\n");
  Append(res,Concatenation(List(t{[2..Length(t)]},function(l)
    if Length(l)>1 then return SPrint(Join(l,"&"),"\\\\\n");
    else return SPrint("\\hline\n\\multicolumn{",Length(t[1]),"}{|c|}{",l[1],
      "}\\\\\n\\hline\n");
    fi;
    end)));
  return SPrint(res,"\\end{supertabular}\n");
end;

FormatTable:=function(t,option)
  if IsBound(option.LaTeX) then return LaTeXTable(t);
  elif IsBound(option.TeX) then return TeXTable(t);
  else return StringTable(t);
  fi;
end;

TeXBracket:=function(s)
  s:=String(s);
  if Length(s)=1 then 
    if s[1] in "abcdefghijklmnopqrstuvwxyABCDEFGHIJKLMNOPQRSTUVWXYZ" then
         return Concatenation(" ",s);
    else return s;
    fi;
  else return Concatenation("{",s,"}");
  fi;
end;

# Formats name^p
StringMonomial:=function(name,p,option)local res;
  p:=String(p);
  if p="0" then return "1";fi;
  if IsBound(option.TeX) then p:=TeXBracket(p);
  elif IsBound(option.Maple) and p[1]='-' then p:=SPrint("(",p,")");fi;
  if p<>"1" then return SPrint(name,"^",BracketIfNeeded(p));
  else return String(name);fi;
end;

# if in an expression like 2*a brackets are needed around a returns (a) 
BracketIfNeeded:=function(a)local par,bracket,i;
  par:=0;bracket:=false;
  for i in [1..Length(a)] do
    if a[i] in "({" then par:=par+1;
    elif a[i] in ")}" then par:=par-1;
    fi;
    if i>1 and a[i] in "+-/" and a[i-1]<>'^' and par=0 then bracket:=true;fi;
  od;
  if bracket then return Concatenation("(",a,")");
  else return a;
  fi;
end;

# formats coeff*p
StringCoefficient:=function(coeff,p,option)
  if not IsString(coeff) then coeff:=Format(coeff,option);fi;
  if coeff="0" then return "0";fi;
  if coeff="1" and Length(p)<>0 then return p;fi;
  if coeff="-1" and Length(p)<>0 then return Concatenation("-",p);fi;
  if p="1" or p="" then return coeff;fi;
  coeff:=BracketIfNeeded(coeff);
  if IsBound(option.GAP) or IsBound(option.Maple) then Append(coeff,"*");fi;
  return ConcatenationString(coeff,p);
end;

StringPolynomial:=function(coeffs,valuation,name,option)local res,x,p;
  res:="";
  for x in [Length(coeffs),Length(coeffs)-1..1] do
    p:=StringCoefficient(coeffs[x],StringMonomial(name,x+valuation-1,option),option);
    if p<>"0" then
      if p[1]<>'-' then Add(res,'+');fi;
      Append(res,p);
    fi;
  od;
  if Length(res)=0 then res:="0";fi;
  if res[1]='+' then res:=res{[2..Length(res)]};fi;
  return String(res);
end;

# print q/d for a quadratic q
StringQuadratic:=function(q,d,option)local res,gcd,c,p;
  c:=[q.a,q.b];gcd:=Gcd(c);c:=c/gcd;
  res:=rec(s:="",d:=gcd/(q.d*d));
  c:=c*Numerator(res.d);res.d:=Denominator(res.d);
  if c[1]<>0 then Append(res.s,String(c[1])); fi;
  if IsBound(option.TeX) then p:=SPrint("\\sqrt ",TeXBracket(q.root));
  elif IsBound(option.Maple) then p:=SPrint("sqrt(",q.root,")");
  else p:=SPrint("ER(",q.root,")");
  fi;
  p:=StringCoefficient(c[2],p,option);
  if p[1]<>'-' and res.s<>"" then Add(res.s,'+');fi;
  Append(res.s,p);
  return res;
end;

StringQuotient:=function(a,b,option)
  b:=Format(b,option);
  if b="1" then return a;fi;
  if IsBound(option.LaTeX) then return SPrint("\\frac",TeXBracket(a),TeXBracket(b));
  elif IsBound(option.TeX) then return SPrint("{",a,"\\over",b,"}");
  else return SPrint(BracketIfNeeded(a),Concatenation("/",b));
  fi;
end;

StringCyclotomic:=function(c,option)local N,root,v,d,q,res,x,try;
  N:=NofCyc(c);
  res:="";
  root:=function(N)
    if IsBound(option.TeX) then 
      if N=4 then return "i";fi;
      return Concatenation("\\zeta_",TeXBracket(N));
    elif IsBound(option.GAP) then return Concatenation("E(",String(N),")");
    else if N=4 then return "I";fi;
      return Concatenation("E",String(N));
    fi;
  end;
  v:=COEFFSCYC(c);
  d:=Lcm(List(v,Denominator));
  if not N in [1,4] and Number(v,x->x<>0)>1 and not IsBound(option.noQuadrat) then
    q:=Quadratic(c*d);
    if q<>false then res:=StringQuadratic(q,d,option);
      return StringQuotient(res.s,res.d,option);
    fi;
    q:=AsRootOfUnity(c*d);
    if q<>false then
      N:=Denominator(q);
      if IsInt(N/2) and (N/2) mod 2=1 then
	q:=q+1/2;if q>1 then q:=q-1;fi;
	res:=Concatenation("-",
		    StringMonomial(root(Denominator(q)),Numerator(q),option));
      else res:=StringMonomial(root(N),Numerator(q),option);
      fi;
      return StringQuotient(res,d,option);
    fi;
    try:=[(1+E(4))/2,(1-E(4))/2,E(3)/2,E(3)^2/2,1+E(3),1-E(3),
           1+E(3)^2,1-E(3)^2,ER(-3)/2];
    for x in try do # hack for unip. deg. coeffs in G_6, G_14, G_27
      q:=Quadratic(c*d/x);
      if q<>false then 
        res:=BracketIfNeeded(StringCyclotomic(2*x,option));
	q:=StringQuadratic(q,2*d,option);
	return StringQuotient(StringCoefficient(q.s,res,option),q.d,option);
      fi;
    od;
  fi;
  return StringQuotient(StringPolynomial(d*v,0,root(N),option),d,option);
end;

# display an object of certain types I recognize:
# scalars,vectors,matrices of cyclotomics or multi-variable
# polynomials over the cyclotomics
Format:=function(arg)local obj,v,res,s,i,m,option;
  obj:=arg[1];
  if Length(arg)=1 then option:=rec();else option:=arg[2];fi;
  if IsString(obj) then 
    if IsBound(option.GAP) then return SPrint("\"",Replace(obj,
	  "\\","\\\\", "\"","\\\"", "\n","\\n"),"\"");
    else return obj;
    fi;
  elif IsList(obj) then
    if IsBound(option.GAP) or IsBound(option.Maple) then 
         return SPrint("[",Join(List([1..Length(obj)],
	   function(i)if IsBound(obj[i]) then return Format(obj[i],option);
	     else return "";fi;end)),"]");
    elif ForAll(obj,IsList) and Length(Set(List(obj,Length)))<=1 then # IsMat
      m:=List(obj,v->List(v,x->Format(x,option)));
      if IsBound(option.LaTeX) then 
        return Join(List(m,v->SPrint(Join(v,"&"),"\\\\\n")),"");
      elif IsBound(option.TeX) then 
        return Join(List(m,v->SPrint(Join(v,"&"),"\\cr\n")),"");
      else
        res:="";
	s:=List(TransposedMat(m),v->1+Maximum(List(v,Length)));
	if Length(s)>=1 then s[1]:=s[1]-1;fi;
	for v in [1..Length(m)] do
	  for i in [1..Length(m[v])] do Append(res,String(m[v][i],s[i]));od;
	  if v<> Length(m) then Append(res,"\n");fi;
	od;
        return String(res);
      fi;
    else return SPrint(Join(List(obj,x->Format(x,option))," "));
    fi;
  elif IsInt(obj) then return String(obj);
  elif IsRat(obj) then return StringQuotient(Format(Numerator(obj),option),
                                              Denominator(obj),option);
  elif IsCyc(obj) then return StringCyclotomic(obj,option);
  elif IsPolynomial(obj) then 
    if IsBound(obj.name) then
      return StringPolynomial(obj.coefficients,obj.valuation,obj.name,option);
    else return StringPolynomial(obj.coefficients,obj.valuation,
        String(obj.baseRing.indeterminate),option);
    fi;
  elif IsBool(obj) then if obj then return "T"; else return "F"; fi;
  elif IsUnknown(obj) then if IsBound(option.GAP) then return("Unknown()");
                           else return "?";fi;
  elif IsRec(obj) then
    if IsBound(obj.operations) and IsBound(obj.operations.Format)
    then return obj.operations.Format(obj,option);
    else return SPrint("rec(",Join(List(RecFields(obj),
        f->SPrint(f,":=",Format(obj.(f),option)))),")");
    fi;
  elif IsString([obj]) then return [obj];
  else return String(obj);
  fi;
end;

FormatGAP:=x->Format(x,rec(GAP:=true));
FormatMaple:=x->Format(x,rec(Maple:=true));
FormatTeX:=x->Format(x,rec(TeX:=true));
FormatLaTeX:=x->Format(x,rec(LaTeX:=true,TeX:=true));
