# the library version of \in for matrix groups is not even able to
# discover that the generators are in the group, thus any call to
# Subgroup fails!

MatGroupOps.\in:=function ( obj, G ) local  l, v, p;
    if not IsMat( obj ) or Length( obj ) <> Length( G.identity ) or 
       Length( obj[1] ) <> Length( G.identity[1] ) 
          or RankMat( obj ) <> Length( G.identity ) 
        or not IsSubset( G.field, Field( Flat( obj ) ) )  then
        return false;
    fi;
    if obj in G.generators then return true;fi;
    G.operations.MakePermGroupP( G );
    l := MatPermPPermMatrix( G, obj );
    return IsPerm( l ) and l in G.permGroupP;
end;

# the library version of \= for fields has a bug: Cyclotomics
# has no generators. This is special cased below

FieldOps.\=:=function (F, G) local  isEql;
  if IsField(F) and IsFinite(F)  then
    if IsField(G) and IsFinite(G)  then
      isEql:=Size(F)=Size(G) and ForAll(F.generators, g->g in G)
			     and ForAll(G.generators, g->g in F);
    elif IsField(G) then isEql:=false;
    else isEql:=DomainOps.\=(F, G);
    fi;
  elif IsField(F) then
    if IsField(G) and IsFinite(G) then isEql:=false;
    elif IsField(G) then
      if not IsBound(F.generators) then
        if not IsBound(G.generators) then
          if F.name="Cyclotomics" and G.name="Cyclotomics" then 
	    return true;
	  else Error("unforeseen case");
	  fi;
	else return false;
	fi;
      elif not IsBound(G.generators) then return false;
      fi;
      isEql:=ForAll(F.generators, g->g in G) and 
	     ForAll(G.generators, g->g in F);
    else isEql:=DomainOps.\=(F, G);
    fi;
  else isEql:=DomainOps.\=(F, G);
  fi;
  return isEql;
end;

# patch now operations deriving from FieldOps
FiniteFieldOps.\= :=  FieldOps.\=;
NumberFieldOps.\= :=  FieldOps.\=;
CyclotomicsOps.\= :=  FieldOps.\=;
RationalsOps.\=   :=  FieldOps.\=;

CharTableOps.CharNames:=function(t,option)local res,dims,a,letters,p,x;
    if IsBound(t.irredinfo) and IsBound(t.irredinfo[1].charname) then
      return List(t.irredinfo,x->x.charname);
    else 
      res:=[];a:="abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ";
      dims:=[];letters:=[];
      for x in t.irreducibles do
	p:=Position(dims,x[1]);
	if p=false then Add(dims,x[1]);Add(letters,'a');p:=Length(dims);
	else letters[p]:=a[Position(a,letters[p])+1];
	fi;
	if IsBound(option.TeX) then Add(res,SPrint("\chi_{",x[1],[letters[p]],"}"));
	else Add(res,SPrint("X",x[1],[letters[p]]));
	fi;
      od;
      return res;
    fi;
end;

# GroupOps has Print but not String!
GroupOps.String:=function(G)local i;
  if IsBound(G.name) then return G.name;
  elif not IsBound(G.parent) then
    if G.generators<>[] then return SPrint("Group( ",Join(G.generators),")");
    else return SPrint( "Group( ", G.identity, " )" );
    fi;
  else return SPrint( "Subgroup( ", G.parent, ", ", G.generators, " )" );
  fi;
end;

# this routine added by F.Celler solves some pb. with polynomials
LaurentPolynomialRingOps.\= := function(R,S)
  if IsBound(R.isLaurentPolynomialRing) and R.isLaurentPolynomialRing
    and IsBound(S.isLaurentPolynomialRing) and S.isLaurentPolynomialRing    
  then return true;
  else return RingOps.\=(R,S);
  fi;
end;

# adds a method 'Coefficient' for polynomials
PolynomialOps.Coefficient:=function(p,i)
  if not IsInt(i) or i<p.valuation or
    Length(p.coefficients)<i+1-p.valuation then return p.baseRing.zero;
  else return p.coefficients[i+1-p.valuation];
  fi;
end;
