#############################################################################
##
#A  util.g              CHEVIE library          Meinolf Geck, Frank L\"ubeck, 
#A                                              Jean Michel, G\"otz Pfeiffer
##
#A  $Id: util.g,v 1.1 1997/01/21 13:46:35 gap Exp $
##
#Y  Copyright (C) 1992 - 1996  Lehrstuhl D f\"ur Mathematik, RWTH Aachen, IWR
#Y  der Universit\"at Heidelberg, University of St. Andrews, and   University 
#Y  Paris VII.
##
##  This file contains routines of general utility which have been
##  developed during the writing of the chevie package. These routines
##  should really put into some other part of the GAP library, they
##  have nothing specific to Coxeter groups.
##

############################################################################
# PositionSublist(s,sub)
# returns the position of the first occurence of sub in s, or false
# if there is no such occurence
PositionSublist:=function(s,sub)local i,j;
  j:=1;
  for i in [1..Length(s)] do
    if s[i]=sub[j] then
      if j=Length(sub) then return i-Length(sub)+1;
      else j:=j+1;
      fi;
    else j:=1;
    fi;
  od;
  return false;
end;

############################################################################
# Replace(s,s1,r1,s2,r2,....)
#  replace in string (or more generally list) s all occurences of s1 by r1
#  then of s2 by r2, etc... return the resulting string or list
#
Replace:=function(arg)local pat,rep,s,l,i;
  s:=arg[1];arg:=arg{[2..Length(arg)]};
  while Length(arg)>0 do
    pat:=arg[1];rep:=arg[2];arg:=arg{[3..Length(arg)]};
    l:=Length(pat);i:=1;
    while i<=Length(s)-l+1 do
      if s{i+[0..l-1]}=pat then 
	s:=Concatenation(s{[1..i-1]},rep,s{[i+l..Length(s)]});
	i:=i+Length(rep);
      else i:=i+1;
      fi;
    od;
  od;
  return s;
end;

#############################################################################
##
#F  Join( <list> [, <delim>]) . . . . . . . Similar to join in Perl
##  
##   Joins the elements of the list, converted to strings, with delim
##   (default delim is comma).
##  
##     [14,2,2,1,1]  --->    "14,2,2,1,1"
##  
Join:=function(arg) local res, i, delim;
  if Length(arg)=2 then delim:=arg[2];else delim:=",";fi;
  res:="";
  for i in [1..Length(arg[1])] do 
    if i>1 then Append(res,delim);fi;
    Append(res,String(arg[1][i]));
  od;
  return String(res);
end;

#############################################################################
##
#F  Split( <string> [, <delim>]) . . . . . . . Similar to split in Perl
##  
##   Splits the string along delim (default delim is comma).
##  
##     "14,2,2,1,"  --->    ["14","2","2","1",""]
##  
Split:=function(arg) local s,res, p, delim;
  s:=arg[1];
  if Length(arg)=2 then delim:=arg[2];else delim:=',';fi;
  res:=[];
  while true do
    p:=Position(s,delim);
    if p=false then Add(res,s);return res;
    else Add(res,s{[1..p-1]});s:=s{[p+1..Length(s)]};
    fi;
  od;
end;

#############################################################################
##
#F  CharRepresentationWords(<rep>,<elts>) . . . . character of representation 
#F  . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .  on words
##  
##  <rep> is a list of matrices representing images of generators,
##  <elts> a list of words in the generators.  Returns the character on <elts>.
##  Tries to be efficient by not computing twice the same product of matrices.
## 
CharRepresentationWords:=function(rep,elts)local l,res,i,f,reps;
  if Length(rep)=0 then return List(elts,x->1);fi;
  l:=[1..Length(elts)]; SortParallel(List(elts,Length),l);
  Print(Length(l),":\c");
  res:=[];reps:=[];
  f:=function(w)local p,j,n;
    if w=[] then return rep[1]^0;fi;
    p:=PositionProperty(l,j->Length(elts[j])>=Length(w));
    for j in l{[p-1,p-2..1]} do
      n:=Length(elts[j]);
      if n<>0 and elts[j]=w{[1..n]} then 
	return reps[j]*f(w{[n+1..Length(w)]});
      fi;
    od;
    return rep[w[1]]*f(w{[2..Length(w)]});
  end;
  for i in l do 
    Print(Position(l,i),".\c");
    reps[i]:=f(elts[i]); res[i]:=TraceMat(reps[i]); od;
  return res;
end;

#############################################################################
##
#F  TwoTree( <m> ) . . . . . . . . . 
# Let m be a matrix whose zeroes are in symmetric positions with respect
# to the diagonal.  Let G be the graph where  vertices i,j are connected
# if m[i][j]<>0. If G is a line this  routine returns it. Else if G is a
# 3-branch tree this routine returns returns [c,b1,b2,b3] where c is the
# central node and b1,b2,b3 are  the branches oriented starting from the
# central node. Otherwise it returns 'false'
#
TwoTree:=function(m)local branch,r,line,l,p,tmp;
  branch:=function(x)
    repeat x:=PositionProperty([1..r],i->m[x][i]<>0 and not i in line);
     if x<>false then Add(line,x);fi;
    until x=false;
  end;
  r:=Length(m);line:=[1];branch(1);l:=Length(line);branch(1);
  line:=line{Concatenation([Length(line),Length(line)-1..l+1],[1..l])};
  l:=Length(line);
  if ForAny([1..l],i->ForAny([1..i-2],j->m[line[j]][line[i]]<>0)) then 
    return false;fi;
  if l=r then return line;fi;
  p:=Difference([1..r],line);
  p:=PositionProperty(line,x->ForAny(m[x]{p},i->i<>0));
  branch(line[p]);
  if Length(line)<>r then return false;fi;
  tmp:=line[p];line:=[line{[p-1,p-2..1]},line{[p+1..l]},line{[l+1..r]}];
  SortParallel(List(line,Length),line);
  return Concatenation([tmp],line);
end;


#############################################################################
##
#F  SublistUnbnd( <l>, <ind> ) . . . . . . . . . Sublist of a list with
#F  possibly unbound entries.
##  
## The writing of this  function was prompted by the fact  that if l has
## some  unbound entries,  l{ind} returns  an error  message instead  of
## doing what is expected (which is  what this routine does). Maybe this
## should be changed?
##
SublistUnbnd:=function(l,ind)local res,i;
  res:=[];
  for i in [1..Length(ind)] do
    if IsBound(l[ind[i]]) then res[i]:=l[ind[i]];fi;
  od;
  return res;
  end;

#############################################################################
##
#F  SymmetricDifference( <M>, <N> ) . . . . . symmetric difference of two sets
##  
SymmetricDifference:=function(x,y)
  return Difference(Union(x,y),IntersectionSet(x,y));
end;

#############################################################################
##
#F  Rotations(<l>) . . . . . All rotations of list l
##  
Rotations:=function(s)local e;e:=Length(s);
  return List([0..e-1],i->s{1+List([i..i+e-1],x->x mod e)});
end;

#############################################################################
##
#F  InverseListsMap( <l> ) . . . . . . . . . . . . . . . . . . . see below
##  
##  <l>  should  be a  list   of  lists  of  positive  integers such  that
##  Concatenation(<l>) contains no element  twice. The function returns  a
##  list r, such that r[i]=[k,l] if <l>[k][l]=i.
##  
InverseListsMap:=function(l)
  local   res,  i,  j;
  res:=[];
  for i in [1..Length(l)] do
    for j in [1..Length(l[i])] do
      res[l[i][j]]:=[i,j];
    od;
  od;
  return res;
end;

Inherit:=function(arg)local r1,r2,ff,f;
  if Length(arg)=3 then ff:=arg[3];else ff:=RecFields(arg[2]);fi;
  for f in ff do arg[1].(f):=arg[2].(f);od;
  return arg[1];
end;

# Given a1,..,an,f returns list of f(a1[i],a2[i],..,an[i])
Zip:=function(arg)local f;f:=arg[Length(arg)];
  return List(TransposedMat(arg{[1..Length(arg)-1]}),v->ApplyFunc(f,v));
end;

# let v1,..,vn be lists of length l1,..,ln
# CartesianPosition([i1,..,in],[l1,..,ln]) gives the position
#            of (v1[i1],..vn[in]) in Cartesian(v1,..vn) while
# CartesianPosition(i,[l1,..,ln]) gives [i1,..,in] such that
#            i=CartesianPosition([i1,..,in],[l1,..,ln])
#            returns false if i is not in [1..Product(l)]
CartesianPosition:=function(pos,l)local res,prod,i;
  if IsInt(pos) then
    res:=[];pos:=pos-1;
    while Length(l)>0 do
      i:=pos mod l[Length(l)];
      Add(res,i);
      pos:=(pos-i)/l[Length(l)];
      l:=l{[1..Length(l)-1]};
    od;
    if pos<>0 then return false;else return Reversed(res+1);fi;
  else
    res:=1;prod:=1;
    for i in [Length(l),Length(l)-1..1] do
      if pos[i]=false then return false;fi;
      res:=res+(pos[i]-1)*prod;
      prod:=prod*l[i];
    od;
    return res;
  fi;
end;

InductionTableOps:=OperationsRecord("InductionTableOps");

InductionTableOps.Print:=function(t)Print(t.what,"Table( ",t.u,", ",t.g,")");end;

InductionTableOps.Format:=function(t,option)local
  arr,res,i,v,table,charGroup,charSubgroup;
  res:=SPrint(t.head(t,option),"\n");
  table:=t.scalar;
  if IsBound(option.charsGroup) then
    table:=table{option.charsGroup};
    charGroup:=t.gNames(t,option){option.charsGroup};
  else charGroup:=t.gNames(t,option);
  fi;
  if IsBound(option.charsSubgroup) then
    table:=table{[1..Length(table)]}{option.charsSubgroup};
    charSubgroup:=t.uNames(t,option){option.charsSubgroup};
  else charSubgroup:=t.uNames(t,option);
  fi;
  arr:=[Concatenation([" "],charSubgroup)];
  for i in [1..Length(table)] do
    v:=[charGroup[i]];
    Append(v,List( table[i], function(a) 
                               if a=0 then return "."; 
                               else return Format(a,option);fi;end));
    Add(arr,v);
  od;
  if IsBound(option.LaTeX) then Append(res,LaTeXTable(arr));
  elif IsBound(option.TeX) then Append(res,TeXTable(arr));
  else Append(res,StringTable(arr));
  fi;
  return res;
end;

InductionTableOps.Display:=function(arg)Print(ApplyFunc(Format,arg));end;

############################################################################
##
#F  InductionTable( <u>, <g> ) . . . . . . . . . . . decomposition of induced 
#F  characters 
##  
##  Let <u>  be  a subgroup of  <g>.   'InductionTable'  returns a  record
##  describing  the  decomposition of  the  irreducible  characters of <u>
##  induced to <g>.
##  
##  The result can be displayed using 'Display'.
##  
##  In fact this function also works for Coxeter cosets.
##  
InductionTable:=function(u,g) local tu,tg,res;
  tu:=CharTable(u); tg:=CharTable(g);
  StoreFusion(tu,tg,FusionConjugacyClasses(u,g));
  res:=rec(scalar:=MatScalarProducts(tu,tu.irreducibles,
                                  Restricted(tg,tu,tg.irreducibles)));
  res.u:=u; res.g:=g; res.what:="Induction";
  res.head:=function(t,option)local uname,gname;
    if IsBound(t.u.operations.ReflectionName) then uname:=ReflectionName(t.u);
    else uname:=String(t.u);fi;
    if IsBound(t.g.operations.ReflectionName) then gname:=ReflectionName(t.g);
    else gname:=String(t.g);fi;
    return SPrint(t.what," from ",uname," to ",gname);end;
  res.uNames:=function(res,option)return CharNames(res.u,option);end;
  res.gNames:=function(res,option)return CharNames(res.g,option);end;
  res.operations:=InductionTableOps;
  return res;
end;

##############################################################################
# 'Value' function for polynomials represented as
#   [ vector of coefficients, valuation]
#
# This function is used in particular for char. tables of Hecke algebras of
# type En, Hn. This seems to be the best "Value" function because many entries
# are sparse polynomials:
#
FastValue:=function(f,x)
  local  val, i, p;
  if IsPolynomial(x) and x=Indeterminate(x.baseRing) then
      return Polynomial(x.baseRing,f[1],f[2]);
  fi;
  if f[1]=[] then
    return 0*x;
  fi;
  val:=0*x;
  p:=f[2];
  for i in f[1] do
    if i<>0 then
      val:=val+i*x^p;
    fi;
    p:=p+1;  
  od;
  return val;
end;

#############################################################################
##
#F  PointsAndRepresentativesOrbits( <g>[, <m>] ) . . . . . . orbits of points 
#F  under permutation group <g>
##  
##  <g>  must  be  a permutation  group.   'PointAndRepresentativesOrbits'
##  returns a  list [orb,rep].  Here  orb is  a list  of the <g>-orbits on
##  [1..LargestMovedPoint(<g>)] or,  if  given, [1..<m>].  rep[i][j] is an
##  element of <g>, such that orb[i][1]^rep[i][j] = orb[i][j].
##  
##  ??? better to give list of points instead of <m> ???
PointsAndRepresentativesOrbits:=function ( arg )
  local  G, orbs, orb, max, g, gs, new, gen, Ggen, p, pnt, fst, img;
  G:=arg[1];

  if Length(arg)>1 then max:=arg[2];
  else max:=PermGroupOps.LargestMovedPoint(G);
  fi;

  Ggen:=G.generators;
  new := BlistList( [ 1 .. max ], [ 1 .. max ] );
  orbs := [  ]; gs:=[];
  fst := 1;
  while fst <> false  do
    orb := [ fst ]; g:=[()]; new[fst] := false;
    p:=1;
    while p<=Length(orb)  do
      for gen  in Ggen  do
	img := orb[p] ^ gen;
	if new[img]  then
	  Add( orb, img ); Add(g,g[p]*gen); new[img] := false;
	fi;
      od;
      p:=p+1;
    od;
    Add( orbs, orb ); Add( gs, g );
    fst := Position( new, true, fst );
  od;
  return [orbs,gs];
end;

GetRootWarned:=[];

############################################################################
# GetRoot(x, [n, [msg]])
#  return an n-th (default 2) root of x when possible, else signals an error.
#  if msg is present and InfoChevie=Print warns about choice of root made
# after printing msg
#
#  For now, it is possible to find an nth-root of:
#  (1) a monomial of the form a.x^b when b is divisible by n and
#        we know how to find an n-th  root of a.
#  (2) a root of unity
#  (3) a rational, when n=2 or x is a perfect power
#  (4) a product of (2) by (3)
#
GetRoot:=function(arg)local x,n,msg,i,j,k,ret,d,n1,error;
  x:=arg[1];
  if Length(arg)=1 or not IsRat(arg[2]) then n:=2;arg:=arg{[2..Length(arg)]};
  else n:=arg[2];arg:=arg{[3..Length(arg)]};fi;
  if n=1 then return x;fi;
  msg:=ApplyFunc(SPrint,arg);
  if msg="" and IsBound(CHEVIE.checkroots) then msg:="checkroots";fi;
  error:=msg<>"no";
  if msg<>"" and error then
    ret:=function(res) 
      if not x in GetRootWarned then
	InfoChevie("#warning: ",msg,": ",Format(res)," chosen as ",Ordinal(n),
	  " root of ",x,"\n");
	AddSet(GetRootWarned,x);
      fi;
      return res;
    end;
  else ret:=x->x;
  fi;
  if x=1 then return ret(1);
  elif IsPolynomial(x) and x.valuation mod n=0 and Length(x.coefficients)=1 then
    return ret(Polynomial(x.baseRing,[GetRoot(x.coefficients[1],n,"no")],x.valuation/n));
  elif IsInt(x) then
    if n=2 then return ret(ER(x));
    elif x>0 then i:=RootInt(x,n); if i^n=x then return ret(i);fi;
    else i:=RootInt(-x,n); 
	 if i^n=-x then return ret(i*E(2*n));fi;
    fi;
  elif IsRat(x) then
    return GetRoot(Numerator(x),n)/GetRoot(Denominator(x),n);
  elif IsCyc(x) then
    i:=AsRootOfUnity(x);
    if i<>false then
      d:=Denominator(i);j:=1;n1:=n;
      repeat k:=Gcd(n1,d);n1:=n1/k;j:=j*k; until k=1;
      return ret(E(j*d)^(Numerator(i)*GcdRepresentation(n1,d)[1]));
    fi;
    i:=CoeffsCyc(x,NofCyc(x));
    j:=Gcd(List(i,Numerator))/Lcm(List(i,Denominator));
    if AsRootOfUnity(x/j)<>false then
      return ret(GetRoot(x/j,n,"no")*GetRoot(j,n,"no"));
    fi;
    i:=Quadratic(x*Denominator(j));
    if i<>false and n=2 then
      k:=List([1,-1],f->(i.a+f*ER(i.a^2-i.root*i.b^2))/(2*i.root*i.d));
      k:=Filtered(List(Filtered(k,IsRat),GetRoot),x->x<>false);
      k:=List(k,x->i.b/(2*x*i.d)+ER(i.root)*x);
      if Length(k)>0 then return k[1]/ER(Denominator(j));fi;
    fi;
  elif IsRec(x) and IsBound(x.operations) and IsBound(x.operations.GetRoot) then
    return ret(x.operations.GetRoot(x,n));
  fi;
  if error then Error(msg,": unable to compute ",n,"-th root of ",x,"\n");fi;
  return false;
end;

############################################################################
# EvalPolRoot(pol, x, n, p)
#  evaluate polynomial pol at E(n)^p*x^(1/n)
#  
#  The point of this routine is to avoid unnecessary root extractions
#  during evaluation (e.g., if pol has no terms of odd degree and n=2,
#  then no root extraction is necessary).
#
EvalPolRoot:=function(pol,x,n,p)local P,l,r; 
  if IsPolynomial(pol) then
    if Length(pol.coefficients)=0 then return 0;fi;
    P:=Concatenation([1..pol.valuation mod n]*0,pol.coefficients);
    P:=List([0..n-1],i->Value(Polynomial(pol.baseRing,
           P{Filtered([1..Length(P)],j->(j-1) mod n=i)},
	   (pol.valuation-pol.valuation mod n)/n),x));
    pol:=Polynomial(pol.baseRing,P{[1..First([n,n-1..1],i->P[i]<>0*P[i])]},0);
    if Length(pol.coefficients)=0 then return 0;fi;
    l:=pol.valuation-1+Filtered([1..Length(pol.coefficients)],
	     i->pol.coefficients[i]<>0*pol.coefficients[i]);
    Add(l,n); r:=Gcd(l);
    pol:=Polynomial(pol.baseRing,
           pol.coefficients{[1,1+r..Length(pol.coefficients)]},
           pol.valuation/r);
    n:=n/r;
    return Value(pol,GetRoot(x,n)*E(n)^p);
  elif IsRec(pol) and IsBound(pol.operations) 
       and IsBound(pol.operations.EvalPolRoot)
  then return pol.operations.EvalPolRoot(pol,x,n,p);
  fi;
end;

###########################################################################
#  This returns the smallest i such that 
#  stst... (i terms)= tsts... (i terms)
# (loops indefinitely if no such i exists)
BraidRelation:=function(s,t)local p1,p2,i;
  p1:=s;p2:=t;i:=1;s:=[s,t];
  while p1<>p2 do
    p1:=p1*s[1+i mod 2];p2:=p2*s[2-i mod 2];
    i:=i+1;
  od;
  return i;
end;

# CheckBraidRelation(gens,rel[,f])
# Check that the braid relation rel[1]=rel[2] holds for gens (in particular
# that no shorter relation holds). If given, call f in case of failure.
CheckBraidRelation:=function(arg)local gens,rel,f,r,l,i,p;
  gens:=arg[1];rel:=arg[2];if Length(arg)=3 then f:=arg[3];else f:=Ignore;fi;
  p:=r->SPrint(IntListToString(r[1]),"=",IntListToString(r[2]));
  l:=gens[rel[1][1]];r:=gens[rel[2][1]];
  i:=1;
  while i<Length(rel[1]) do
    if l=r then 
      f(" relation ",p(rel)," already holds as ",p(List(rel,x->x{[1..i]})));
      return false;
    fi;
    i:=i+1;
    l:=l*gens[rel[1][i]];r:=r*gens[rel[2][i]];
  od;
  if l=r then return true;fi;
  f(" relation ",p(rel),"failed");return false;
end;

#############################################################################
# The       next      function      is       a      replacement      for
# Indeterminate(ApplyFunc(DefaultRing,V))   since   GAP   3.3   function
# DefaultRing  fails  for  DefaultRing(X(Cyclotomics),E(3))  instead  of
# returning X(Cyclotomics)

ChevieIndeterminate:=function(V)local domain,myin,elm;
  myin:=function(a,b)
    if IsRec(b) and IsBound(b.domain) then
      if a in b.domain then return true;
      else while IsBound(b.baseRing) do 
             b:=b.baseRing; if a in b then return true;fi;od;
      fi;
    fi;
    return false;
  end;
  if ForAll( V, IsInt )  then return X(Integers);
  elif ForAll( V, IsRat )  then return X(Rationals);
  elif ForAll( V, IsCyc )  then return X(Cyclotomics);
  else
    for elm  in V  do
      if ForAll( V, l-> myin(l,elm)) then return X(elm.domain); fi;
    od;
#   Error( "sorry, the elements of <V> lie in no common ring domain" );
    return Mvp("foo");
  fi;
end;
