#############################################################################
##
#A  coxeter.g            CHEVIE library          Meinolf Geck, Frank Luebeck, 
#A                                            Jean Michel and G\"otz Pfeiffer
##
#A  $Id: coxeter.g,v 1.2 1997/02/20 11:09:42 werner Exp $
##
#Y  Copyright (C) 1992 - 1996  Lehrstuhl D f\"ur Mathematik, RWTH Aachen, IWR
#Y  der Universit\"at Heidelberg, University of St. Andrews, and   University 
#Y  Paris VII.
##
##  This file contains general functions that deal with Coxeter groups.
##

# returns a reflectiontype + unparsable part of arg
ParseTypeFromArg:=function(arg)local res,t,shift;
  res:=[];
  shift:=function()local tmp; 
    tmp:=arg[1];arg:=arg{[2..Length(arg)]};return tmp;end;
  while Length(arg)>0 and IsString(arg[1]) and arg[1]<>[] do
    t:=rec(series:=shift(),rank:=shift(),operations:=ReflTypeOps); 
    if t.rank=1 then t.series:="A";fi;
    if t.series="B" then t.cartanType:=2;
    elif t.series="C" then t.series:="B";t.cartanType:=1;
    elif t.series="Bsym" then t.series:="B";t.cartanType:=ER(2);
    elif t.series="B?" then t.series:="B";t.cartanType:=shift();
    elif t.series="D" and t.rank=3 then t.series:="A";
    elif t.series="G" then t.series:="G";t.cartanType:=1;
    elif t.series="Gsym" then t.series:="G";t.cartanType:=ER(3);
    elif t.series="G?" then t.series:="G";t.cartanType:=shift();
    elif t.series="F" then t.series:="F";t.cartanType:=1;
    elif t.series="Fsym" then t.series:="F";t.cartanType:=ER(2);
    elif t.series="F?" then t.series:="F";t.cartanType:=shift();
    elif t.series="H" and t.rank=2 then t.series:="I";t.bond:=5;
    elif t.series="I" then t.bond:=shift();
      if t.bond mod 2=0 then t.cartanType:=1;fi;
    elif t.series="Isym" then t.bond:=shift();t.series:="I";
      t.cartanType:=E(2*t.bond)+E(2*t.bond)^-1;
    elif t.series="I?" then t.bond:=shift();t.series:="I";
      t.cartanType:=shift();
    fi;
    Add(res,t);
  od;
  return [res,arg];
end;

#############################################################################
##
#F  CartanMat( <type>, <rk>, ... )  
#F    returns a Cartan matrix of given type(s) and rank(s)
##
##  E.g., 'CartanMat( "F", 4 );'
##  For type I2(n), a third argument is needed: 'CartanMat( "I", 2, n )'.
##  If several couples <type>, <rk> are given, the direct sum of the
##  corresponding Cartan matrices is returned.
##
CartanMat:=function (arg) local t, res;
  if IsRec(arg[1]) then return ChevieDispatcher("CartanMat",arg);fi;
  return ApplyFunc(DiagonalMat,List(ApplyFunc(ParseTypeFromArg,arg)[1],
    CartanMat));
end;

#############################################################################
##
#F  CoxeterTypeFromCartanMat( <C> )  . . . . type of a Cartan matrix
##  
##  Returns  a  list  of records (series:=,indices:=,rank:=) which  describe 
##  irreducible components of the Cartan matrix <C>, such that
##    C{indices}{indices}=CartanMat(series,rank)
##  A field .bond is added for type "I".
##  For series B,G,I,F a field .cartanType is added containing the ratio
##  of one conjugacy class of roots to the other
##  
##  If <C> is not the Cartan matrix of a finite Coxeter group this function
##  returns 'false'. So it can be used to check a matrix for this property.
  
CoxeterTypeFromCartanMat:=C->List(DecomposedMat(C),
  function(s)local co,m,r,x,l,m1,t,tmp,rev;
    m:=C{s}{s}; r:=Length(m); l:=TwoTree(m);
    if l=false then return false;fi;
    t:=rec(operations:=ReflTypeOps,rank:=r);
    if not IsList(l[Length(l)]) then # line: types A,B,C,F,G,H,I
      m1:=m{l}{l};
      co:=i->m1[i][i+1]*m1[i+1][i];
      rev:=function()l:=Reversed(l);m1:=m{l}{l};end;
      if r=1 then t.series:="A";
      elif r=2 then
	if co(1)=1 then t.series:="A";
	elif co(1)=2 then t.series:="B"; 
          if m1[1][2]=-1 or (m1[1][2]<>-2 and m1[2][1]>m1[1][2]) then rev();fi;#B2 preferred to C2
	  t.cartanType:=m1[1][2]/-1;
	elif co(1)=3 then t.series:="G"; 
          if m1[1][2]<>-1 and m1[2][1]>m1[1][2] then rev();fi;
          t.cartanType:=m1[1][2]/-1;
	else x:=NofCyc(co(1));
          if m1[2][1]=-1 or (m1[1][2]<>-1 and m1[2][1]>m1[1][2])then rev();fi;
	  if co(1)=2+E(x)+E(x)^-1 then t.bond:=x; else t.bond:=2*x; fi;
	  if t.bond=1 then return false;fi;
	  t.series:="I";
	  if t.bond mod 2=0 then t.cartanType:=m1[1][2]/-1;fi;
	fi;
      else
        if co(r-1)<>1 then rev();fi;
	if co(1)=1 then
	  if co(2)=1 then t.series:="A";
	  elif r<>4 then return false;
	  else
	    if m1[2][3]<m1[3][2] then rev();fi;
	    t.series:="F"; t.cartanType:=m1[2][3]/-1;
	  fi;
	elif co(1)=2 then t.series:="B"; t.cartanType:=m1[1][2]/-1;
	elif not r in [3,4] then return false;
	elif co(1)=(3+ER(5))/2 then t.series:="H";
	else return false;
	fi; 
      fi;
      t.indices:=l;
    else # types D,E
      tmp:=l[1];l:=l{[2..Length(l)]};
      if Length(l[2])=1 then t.series:="D";
       t.indices:=Concatenation(l[1],l[2],[tmp],l[3]);
      elif not r in [6..8] then return false;
      else
       t.series:="E";
       t.indices:=Concatenation([l[2][2],l[1][1],l[2][1],tmp],l[3]);
      fi;
    fi;
    if CartanMat(t)<>m{t.indices}{t.indices} then return false;fi; 
       # countercheck if we get C from type
    t.indices:=s{t.indices};return t;
  end);

# JM  10-2-2001  : I  suppressed  the  sorting.  I understand  this  was
# done  to make  easier to  test isomorphism  between Coxeter  groups. I
# believe  this code  should be  confined to  the isomorphism  test. The
# reason  is  that it  affects  the  way  character tables  of  products
# are  constructed, which  caused  me  all kinds  of  trouble in  trying
# to  compute  HC and  Lusztig  induction  of unipotent  characters  (if
# one  wants to  re-introduce  this  sorting, it  should  be  done in  a
# separate routine  e.g. 'IsomorphismType'  so it  can be  called easily
# whenever one constructs  a group from its type --  otherwise this last
# operation becomes painful -- the  normalization should also be able to
# normalize types of  cosets, Spets, complex reflection groups  if it is
# re-introduced)
#
#  # Sorting via rank and alphabetically:
#  tmp:=List(type,a->[a.rank,a.series,Minimum(a.indices)]);
#  SortParallel(tmp,type,function(a,b)
#                         return a[1]>b[1] or (a[1]=b[1] and a[2]<b[2])
#                            or (a[1]=b[1] and a[2]=b[2] and a[3]<b[3]); 
#               end);

#############################################################################
#F  ReflectionType( <W> )  . type of a Reflection group or Spets
##  
##  returns  a  list  of records (series:=,indices:=,rank:=) 
##  which  describe  irreducible components of W.
##
ReflectionType:=function(C)
  # dispatcher function, if not called with Cartan matrix:
  if IsRec(C) then return ChevieAttributeDispatcher("ReflectionType","type",[C]);
  else return CoxeterTypeFromCartanMat(C);
  fi;
end;

IsomorphismType:=function(W)local t;
  t:=Collected(List(ReflectionType(W),ReflectionName));
  return String(Concatenation(List(t,function(x)
    if x[2]=1 then return x[1];else return SPrint(x[1],"^",x[2]);fi;end)));
end;

#############################################################################
##
#F  RootsIntCartan( <C> ) . . . . . . . . root system from Cartan matrix with
#F  integer entries
##  
##  
##  Returns the set of roots determined by  <C>, as linear combinations of
##  the simple  ones.  The first half  are the  positive roots ordered  by
##  their height.  The second half   contains  the corresponding  negative
##  roots in the same order.
##  
##  The simple roots come first and they are ordered as given by <C>.
##  
RootsIntCartan:=function(C)
  local RootsIntCartan1, a, l, n, p, v, i, x, tmp, res;
  
  # returns list l where l[i] is the list of positive roots of height i:
  RootsIntCartan1:=function(C) local res, I, l, a, pmq, i, j;
    l:=Length(C);
    I:=C^0;
    res:=[I];
    res[2]:=[];
    for i in [1..l] do
      for j in [i+1..l] do
        if C[i][j]<>0 then Add(res[2],res[1][i]+res[1][j]); fi;
      od;
    od;
    i:=2;
    while res[i]<>[] do
      res[i+1]:=[];
      for a in res[i] do
        for j in [1..l] do
          pmq:=C[j]*a;
          if pmq<0 or i-pmq-1>0 and a[j]>pmq and 
                                          a-(pmq+1)*I[j] in res[i-pmq-1] then 
            if not (a+I[j] in res[i+1]) then Add(res[i+1],a+I[j]); fi;  
          fi;
        od;
      od;
      i:=i+1;
    od;
    Unbind(res[i]);

    return res;  
  end;

  l:=Length(C);
  n:=0*[1..l];
  
  # to compute the irreducible subsytems separately:
  p:=DecomposedMat(C);

  # simple roots must be ordered as given (res[i] becomes the 
  # set of roots of height i): 
  res:=[C^0];
  for a in p do
    tmp:=RootsIntCartan1(C{a}{a});
    for i in [2..Length(tmp)] do
      if not IsBound(res[i]) then res[i]:=[]; fi;
      for x in tmp[i] do v:=ShallowCopy(n); v{a}:=x; Add(res[i],v); od;
    od;
  od;
  
  res:=Concatenation(res);
  return rec(roots:=Concatenation(res,-res));
end;
    
#############################################################################
##
#F  RootsAndMatgens( <C> ) . . the root system of the Cartan Matrix <C>
##
##  Returns a record with fields .cartan, .matgens, .roots, .N, 
##    .semisimpleRank
##  cartan         : the Cartan matrix <C>,
##  semisimpleRank : the size of <C>
##  roots          : the root vectors
##  N              : the number of positive roots
##  matgens        : the matrices of the reflections w.r.t. the simple roots
##  
RootsAndMatgens := function(C)local W,R, a, i, v, e;
  W:=rec(cartan:=C);
  W.semisimpleRank:=Length(W.cartan);
  if Length(W.cartan)=0 then W.matgens:=[];W.roots:=[];
  else 
    e := W.cartan^0;
    W.matgens := List([1..W.semisimpleRank], i->Reflection(e[i],W.cartan[i]));
    if ForAll( Concatenation( W.cartan ), IsInt )  then
        R:=RootsIntCartan(W.cartan);
#       W.degrees:=R.degrees;
        R:=R.roots;
    else
        R := W.cartan^0;
        for a  in R  do
            for i  in [1..Length(W.matgens)] do
                if a <> R[i] then
                    v := a * W.matgens[i];
                    if not v in R  then Add( R, v ); fi;
                fi;
            od;
        od;
        R:=Concatenation(R,-R);
    fi;
    W.roots:=R;
  fi;
  W.N := Length( W.roots )/2;
  return W;
end;

########################################################################
##
#F  HighestShortRoot( <W> ) . . . . . . . . . . . . . highest short root 
##
##  If <W> is irreducible then 'HighestShortRoot' computes the unique 
##  short root of maximal height. Note that if all roots have the same 
##  length then this is the  unique root of maximal height. The latter 
##  can be extracted by  W.roots[W.N]. If <W> is not irreducible then
##  an error message is returned.
##
HighestShortRoot:=function(W)
  if Length(ReflectionType(W))=1 then 
    return First([W.N,W.N-1..1],i->W.rootLengths[i]=1);
  else Error(" group is not irreducible\n");
  fi;
end;

#############################################################################
##
#F  SimpleRootsSubsystem( <W>, <l> ) . . . . . . . simple roots for subsystem
#F  of reflection subgroup
##  
##  <l> must  be a  subset of [1..2*Parent(<W>).N]. 'SimpleRootsSubsystem'
##  returns the  set of simple roots of  the root system of the reflection
##  subgroup              of       Parent(<W>),      generated          by
##  Reflections(Parent(<W>)){<l>}.
##  
SimpleRootsSubsystem:=function(W,l)
  local orb, refperm, gen, x, tmp, i, j, k, n, s, rk, ll, bl, erg;
  
  ll:=Set(l);
  if IsBound(W.parent) then W:=Parent(W); fi;
  
  # trivial case (fast handling of parabolic subgroups):
  if IsSubset([1..W.semisimpleRank],ll) then return ll; fi;
  
  refperm:=Reflections(W);refperm:=Concatenation(refperm,refperm);
  gen:=refperm{ll};
  
  # compute orbit:
  tmp:=Set(Concatenation(
               PermGroupOps.Orbits(rec(generators:=gen),ll,OnPoints)));
  orb:=tmp{[1..Length(tmp)/2]};
  
  erg:=Set([]);
  
  # first check input (get the result faster, if input is set of simple roots):
  for i in ll do
    if i<=W.N then
      n:=0;
      for j in orb do if n<2 and j^refperm[i]>W.N then n:=n+1; fi; od;
      if n=1 then AddSet(erg,i); fi;
    fi;  
  od;
  if erg=ll then return erg; fi;
  
  for i in orb do
    if not i in ll then
      n:=0;
      for j in orb do if n<2 and j^refperm[i]>W.N then n:=n+1; fi; od;
      if n=1 then AddSet(erg,i); fi;
    fi;  
  od;
  
  return erg;
end;

#############################################################################
##
#V  CoxeterGroupOps  . . . . . .  operation record for finite Coxeter groups
##
CoxeterGroupOps := OperationsRecord("CoxeterGroupOps",ComplexGroupOps);
Inherit(CoxeterGroupOps,AbsCoxOps);
# this overrides from ComplexGroupOps 
# [ "CartanMat", "Elements", "EltWord", "Hecke", "ReducedInRightCoset", 
#   "ReflectionSubgroup", "ReflectionType"]
# we want to keep:
CoxeterGroupOps.CartanMat:=ComplexGroupOps.CartanMat;

#############################################################################
##
#F  CoxeterGroupOps.String( <W> ) . . . . . . . . . . . 'String' method
##  
CoxeterGroupOps.String:=function(W)local res;
  if IsBound(W.parent) then
    res:=SPrint("ReflectionSubgroup(",W.parent,", ",
                               W.rootInclusion{W.generatingReflections});
    if not IsBound(W.omega) or Size(W.omega)=1 then Append(res,")");
    else PrintToString(res,", ",String(W.omega),")");
    fi;
    return res;
  else return SPrint("CoxeterGroup(",Join(List(W.createArgs,function(a)
      if IsString(a) then return SPrint("\"",a,"\"");
      else return a; fi;end)),")");
  fi;
end;

#############################################################################
##
#F  Reflections( <W> )  . . . . . . . . . . . .  the reflections in W
##
## 'Reflections'  returns  the set of all reflections in the Coxeter group
## <W>, as a list of permutations. The i-th entry in this list
##  is the reflection along the i-th root in <W>.roots.
##  
CoxeterGroupOps.Reflections:=function(W)local i;
  if Length(W.reflections)<2*W.N then
    W.reflections{[W.N+1..2*W.N]}:=List([1..W.N],i->Reflection(W,i));
  fi;
  return W.reflections{[1..W.N]};
end;
#############################################################################
##
#F  CoxeterLength( <W> , <w> )  . . . . . length of a permutation w in W
##
##  Counts the number of positive roots sent by w to negative roots.
##  
CoxeterGroupOps.CoxeterLength:=function(W,w)local res,N,r;
# return Number(OnTuples(W.rootInclusion{[1..W.N]},w),r->r>W.parentN);
  res:=0;N:=W.parentN; # Code below is 3.5 times faster than above line
  for r in OnTuples(W.rootInclusion{[1..W.N]},w) do
    if r>N then res:=res+1;fi;
  od;
  return res;
end;

CoxeterGroupOps.BaseX:=function(W)local tmp;
  if not IsBound(W.forMatX) then
    tmp:=W.simpleRoots{[1..W.semisimpleRank]};
    if W.semisimpleRank<W.rank then
      if Length(tmp)=0 then tmp:=IdentityMat(W.rank);
      else Append(tmp,NullspaceMat(TransposedMat(W.simpleCoroots)));
      fi;
    fi;
    W.forMatX:=[0*[W.semisimpleRank+1..W.rank],
                IdentityMat(W.rank){[W.semisimpleRank+1..W.rank]},
                tmp,tmp^-1];
  fi;
  return W.forMatX[3];
end;

#############################################################################
##
#F  MatXPerm( <W>, <w> )  . . . . . . . . . . . . convert a permutation to
#F  corresponding matrix operating on X
#F  MatYPerm( <W>, <w> )  . . . . . . . . . . . . convert a permutation to
#F  corresponding matrix operating on Y
##  
##  Let <w>  be a  permutation with the  following property: The images of
##  the        simple        roots         of       <W>      (       i.e.,
##  <W>.rootInclusion{[1..<W>.semisimpleRank]}  )  must be roots of  <W> (
##  i.e., in  the set <W>.rootInclusion ).   
##  
##  Let X_1 be the  sublattice of X  consisting of the elements orthogonal
##  to  all   coroots of <W>.    'MatXPerm'  returns the unique invertible
##  matrix which, as matrix acting  on X, maps the  simple roots of <W> as
##  indicated by <w> and which induces the identity map on X_1.
##  
##  If in particular <w> is  an element of  <W> then 'MatXPerm( <W>, <w> )
##  is  the matrix   representing <w> as  linear map   on X. This  follows
##  immediately from the formula for the generating reflections of <W>.
##  
##  'MatYPerm(  <W>, <w> )'  returns   the transposed of 'MatXPerm(   <W>,
##  <w>^-1 )'.  If in particular <w> is an element  of <W> then 'MatYPerm(
##  <W>, <w> )' is the matrix representing <w> as linear map on Y.
##  
CoxeterGroupOps.MatXPerm:=function(W, w)local tmp, M;
  
  if W.semisimpleRank=0 then return IdentityMat(W.rank);fi;

  tmp:=CoxeterGroupOps.BaseX(W);
  tmp:=W.forMatX;
    # operation is trivial on space orthogonal to all simple coroots
  M:=List(W.roots{W.rootRestriction{OnTuples(W.rootInclusion{
            [1..W.semisimpleRank]},w)}},i->Concatenation(i,tmp[1]));
  return tmp[4]*Concatenation(M,tmp[2])*tmp[3];
end;

MatYPerm:=function(W,w)
  return TransposedMat(MatXPerm(W,w^-1));
end;

#############################################################################
##
#F  AddComponentsCoxeterGroup ( ... )  only used internally by 'CoxeterGroup'
#F  and 'ReflectionSubgroup'
##  
# arguments: res [,parent]
AddComponentsCoxeterGroup :=function(arg) local res, parent, tmp, i, p, f;

  res:=arg[1];
  if Length(arg)>1 then parent:=arg[2]; else parent:=res; fi;

  if IsBound(parent.rank) then
    res.rank:=parent.rank;
  elif res.semisimpleRank>0 then
    res.rank:=Length(res.simpleRoots[1]);
  else
    res.rank:=0;
  fi;

  # root lengths from parent group, if already known:
  if IsBound(parent.rootLengths) then
    res.rootLengths:=parent.rootLengths{res.rootInclusion};
  fi;

  # for each root in a parent group we determine a simple root in 
  # the same W-orbit:
  tmp:=PointsAndRepresentativesOrbits(res,2*res.parentN);
  res.orbitRepresentative:=[];
  res.orbitRepresentativeElements:=[];
  for i in [1..Length(tmp[1])] do 
    if tmp[1][i][1] in res.rootInclusion then
      res.orbitRepresentative{res.rootRestriction{tmp[1][i]}}
                                                :=0*tmp[1][i]+tmp[1][i][1];
      res.orbitRepresentativeElements{res.rootRestriction{tmp[1][i]}}
                                                               :=tmp[2][i];
    fi;
  od;

  f:=function(t)local i,j,k,c;
    i:=t.indices;
    c:=CartanMat(t);
    for j in [1..Length(i)] do for k in [1..j-1] do
      if c[j][k]<>c[k][j] then
	res.rootLengths[res.orbitRepresentative[i[j]]]:=-c[k][j];
	res.rootLengths[res.orbitRepresentative[i[k]]]:=-c[j][k];
	return;
      fi;   
    od;od;
    for j in i do res.rootLengths[res.orbitRepresentative[j]]:=1;od;
  end;

  # root lengths (for parent group):
  if not IsBound(res.rootLengths) then
    res.rootLengths:=[];
    # first for the simple roots using the classification:
    for tmp in ReflectionType(res) do f(tmp);od;
    for i in [1..2*res.N] do
      res.rootLengths[i]:=res.rootLengths[res.orbitRepresentative[i]];
    od;
  fi;
  
  # the set of coroots: i-th entry is coroot to roots[i]
  if res.cartan=TransposedMat(res.cartan) then
    res.coroots:=res.roots;
  elif not IsBound(parent.coroots) then
    
    # the dual group as permutation group on the coroots:
    tmp:=RootsAndMatgens(TransposedMat(res.cartan));
    tmp.identity:=();
    tmp.reflections:=List(tmp.roots*tmp.matgens,SortingPerm)/
                                                   SortingPerm(tmp.roots);
    res.coroots:=[];
    for i in [1..res.N] do
      p:=res.orbitRepresentative[i];
      p:=p^Product(tmp.reflections{CoxeterWord(res,res.orbitRepresentativeElements[i])});
      Add(res.coroots,tmp.roots[p]);
    od;
    Append(res.coroots,-res.coroots);
  else
    tmp:=[parent.coroots{res.rootInclusion},
          RootsAndMatgens(TransposedMat(res.cartan)).roots{[1..res.N]},
          parent.coroots{res.rootInclusion{[1..res.semisimpleRank]}}];
    res.coroots:=[];
    for p in tmp[2] do
      res.coroots[Position(tmp[1],p*tmp[3])]:=p;
    od;
    Append(res.coroots,-res.coroots);
  fi;

  # Order of the Coxeter group 
  # (This doesn't use a stabilizer chain but allows a much faster 
  # construction of a stabilizer chain)
  # here also the 'ReflectionType' is computed.
  if ReflectionType(res)=false then InfoChevie("# Unknown Cartan Matrix\n");
  else Size(res);
  fi;

  res.name:=ReflectionName(res);
end;

#############################################################################
##  
#F  CoxeterGroup( <cartan>[, <omega>])
#F  CoxeterGroup( <roots>, <coroots>[, <omega>])
#F  CoxeterGroup( <type>, <rank>[, <type>, <rank> ...][, "sc"][, <omega>] )
#F  CoxeterGroup( <rec> )
##  
CoxeterGroup:=function(arg)local l, sc, om, W, typ, a, tmp;

  l:=Length(arg);
  
  # get back the .coxeter entry of a record:
  if l=1 and IsRec(arg[1]) then
    if IsBound(arg[1].coxeter) then return arg[1].coxeter;
    elif IsBound(arg[1].operations) and IsBound(arg[1].operations.CoxeterGroup)
    then return arg[1].operations.CoxeterGroup(arg[1]);
    fi;
  fi;
  
  if l>0 and IsGroup(arg[l]) then
    om:=arg[l];
    l:=l-1;
  fi;
  if l>0 and arg[l]="sc" then
    sc:=true;
    l:=l-1;
  else
    sc:=false;
  fi;

  W:=rec();
  if l=2 and IsMat(arg[1]) then
    if sc then 
      Error("CoxeterGroup: argument \"sc\" not allowed ",
            "together with 2 matrices\n");
    fi;
    W.simpleRoots:=arg[1];
    W.simpleCoroots:=arg[2];
    W.cartan:=W.simpleCoroots*TransposedMat(W.simpleRoots);
  else
    if l=0 then W.cartan:=[];
    elif IsString(arg[1]) or IsRec(arg[1]) then
	W.cartan:=ApplyFunc(CartanMat,arg{[1..l]});
    else W.cartan:=arg[1];
    fi;
    if Length(W.cartan)=0 then W.simpleRoots:=[];W.simpleCoroots:=[];
    elif sc then
      W.simpleRoots:=TransposedMat(W.cartan); W.simpleCoroots:=W.cartan^0;
    else
      W.simpleRoots:=W.cartan^0; W.simpleCoroots:=W.cartan;
    fi;
  fi;

  tmp:=RootsAndMatgens(W.cartan);Inherit(W,tmp);
  W.rootInclusion := [1..2*W.N];
  W.rootRestriction:=[1..2*W.N];
  W.parentN:=W.N;
  W.isCoxeterGroup := true;

  if W.roots=[] then
    tmp:=Group([],()); # because []*[] not allowed in GAP!
  else
    tmp:=Group(List(W.roots*W.matgens,SortingPerm)/SortingPerm(W.roots),());
  fi;

  Inherit(tmp,W);W:=tmp;

# new stuff. JM 17-3-99
  W.reflections:=ShallowCopy(W.generators);
  W.nbGeneratingReflections:=Length(W.reflections);
  W.generatingReflections:=[1..W.nbGeneratingReflections];
  W.reflectionsLabels:=W.rootInclusion{W.generatingReflections};

  W.operations:=CoxeterGroupOps;
  AddComponentsCoxeterGroup(W);

  if IsBound(om) then
    if IsPermGroup(om) then
      if W.semisimpleRank<> W.rank then
        Error("#I Omega can be permutation ",
              "group only for semisimple datum.\n");
      fi;
      om:=Group(List(om.generators,x->MatXPerm(W,x)), IdentityMat(W.rank));
    fi;
    if not ForAll(om.generators,M->IsNormalizing(W.roots*W.simpleRoots,M) and
               IsNormalizing(W.coroots*W.simpleCoroots,TransposedMat(M))) then
      Error("#I Generators of Omega should normalize roots and coroots.\n");
    fi;
    W.omega:=om;
    tmp:=W.roots*W.simpleRoots;
    W.omega.elementsPerm:=List(Elements(W.omega),M->PermListList(tmp,tmp*M));
  fi;
  
  # for printing:
  # simplify arguments: first check if one matrix in argument is superfluous:
  if l=2 and W.rank=W.semisimpleRank and IsMat(arg[1]) then
    tmp:=IdentityMat(W.rank);
    if arg[1]=tmp then
      arg:=arg{[2..Length(arg)]};
      l:=1;
    elif arg[2]=tmp then
      arg[1]:=TransposedMat(arg[1]);
      arg[2]:="sc";
      l:=1;
    fi;
  fi;

  # check in case of one matrix if it can be created by 'CartanMat':
  if l=1 then
    tmp:=List(ReflectionType(W),ShallowCopy);
    Sort(tmp,function(a,b) return a.indices<b.indices; end);
    typ:=[];
    for a in tmp do
      if typ<>false then
        if a.series in ["B","C"] and Length(a.indices)=2 then
          if a.indices[1]+1=a.indices[2] then
            Append(typ,[a.series,a.rank]);
          elif a.indices[2]+1=a.indices[1] then
            Append(typ,Difference(["B","C"],[a.series]));
            Add(typ,2);
          else
            typ:=false;
          fi;
        else
          if a.indices=[a.indices[1]..a.indices[1]+a.rank-1] then
            if a.series<>"I" then
              Append(typ,[a.series,a.rank]);
            else
              Append(typ,[a.series,a.rank,a.bond]);
            fi;
          else
            typ:=false;
          fi;
        fi;
      fi;
    od;
    if typ<>false then
      arg:=Concatenation(typ,arg{[2..Length(arg)]});
    fi;
  fi;
  W.createArgs:=ShallowCopy(arg);

  AbsCoxOps.CompleteCoxeterGroupRecord(W);
  Inherit(W.operations,CoxeterGroupOps);
  return W;
end;

#############################################################################
##
#F  OmegaSubgroup( <W>, <ws>[, <oms> ]) . . . . . . . . . . . . . subgroup of 
#F  < <W>, <W>.omega > as matrix group
##  
##  <ws> must be list of  elements of <W>,  <oms> must be list of elements
##  of  <W>.omega.  'OmegaSubgroup'  returns  the   matrix group (matrices
##  acting on X) generated by the elements in <ws> and <oms>.
##  
OmegaSubgroup:=function(arg)
  local W, wgen, ogen, gen, tmp, erg;
  
  if IsBound(arg[1].parent) then
    W:=arg[1].parent;
  else
    W:=arg[1];
  fi;
  
  wgen:=arg[2];
  if Length(arg)>2 then
    ogen:=arg[3];
  else
    ogen:=[];
  fi;
  
  if not ForAll(wgen,x->x in W) and (
          (IsBound(W.omega) and ForAll(ogen,x->x in W.omega)) or ogen=[]) then
    Error("#I second argument must be list of elements in parent of",
          " first argument W,\n",
          "#I third argument must be list of elements of W.omega.\n");
  fi;
  erg:=Group(Concatenation(List(wgen,x->MatXPerm(W,x)),ogen),
             IdentityMat(W.rank));
  tmp:=W.roots*W.simpleRoots;
  erg.elementsPerm:=List(Elements(erg),M->PermListList(tmp,tmp*M));
  
  erg.Wgenerators:=wgen;
  erg.Omegagenerators:=ogen;
  return erg;
end;

#############################################################################
##
#F  ReflectionSubgroup( <W>, <gens>[, <Omega> ] ) . . . . . 
##  

CoxeterGroupOps.ReflectionSubgroup:=function(arg)
  local W, simpleroots, tmp, Om, res, i, n,  r, gens, pos;
  
  if IsBound(arg[1].parent) then W:=arg[1].parent;
  else W:=arg[1];
  fi;
  
##  gens:=arg[1].rootInclusion{arg[2]};
  # this is now changed, generators must be given as reflections of the
  # parent group:
  gens:=arg[2];
  if IsSubset(W.reflectionsLabels,gens) then
    gens:=List(gens,x->W.rootInclusion[Position(W.reflectionsLabels,x)]);
  elif not IsSubset(W.rootInclusion,gens) then
    Error("argument should be labels for simple reflexions or a subset of W.rootInclusion");
  fi;
  
  if Length(arg)>2 then Om:=arg[3]; fi;
  
  # checking if subgroup in cache:
  res:=rec(callarg:=[gens]);
  if IsBound(Om) then Add(res.callarg,Om.generators); fi;
  res:=CHEVIE.GetCached(W,"ReflectionSubgroups",res,function(x,res)return
    x.callarg=res.callarg;end);
  if IsBound(res.isGroup) then return res;fi;
  
  # determine a set of simple roots:
  simpleroots:=SimpleRootsSubsystem(W,gens);
  if Set(gens)=simpleroots and Length(gens)=Length(Set(gens)) then
    simpleroots:=gens;
  fi;
  
  gens:=Reflections(W){simpleroots};

  # The command 'Subgroup' returns W, if Set(gens)=Set(W.generators).
  # To avoid this we substitute it by some more lines.
  res.isDomain   := true;
  res.isGroup    := true;
  res.parent     := W;
  res.identity   := ();
  res.generators := ShallowCopy(gens);
  res.isPermGroup:= true;
  for i in [1..Length(gens)] do
    res.(i):=gens[i];
  od;
 
  res.isCoxeterGroup:=true;
  res.parentN:=W.N;

  # case of empty simpleroots:
  if simpleroots=[] then
    res.simpleRoots:=[];
    res.simpleCoroots:=[];
    res.cartan:=[];
    res.matgens:=[];
    res.roots:=[];
    res.N:=0;
    res.rootInclusion:=[];
    res.rootRestriction:=[];
    res.semisimpleRank:=0;
    res.reflections:=[];
  else
    res.simpleRoots:=W.roots{simpleroots}*W.simpleRoots;
    res.simpleCoroots:=W.coroots{simpleroots}*W.simpleCoroots;
    res.cartan:=W.coroots{simpleroots}*
                        W.cartan*TransposedMat(W.roots{simpleroots});
    Inherit(res,RootsAndMatgens(res.cartan));

    if res.roots<>[] then r:=res.roots*W.roots{simpleroots};
                   else r:=[];
    fi;
    res.rootInclusion:=ListBlist([1..2*W.N],BlistList(W.roots,r));
    res.rootInclusion:=Permuted(res.rootInclusion,
                            PermListList(r,W.roots{res.rootInclusion}));

    res.rootRestriction:=[];
    res.rootRestriction{res.rootInclusion}:=[1..2*res.N];

    res.reflections:=W.reflections{res.rootInclusion};
  fi;
  
  if IsBound(Om) then
    tmp:=Set(res.rootInclusion);
    if not ForAll(Om.elementsPerm,x->
                             Set(OnTuples(res.rootInclusion,x))=tmp) then
      Error("#I Omega as permutation group must normalize the ",
            "CoxeterGroup subgroup.\n");
    fi;
    res.omega:=Om;
  fi;
  

  res.operations:=CoxeterGroupOps;
  AddComponentsCoxeterGroup (res,W);

# new stuff. JM 17-3-99
  res.nbGeneratingReflections:=res.semisimpleRank;
  res.reflectionsLabels:=res.rootInclusion{[1..res.nbGeneratingReflections]};
  AbsCoxOps.CompleteCoxeterGroupRecord(res);
  Inherit(res.operations,CoxeterGroupOps);
  return res;
end;

#############################################################################
##
#F  LeftDescentSet( W, x)  the set of generators s such that sx < x
##  
##  the generators are numbered by the corresponding reflections of the parent
##  group
##  
CoxeterGroupOps.LeftDescentSet:=function(W,w) 
 return Filtered(W.rootInclusion{W.generatingReflections},i->i^w>W.parentN);
end;

#############################################################################
##
#F  CoxeterGroupOps.ReflectionType( <W> ) . . . . . type for CoxeterGroup record
##  
##  This is essentially the same as ReflectionType(<W>.cartan), but with
##  one exception: A subsystem  of type "A" is said to  be of type "~A",
##  iff the following conditions are fulfilled:
##  
##   - <W> is not a parent group
##   - the subsystem is contained in an irreducible subsystem of 
##     Parent(<W>), which is of type "B", "C", "G" or "F"
##   - the subsystem consists of short roots
##  
CoxeterGroupOps.ReflectionType:= function(W)
  local p, a, b, res;
  res:=ReflectionType(W.cartan);
  
  if false in res then InfoChevie("# Unknown Cartan matrix\n");return false;fi;

  # parent or all roots of parent have same length:
  if not IsBound(W.parent) or Length(W.parent.rootLengths)=0 or
    Maximum(W.parent.rootLengths)=1 then return res;
  fi;
  
  for a in res do
    p:=W.rootInclusion[a.indices[1]];
    if a.series="A" and Parent(W).rootLengths[p]=1 then
      p:=Parent(W).orbitRepresentative[p];
      for b in ReflectionType(Parent(W)) do
        if p in b.indices and b.series in ["B","C","F","G"] then
          a.tilde:=true;
        fi;
      od;
    fi;
  od;
  
  return res;
end;
        
#############################################################################
##
#F  CoxeterGroupOps.FusionConjugacyClasses( <W1>, <W2> )  . . . . . overwrite
#F  'FusionConjugacyClasses' for Coxeter groups
##  
##  'FusionConjugacyClasses'   is overwritten   for permutation groups  by
##  reading classinv.g.
##  
CoxeterGroupOps.FusionConjugacyClasses:=PermGroupOps.FusionConjugacyClasses;

#############################################################################
##
#F  CoxeterGroupOps.\=( <W1>, <W2> )  . . . equality test for Coxeter groups
##  
##  Here  two CoxeterGroup records  are defined to  be equal if  they are 
##  equal as permutation  groups, and if the simple  roots,  the simple 
##  coroots and the groups .omega are equal.
##  
if false then
CoxeterGroupOps.\= := function(W1,W2)
  if not (IsRec(W1) and IsRec(W2) and IsCoxeterGroup(W1) and IsCoxeterGroup(W2))
  then return false;
  elif W1.simpleRoots<>W2.simpleRoots 
                     or W1.simpleCoroots<>W2.simpleCoroots then
    return false;
  elif PermGroupOps.\=(W1,W2)=false then
    return false;
  elif IsBound(W1.omega) then
    if IsBound(W2.omega) then
      return W1.omega=W2.omega;
    else
      return Size(W1.omega)=1;
    fi;
  else
    if IsBound(W2.omega) then
      return Size(W2.omega)=1;
    else
      return true;
    fi;
  fi;
end;
fi;

#############################################################################
##
#F  ReflectionDegrees( <W>)  . . . . . . . degrees as a reflection group
##
##  Returns the degrees of the  reflection group W,
##  as an increasing list of integers.
##  

# The method is  no more used now.  HasTypeOps.ReflectionDegrees is used
# instead. This stays here for  documenting another algorithm, which may
# be  useful  for  checking  purposes.  Note  that  for  non-irreducible
# groups this  function returns the  reflection degrees sorted,  wich is
# incompatible  with  HasTypeOps.ReflectionDegrees  and  thus  with  the
# assumption in PhiFactors in spets.g .
if false then
CoxeterGroupOps.ReflectionDegrees:=function(W)local p,res;
  if not ForAll(Concatenation(W.cartan),IsInt)  then 
    Error("no algorithm for non-Weyl Coxeter groups");
  fi;
# Crystallographic group: general algorithm
    p:=List(Collected(List(W.roots{[1..W.N]},Sum)),x->x[2]);
    if p=[] then res:= [];
    else res:= Reversed(1+List([1..Maximum(p)],x->Number(p,y->y>=x)));
    fi;
    return Concatenation(List([W.semisimpleRank+1..W.rank],x->1),res);
end;
fi;
    
CoxeterGroupOps.ReflectionCoDegrees:=W->ReflectionDegrees(W)-2;

#############################################################################
##
#F  ReflectionCharValue( <W>, <w> ) ... The reflection character on w
##
##  'ReflectionCharValue' returns  the value of the reflection character of W 
##  at the element w.
##  
CoxeterGroupOps.ReflectionCharValue:=function(W,w)
  return W.rank-W.semisimpleRank+Sum([1..W.semisimpleRank],
     i->W.roots[W.rootRestriction[W.rootInclusion[i]^w]][i]);
end;

CoxeterGroupOps.IsLeftDescending:=function(W,w,i)
   return W.rootInclusion[i]^w>W.parentN; end;

CoxeterGroupOps.FirstLeftDescending:=function(W,w)local i;
  for i in W.generatingReflections do
    if W.rootInclusion[i]^w>W.parentN then return i;fi;
  od;
  return false;
end;

#############################################################################
##
#F  PermMatX( <W> , <M> )  . . . . . . . . .  convert a matrix in X which
#F  preserves the roots to permutation of the roots of Parent(<W>)
#F  PermMatXCoxeterElement( <W> , <M> )  . . . . convert a matrix in X which
#F  represents an element of Parent(<W>) to corresponding permutation
#F  PermMatY( <W> , <M> )  . . . . . . . . .  convert a matrix in Y which
#F  preserves the coroots to permutation of the roots of Parent(<W>)
##  

# this is faster than 'PermMatX', but works for elements of the Coxeter 
# group only:
PermMatXCoxeterElement:=function(W, M) local tmp;
  
  # we always have to look at the parent group:
  if IsBound(W.parent) then W:=W.parent; fi;

  # the images of the points Base(W) determine the element uniquely:
  tmp:=W.roots*W.simpleRoots;
  tmp:=List(Base(W),p->Position(tmp,tmp[p]*M));
  return RepresentativeOperation(W,Base(W),tmp,OnTuples);
end;

CoxeterGroupOps.PermMatX:=function(W, M) local tmp;

  # we always have to look at the parent group:
  if IsBound(W.parent) and W<>W.parent then
    return W.parent.operations.PermMatX(W.parent,M);
  fi;

  tmp:=W.roots*W.simpleRoots;
  return PermListList(tmp,tmp*M);
end;

PermMatY:=function(W, M)
  return PermMatX(W, TransposedMat(M))^-1;
end;

CoxeterGroupOps.ProductRootEmbed:=function(W1,W2)local p;
  p:=SortingPerm(List(Concatenation(W1.roots,W2.roots),
       function(x)local s;s:=Sum(x);if not IsInt(s) then s:=evalf(s);fi;
         if s>0 then return s; else return W1.N+W2.N-s;fi;end));
  return [OnTuples([1..2*W1.N],p),OnTuples(2*W1.N+[1..2*W2.N],p)];
end;

CoxeterGroupOps.\*:=function(W1,W2)local g;
  g:=ComplexGroupOps.\*(W1,W2);
  if not IsCoxeterGroup(W1) then return g;fi;
  g.cartan:=DiagonalMat(W1.cartan,W2.cartan);
  g.simpleRoots:=Concatenation(
     List(W1.simpleRoots,x->Concatenation(x,[1..W2.rank]*0)),
     List(W2.simpleRoots,x->Concatenation([1..W1.rank]*0,x)));
  g.simpleCoroots:=Concatenation(
     List(W1.simpleCoroots,x->Concatenation(x,[1..W2.rank]*0)),
     List(W2.simpleCoroots,x->Concatenation([1..W1.rank]*0,x)));
  g.N:=W1.N+W2.N;
  g.parentN:=g.N;
  AbsCoxOps.CompleteCoxeterGroupRecord(g);
  Inherit(g.operations,CoxeterGroupOps);
  AddComponentsCoxeterGroup(g);
  return g;
end;

# the lines below trick the auto system to do the right thing
CoxeterGroupOps.Hecke:=CoxeterGroupOpsHecke;
CoxeterGroupOps.RelativeGroup:=CoxeterGroupOpsRelativeGroup;
CoxeterGroupOps.ClassInvariants:=CoxeterGroupOpsClassInvariants;
