#############################################################################
##
#A  dispatch.g               CHEVIE library                       Jean Michel
##
#Y  Copyright (C) 1992 - 1999  Lehrstuhl D fur Mathematik, RWTH Aachen,
#Y  University of St. Andrews, and  University Paris VII.
##
##   This file contains generic functions defined in the package 'chevie'
##   Jean Michel 7-9-96. Uniformized 4-99
##
##  Note: Inheritance diagram for Chevie types:
##
##          PermGroupOps
##              |
##          PermRootOps    HasTypeOps   
##              |          /      |      
##              |         /       |      
##            ComplexGroupOps   SpetsOps
##                     |          |
##                     |        CoxeterCosetOps
##                     |
##                     |        AbsCoxOps  MatGroupOps
##                     |         /   \      |
##                     |        /     \     |
##                 CoxeterGroupOps    GenCoxOps
##                                        |
##                                  AffineCoxeterGroupOps
##
##  in additions AbsCox groups which happen to be finite inherit from HasType

ChevieDispatcher:=function(name,a)local obj;
  obj:=a[1];
  if IsRec(obj) and IsBound(obj.operations) and IsBound(obj.operations.(name))
  then return ApplyFunc(obj.operations.(name),a);
  else Error(obj," has no method for ",name);
  fi;
end;
  
AlphaInvolution:=
  function(arg)return ChevieDispatcher("AlphaInvolution",arg);end;

AltInvolution:=
  function(arg)return ChevieDispatcher("AltInvolution",arg);end;

BetaInvolution:=
  function(arg)return ChevieDispatcher("BetaInvolution",arg);end;

BraidRelations:=
  function(arg)return ChevieDispatcher("BraidRelations",arg);end;

BrieskornNormalForm:=
  function(arg)return ChevieDispatcher("BrieskornNormalForm",arg);end;

CharName:=function(arg)
  if Length(arg)=2 then Add(arg,rec());fi;
  return ChevieDispatcher("CharName",arg);end;

Coefficient:=
  function(arg)return ChevieDispatcher("Coefficient",arg);end;

CoxeterElements:=
  function(arg)return ChevieDispatcher("CoxeterElements",arg);end;

CoxeterLength:=
  function(arg)return ChevieDispatcher("CoxeterLength",arg);end;

CoxeterWord:=
  function(arg)return ChevieDispatcher("CoxeterWord",arg);end;

CoxeterWords:=
  function(arg)return ChevieDispatcher("CoxeterWords",arg);end;

EltWord:=
  function(arg)return ChevieDispatcher("EltWord",arg);end;

FakeDegrees:=function(arg)
  return ChevieDispatcher("FakeDegrees",arg);end;

FirstLeftDescending:=
  function(arg)return ChevieDispatcher("FirstLeftDescending",arg);end;

GenericOrder:=
  function(arg)return ChevieDispatcher("GenericOrder",arg);end;

Hecke:=
  function(arg)return ChevieDispatcher("Hecke",arg);end;

HeckeCharValues:=
  function(arg)return ChevieDispatcher("HeckeCharValues",arg);end;

HeckeSubAlgebra:=
  function(arg)return ChevieDispatcher("HeckeSubAlgebra",arg);end;

HighestPowerGenericDegrees:=
  function(arg)return ChevieDispatcher("HighestPowerGenericDegrees",arg);end;

Invariants:=
  function(arg)return ChevieDispatcher("Invariants",arg);end;

IsLeftDescending:=
  function(arg)return ChevieDispatcher("IsLeftDescending",arg);end;

LeftDescentSet:=
  function(arg)return ChevieDispatcher("LeftDescentSet",arg);end;

LowestPowerGenericDegrees:=
  function(arg)return ChevieDispatcher("LowestPowerGenericDegrees",arg);end;

MatXPerm:=
  function(arg)return ChevieDispatcher("MatXPerm",arg);end;

NrConjugacyClasses:=
  function(arg)return ChevieDispatcher("NrConjugacyClasses",arg);end;

PermMatX:=
  function(arg)return ChevieDispatcher("PermMatX",arg);end;

ReducedInRightCoset:=
  function(arg)return ChevieDispatcher("ReducedInRightCoset",arg);end;

ReflectionCharValue:=
  function(arg)return ChevieDispatcher("ReflectionCharValue",arg);end;

ReflectionSubgroup:=
  function(arg)return ChevieDispatcher("ReflectionSubgroup",arg);end;

Reflections:=
  function(arg)return ChevieDispatcher("Reflections",arg);end;

ReflectionEigenvalues:=
  function(arg)return ChevieDispatcher("ReflectionEigenvalues",arg);end;

RelativeGroup:=
  function(arg)return ChevieDispatcher("RelativeGroup",arg);end;

RelativeCoset:=
  function(arg)return ChevieDispatcher("RelativeCoset",arg);end;

Representations:=
  function(arg)return ChevieDispatcher("Representations",arg);end;

SchurElement:=
  function(arg)return ChevieDispatcher("SchurElement",arg);end;

TorusOrder:=
  function(arg)return ChevieDispatcher("TorusOrder",arg);end;

Valuation:=
  function(arg)return ChevieDispatcher("Valuation",arg);end;

ChevieAttributeDispatcher:=function(name,fieldname,a)local obj;
  obj:=a[1];
  if not IsRec(obj) then Error(obj," has no method for ",name);fi;
  if not IsBound(obj.(fieldname)) then
    obj.(fieldname):=ChevieDispatcher(name,a);
  fi;
  return obj.(fieldname);
end;

ChevieCharInfo:=function(arg)
  return ChevieAttributeDispatcher("ChevieCharInfo","charInfo",arg);end;

ChevieClassInfo:=function(arg)
  return ChevieAttributeDispatcher("ChevieClassInfo","classInfo",arg);end;

ReflectionDegrees:=function(arg)
  return ChevieAttributeDispatcher("ReflectionDegrees","degrees",arg);end;

ReflectionCoDegrees:=function(arg)
  return ChevieAttributeDispatcher("ReflectionCoDegrees","ReflectionCoDegrees",arg);end;

SchurElements:=function(arg)
  return ChevieAttributeDispatcher("SchurElements","schurElements",arg);end;

ReflectionName:=function(arg)local res;
   if Length(arg)=1 then Add(arg,rec());fi;
   if IsList(arg[1]) then 
     res:=List(arg[1],x->ReflectionName(x,arg[2]));
     if IsBound(arg[2].TeX) then return Join(res,"\\times ");
     else return Join(res,"x");
     fi;
   fi;
   return ChevieDispatcher("ReflectionName",arg);
end;

PositionSgn:=function(W)local ci;
  ci:=ChevieCharInfo(W);
  if not IsBound(ci.positionSgn) then return false;
  else return ci.positionSgn;
  fi;
end;

LowestPowerFakeDegrees:=function(W)local ci;
  ci:=ChevieCharInfo(W);
  if not IsBound(ci.b) then ci.b:=PermRootOps.LowestPowerFakeDegrees(W);fi;
  return ci.b;
end;

HighestPowerFakeDegrees:=function(W)local ci;
  ci:=ChevieCharInfo(W);
  if not IsBound(ci.B) then ci.B:=PermRootOps.HighestPowerFakeDegrees(W);fi;
  return ci.B;
end;

CharParams:=function(W)local ci;
  if IsRec(W) and IsBound(W.operations) and IsBound(W.operations.ChevieCharInfo)
  then ci:=ChevieCharInfo(W);
    if not IsBound(ci.charparams) then ci.charparams:=CharParams(CharTable(W));
    fi;
    return ci.charparams;
  elif IsCharTable(W) then
    if IsBound(W.irredinfo) and IsBound(W.irredinfo[1].charparam) then
      W.charparam:=List(W.irredinfo,x->x.charparam);
    else W.charparam:=[1..Length(W.irreducibles)];
    fi;
    return W.charparam;
  elif IsGroup(W) then return CharParams(CharTable(W));
  else Error(W, " has no method for CharParams");
  fi;
end;

CharNames:=function(arg)local W,option,res;
  W:=arg[1];
  if Length(arg)=2 then option:=arg[2];
  else option:=rec();
  fi;
  if IsRec(W) and IsBound(W.operations) and IsBound(W.operations.CharNames) then
   if not IsBound(option.TeX) and IsBound(W.charNames) then return W.charNames;fi;
   res:=W.operations.CharNames(W,option);
   if not IsBound(option.TeX) then W.charNames:=res;fi;
   return res;
  elif IsGroup(W) then return CharNames(CharTable(W));
  else Error(W," has no method for CharNames");
  fi;
end;

AlmostCharNames:=function(arg)local W,option,res;
  W:=arg[1];
  if Length(arg)=2 then option:=arg[2];
  else option:=rec();
  fi;
  if IsRec(W) and IsBound(W.operations) and IsBound(W.operations.AlmostCharNames) then
   if not IsBound(option.TeX) and IsBound(W.almostCharNames) then return W.almostCharNames;fi;
   res:=W.operations.AlmostCharNames(W,option);
   if not IsBound(option.TeX) then W.almostCharNames:=res;fi;
   return res;
  else Error(W," has no method for AlmostCharNames");
  fi;
end;

PositionId:=function(W)local ci;
  if IsCharTable(W) then
    if not IsBound(W.positionId) then
      W.positionId:=Position(W.irreducibles,W.irreducibles[1]*0+1);
    fi;
    return W.positionId;
  elif IsGroup(W) then ci:=ChevieCharInfo(W);
    if not IsBound(ci.positionId) then 
      ci.positionId:=PositionId(CharTable(W));
    fi;
    return ci.positionId;
  else Error(W, " has no method for PositionId");
  fi;
end;

#############################################################################
##
#F  PrintDiagram(<type>) . . . . . . . .Prints diagram for a ReflectionType
#F  PrintDiagram(<W>) . Prints diagram of Coxeter Datum or Hecke algebra, etc...
##  
##  The first form allows  PrintDiagram(ReflectionType(<cartan matrix>));
##

PrintDiagram:=function(W)local t;
  if IsRec(W) then ChevieDispatcher("PrintDiagram",[W]);return;fi;
  for t in W do PrintDiagram(t);od;
end;

HeckeClassPolynomials:=function(arg)local H;
  if IsBound(arg[1].coset) then H:=arg[1].coset;
  else H:=arg[1].hecke;
  fi;
  return ApplyFunc(H.operations.HeckeClassPolynomials,arg);
end;

Specialization:=function(H1,H2,f)
  if Group(H1)<>Group(H2) then 
    Error("Hecke algebras should be algebras for the same reflection group");fi;
  if ForAny([1..Length(H1.parameter)],i->ForAny([1,2],j->
      f(H1.parameter[i][j])<>H2.parameter[i][j])) then
    Error("f should send parameters of H1 to those of H2");fi;
  return t->t.operations.Specialization(t,H2,f);
end;
