#############################################################################
##
#A  hastype.g  CHEVIE library    Jean Michel  5-1999
##
#Y  Copyright (C) 1999 Lehrstuhl D fur Mathematik, RWTH Aachen,
#Y   and   University Paris VII.
##
##  This  file contains  operations  for reflexion  groups or  reflexion
##  cosets  which are  classified  (whose  decomposition in  irreducible
##  components is known). It builds answers by combining the answers for
##  irreducible types fetched from the CHEVIE tables. Here 'irreducible'
##  includes descents of  scalars. The code for irreducible  types is in
##  refltype.g .
##
##  A group or coset is classified  when its field .type has been filled
##  This field is accessed by the function 'ReflectionType' .

# ReflectionGroup from a ReflectionType
ReflectionGroup:=function(arg)local t,g,res,o,i,r;
  if Length(arg)=0 then return CoxeterGroup();fi;
  for t in arg do
    if t.series<>"ST" then g:=CoxeterGroup(t);
    elif IsBound(t.ST) then g:=ComplexReflectionGroup(t.ST);
    else g:=ComplexReflectionGroup(t.p,t.q,t.rank);
    fi;
    if t.series="ST" and IsBound(t.cartanType)then
      g.roots:=ShallowCopy(g.roots);
      g.simpleCoroots:=ShallowCopy(g.simpleCoroots);
      o:=Set(g.orbitRepresentative);
      if Length(o)=2 then
        o:=Filtered([1..Length(g.roots)],i->g.orbitRepresentative[i]=o[2]);
	g.roots{o}:=g.roots{o}*t.cartanType;
	o:=Intersection(o,[1..Length(g.simpleCoroots)]);
	g.simpleCoroots{o}:=g.simpleCoroots{o}/t.cartanType;
      else for i in [2..Length(o)] do
        r:=Filtered([1..Length(g.roots)],j->g.orbitRepresentative[j]=o[i]);
	g.roots{r}:=g.roots{r}*t.cartanType[i-1];od;
	r:=Intersection(r,[1..Length(g.simpleCoroots)]);
	g.simpleCoroots{r}:=g.simpleCoroots{r}/t.cartanType[i-1];
      fi;
      g.matgens:=List(g.generatingReflections,
                  j->Reflection(g.roots[j],g.simpleCoroots[j]));
    fi;
    if IsBound(res) then res:=res*g;
    else res:=g;
    fi;
  od;
  return res;
end;

HasTypeOps:=OperationsRecord("HasTypeOps");

###########################################################################
##
#F  CharName(W,p[,option]) . . . . . . . . Name of character with .charparam p
#    option may be "TeX"
##
HasTypeOps.CharName:=function(W,p,option)local t;
  t:=ReflectionType(W);
  return Join(List([1..Length(t)],i->CharName(t[i],p[i],option)));
end;

HasTypeOps.CharNames:=function(W,option)local ci;
  ci:=ChevieCharInfo(W);
  if IsBound(option.TeX) then return List(ci.charparams,x->CharName(W,x,option));
  else return ci.charnames;
  fi;
end;

HasTypeOps.NrConjugacyClasses:=W->Product(List(ReflectionType(W),
  NrConjugacyClasses));

#############################################################################
##
#F  HasTypeOps.ConjugacyClasses( <W> ) . . . . . . . ConjugacyClasses,
#F  using the classification
##  
HasTypeOps.ConjugacyClasses:=function(W)local t;
  t:=ChevieClassInfo(W);
  if IsSpets(W) then
    if t=false then return SpetsOps.ConjugacyClasses(W);
    else return List(t.classtext,x->ConjugacyClass(Group(W),
                                            EltWord(Group(W),x)*W.phi));
    fi;
  else
    if t=false then return PermGroupOps.ConjugacyClasses(W);
    else return List(t.classtext,w->ConjugacyClass(W,EltWord(W,w)));
    fi;
  fi;
end;

#############################################################################
##
#F  HasTypeOps.ChevieClassInfo( <W> ) . . . see corresponding dispatcher 
#F  function
##
HasTypeOps.ChevieClassInfo:=function(W) local tmp,res,inc;
  if IsSpets(W) then inc:=Group(W).rootInclusion;
                else inc:=W.rootInclusion;
  fi;
  tmp:=List(ReflectionType(W),function(t)local r,i;
    i:=ShallowCopy(ChevieClassInfo(t));
    if IsBound(t.orbit) then t:=t.orbit[1];fi;
    r:=MappingPermListList([1..Length(t.indices)],inc{t.indices});
    if i=false then return false;fi;
    i.classtext:=List(i.classtext,x->OnTuples(x,r));
    return i;end);
  
  if ForAny(tmp,x->x=false) then return false;fi;
  if Length(tmp)=1 then res:=tmp[1];# keep extra fields for irreducible type
  else res:=rec();
  fi; 
  res.classtext:=List(Cartesian(List(tmp,x->x.classtext)),Concatenation);
  res.classnames:=List(Cartesian(List(tmp,x->x.classnames)),Join);
  if ForAll(tmp,x->IsBound(x.classparams)) then 
     res.classparams:=Cartesian(List(tmp,x->x.classparams));
  fi;
  return res;
end;

#############################################################################
##
#F  WordsClassRepresentatives( <W> ) . . . . representatives of conjugacy 
#F  classes of <W> as Words. Classification is used.
##  
WordsClassRepresentatives:=W->ChevieClassInfo(W).classtext;

#############################################################################
##
#F  HasTypeOps.BraidRelations( <W> ) . . . see corresponding dispatcher 
#F  function
##
HasTypeOps.BraidRelations:=W->Concatenation(List(ReflectionType(W),
  t->List(BraidRelations(t),x->List(x,y->W.rootInclusion{y}))));
  
#############################################################################
##
#F  HasTypeOps.CharTable( <W> ) . . . . . . . . . . . 'CharTable', using
#F  the classification
##
HasTypeOps.CharTable:=function(W)local t, l, res, cl, d;

  l:=List(ReflectionType(W),ReflTypeOps.CharTable);
  if ForAny(l,tbl->tbl=false) then 
    if IsSpets(W) then res:=SpetsOps.CharTable(W);
    else res:=PermGroupOps.CharTable(W);
    fi;
  else
    if Length(l)>0 then res:=l[1];
    else 
      res:=Group(());res.name:="";
      res:=GroupOps.CharTable(res);
      Unbind(res.group);
      res.classtext:=[[]];
    fi;

    for t in l{[2..Length(l)]} do
      res:=CharTableDirectProduct(res,t); 
      Unbind(res.fusionsource); Unbind(res.fusions);
    od;
  fi;
  
  res.irredinfo:=List(ChevieCharInfo(W).charparams,
                      x->rec(charparam:=x,charname:=CharName(W,x)));

  # for a coset, InitClassesCharTable which is called from
  # CharTableDirectProduct uses a wrong value of size (centralizers[1])
  # so we must fix the classes
  res.classes:=List(res.centralizers,x->res.size/x);

  t:=ChevieClassInfo(W);
  if t<>false then Inherit(res,t);fi;

  if not IsSpets(W) and not IsBound(res.powermap) then
    Print("warning: powermap not bound\n");
    res.powermap:=[];
    cl:=List(ConjugacyClasses(W),Representative);
    for d in Reversed(Set(Factors(res.size))) do
    res.powermap[d]:=List(cl,x->PositionClass(W,x^d));
    od;
  fi;

  res.operations.StringEntry := function(x)
    if x=0*x then return ".";else return Format(x);fi;end;

  if IsBound(W.name) then res.name:=W.name;
  else res.name:=ReflectionName(W);
  fi;
  res.identifier:=res.name;
  return res;
end;

#############################################################################
##
#F  HasTypeOps.Representations( <W> ) . .  'Representations', using
#F  the classification
##
HasTypeOps.Representations:=function(arg)local W,l,t,inds,rep;
  W:=arg[1]; t:=ReflectionType(W); l:=List(t,NrConjugacyClasses);
  if Length(arg)=2 then inds:=arg[2];else inds:=[1..Product(l)];fi;
  rep:=function(i)local reps,tensorat,dims,ind,res;
    if Length(t)=1 then return ReflTypeOps.Representation(t[1],i);fi;
    ind:=CartesianPosition(i,l);if ind=false then return false;fi;
    reps:=List([1..Length(t)],i->ReflTypeOps.Representation(t[i],ind[i]));
    if ForAny(reps,r->r=false) then return false;fi;
    tensorat:=function(i,m,dims) # id(d1)tensor id(d2)..tensor(m) at i tensor..
      return List([1..Product(dims)],k->List([1..Product(dims)],
      function(l)local r,c;
	r:=CartesianPosition(k,dims); c:=CartesianPosition(l,dims);
	if ForAny([1..Length(dims)],j->r[j]<>c[j] and j<>i) then return 0;
	else return m[r[i]][c[i]];fi;end));
    end;
    if IsSpets(W) then
      dims:=List(reps,x->Length(x.F[1]));
      res:=rec(gens:=[],
	F:=List([1..Product(dims)],k->List([1..Product(dims)],
          function(l)local r,c;
	    r:=CartesianPosition(k,dims); c:=CartesianPosition(l,dims);
	   return Product([1..Length(reps)],i->reps[i].F[r[i]][c[i]]);end)));
      for i in [1..Length(t)] do
        res.gens{t[i].indices}:=List(reps[i].gens,m->tensorat(i,m,dims));
      od;
    else
      dims:=List(reps,x->Length(x[1]));res:=[];
      for i in [1..Length(t)] do
        res{t[i].indices}:=List(reps[i],m->tensorat(i,m,dims));
      od;
      return res;
    fi;
  end;
  if IsList(inds) then return List(inds,rep);
  else return rep(inds);
  fi;
end;

############################################################################
##
#F  FakeDegree( <W>, <p>, <q> ) . . .Fake Degree of char with charparam <p>
##  
##  This   returns the   polynomial  describing the   multiplicity of  the
##  character chi with .charparam  p in the graded  version of the regular
##  representation  given  by the quotient  S/I  where S  is the symmetric
##  algebra of the reflection representation and  I is the ideal generated
##  by the homogenous invariants of positive degree in S.
##  
FakeDegree:=function(W,p,q)local t;
  t:=ReflectionType(W);
  t:=List([1..Length(t)],i->ReflTypeOps.FakeDegree(t[i],p[i],q));
  if ForAny(t,x->x=false) then return false;
  else return Product(t)*q^0;
  fi;
end;

############################################################################
##
#F  FakeDegrees( <W>, <q> ) . . . . . .  Fake Degrees of group <W>
##
##  This returns the list of polynomials describing the multiplicity of
##  each character in the graded version of the regular representation given
##  by the quotient S/I where S is the symmetric algebra of the reflection
##  representation and I is the ideal generated by the homogenous invariants
##  of positive degree in S.
##  The ordering of the result corresponds to the ordering of the characters 
##  in the  CharTable.
##
#
HasTypeOps.FakeDegrees:=function(W,q)local P,p,f;
  P:=[];
  for p in CharParams(W) do
    f:=FakeDegree(W,p,q);
    if f=false then 
      if IsSpets(W) then return SpetsOps.FakeDegrees(W,q);
      else return PermRootOps.FakeDegrees(W,q);
      fi;
    fi;
    Add(P,f);
  od;
  return P;
end;

#############################################################################
##
#F  ReflectionDegrees( <W>)  . . . . . . . degrees as a reflection group
##
##  Returns the degrees of the  reflection group W.
##  
HasTypeOps.ReflectionDegrees:=W->Concatenation([W.semisimpleRank+1..W.rank]*0+1,
    Concatenation(List(ReflectionType(W),ReflectionDegrees)));

HasTypeOps.ReflectionCoDegrees:=W->Concatenation([W.semisimpleRank+1..W.rank]*0-1,
   Concatenation(List(ReflectionType(W),ReflectionCoDegrees)));


#############################################################################
##
#F  HasTypeOps.PrintDiagram( <W> )  . see dispatcher function
##  
#############################################################################
##
#F  SpetsOps.PrintDiagram( <WF> ) . . . . prints twisted Dynkin
#F  diagram of <WF>
##  
##  In  addition to 'PrintDiagram(  Group(<WF>) )'  information 
##  about the action of <WF>.phi on the Dynkin diagram is printed.
##  
HasTypeOps.PrintDiagram:=function(W)
  PrintDiagram(List(ReflectionType(W),function(t)
    t:=ShallowCopy(t);
    if IsSpets(W) then
      t.orbit:=List(t.orbit,function(a)local H;
	a:=ShallowCopy(a); H:=Group(W);
	a.indices:=H.reflectionsLabels{a.indices};
	return a;end);
      if t.twist<>() then
	t.twist:=RestrictedPerm(W.phi^Length(t.orbit),t.orbit[1].indices);
      fi;
    else
      t.indices:=W.reflectionsLabels{t.indices};
    fi;
    return t;end));
end;

#############################################################################
##
#F  HasTypeOps.ReflectionName(<W>)  see dispatcher function
##  Gives a string which describes the isomorphism type of W.  
##  
##  For a Spets, An orbit of
##  phi on the  components is put in brackets  if of length x greater than
##  1, and  is  preceded by  the  order of  phi^x on it,  e.g. 2(A2xA2xA2)
##  denotes 3 components A2  permuted by phi,  and such that phi^3 induces
##  the non-trivial diagram automorphism on any of them, while 3D4 denotes
##  an orbit of  length 1 on  which phi is of order  3.  
##    If the coset is not semi-simple, its toral part is added as a
##  product of cyclotomic factors over the minimum possible field.
##  
HasTypeOps.ReflectionName:=function(W,option)local res,t,i,total,l;
  t:=ReflectionType(W); res:="";total:=0;
  for i in [1..Length(t)] do
    Append(res,ReflectionName(t[i],option));
    if IsSpets(W) then
      l:=Group(W).rootInclusion{Concatenation(List(t[i].orbit,T->T.indices))};
    else l:=W.rootInclusion{t[i].indices};
    fi;
    if l<>total+[1..Length(l)] then
 # added JM 3 oct. 2000 so that different reflection subgroups have
 # different reflection names: this is necessary in 'storefusion'
      if IsBound(option.TeX) then 
           PrintToString(res,"\\langle ",Join(l),"\\rangle ");
      else PrintToString(res,"<",Join(l),">");fi;
    fi;
    if i<Length(t) then 
      if IsBound(option.TeX) then Append(res,"\\times ");
      else Append(res,"x");fi;
    fi;
    total:=total+Length(l);
  od;
  if IsSpets(W) then 
    t:=Group(W).rank-Group(W).semisimpleRank ;
    if t>0 then
      i:=Copy(option);i.GAP:=true;
      if res<>"" then Append(res,".");fi;
      Append(res,Format(CycPol(Concatenation([1,0],List(ListBlist(PhiFactors(W),
        List(ReflectionDegrees(Group(W)),x->x=1)),AsRootOfUnity))),i));
    fi;
  else
    t:=W.rank-W.semisimpleRank ;
    if t>0 then
      if res<>"" then Append(res,".");fi;
      Append(res,"(q-1)"); 
      if t>1 then 
        if IsBound(option.TeX) then PrintToString(res,"^{",t,"}");
        else PrintToString(res,"^",t);fi;
      fi;
    fi;
  fi;
  return String(res);
end;

HasTypeOps.ChevieCharInfo:=function(W)local res,t,p,f;
  t:=ReflectionType(W);
  p:=List(t,ChevieCharInfo);
  if Length(t)=1 then res:=ShallowCopy(p[1]);# keep additional fields for irreducible type
  else res:=rec(positionRefl:=false);
  fi; 
  for f in ["positionId","positionSgn"] do
    if ForAll(p,x->IsBound(x.(f))) then 
      res.(f):=CartesianPosition(List(p,x->x.(f)),
	   List(t,x->Length(CharParams(x))));
    fi;
  od;
  if ForAll(p,x->IsBound(x.charparams)) then 
    res.charparams:=Cartesian(List(p,x->x.charparams));
  fi;
  for f in ["b","B","a","A"] do
    if ForAll(p,x->IsBound(x.(f))) then 
      res.(f):=List(Cartesian(List(p,x->x.(f))),Sum);
    fi;
  od;
  res.charnames:=List(res.charparams,i->HasTypeOps.CharName(W,i,rec()));
  return res;
end;

HasTypeOps.LowestPowerGenericDegrees:=function(W)local ci;
  ci:=ChevieCharInfo(W);
  if not IsBound(ci.a) then Error("no LowestPowerGenericDegrees for ",W);fi;
  return ci.a;
end;

HasTypeOps.HighestPowerGenericDegrees:=function(W)local ci;
  ci:=ChevieCharInfo(W);
  if not IsBound(ci.a) then Error("no HighestPowerGenericDegrees for ",W);fi;
  return ci.a;
end;

HasTypeOps.Invariants:=function(W)local V;
  V:=Parent(W);
  return Concatenation(List(ReflectionType(W),function(t)local H,ir;
    H:=ReflectionGroup(t);
    if CartanMat(W,t.indices)<>CartanMat(H) then
      Error("not standard cartan matrix: cannot compute invariants");
    fi;
    ir:=PermRootOps.IndependentRoots(H);
    return List(Invariants(t),f->function(arg)
#     Print("t=",t," H=",H,"\n");
      return ApplyFunc(f,H.simpleCoroots{ir}^-1*List(ir,i->
       PermRootOps.Coroot(V,W.rootInclusion[t.indices[i]]))*arg);end);
      end));
end;
