#############################################################################
##
#A  hekcoset.g                     CHEVIE library                 Jean Michel
##
#A  $Id: hekcoset.g,v 1.1 1997/01/21 13:46:31 gap Exp $
##
#Y  Copyright (C) 1996  Lehrstuhl D f\"ur Mathematik, RWTH Aachen, and
#Y  Universite Paris VII.
##
##  This file contains  GAP functions for working with Hecke cosets and
##  their character tables.
##

if not IsBound(HeckeAlgebraOps.InitKL) then ReadChv("prg/kl");fi;

##############################################################################
HeckeCosetOps:=OperationsRecord("HeckeCosetOps",DomainOps);
HeckeCosetOps.T:=HeckeAlgebraOps.T;

HeckeCosetOps.Hecke:=HF->HF.hecke;
HeckeCosetOps.Group:=HF->Group(HF.hecke);

HeckeCosetOps.String:=function(HF)local H,W,i,o,para,root,res;
  H:=Hecke(HF);W:=Group(H);
  o:=List(W.EigenvaluesGeneratingReflections,x->1/x);
  para:=ShallowCopy(H.parameter);
  if Length(para)>0 and ForAll(para,x->x=para[1]) then para:=[para[1]];fi;
  for i in [1..Length(para)] do
    if IsList(para[i]) and para[i]{[2..Length(para[i])]}=
               List([1..o[i]-1],j->E(o[i])^j) then para[i]:=para[i][1];fi;
  od;
  if Length(para)=1 and not IsList(para[1]) then para:=para[1];fi;
  if ForAny([1..Length(o)],i->IsBound(H.rootParameter[i])) then
    root:=ShallowCopy(H.rootParameter);
    if Length(root)>0 and ForAll(root,x->x=root[1]) then root:=root[1];fi;
  fi;
  res:=SPrint("Hecke(",CoxeterCoset(HF));
  if para<>1 or (IsBound(root) and root<>1) then PrintToString(res,",",para);fi;
  if IsBound(root) and root<>1 then PrintToString(res,",",root);fi;
  PrintToString(res,")");
  return res;
end;

HeckeCosetOps.Print:=function(H)Print(String(H));end;

HeckeCosetOps.CharTable:=function(HF)
  local t,res,r,tbl,o,WF,H,tmp,param,sqrtparam;
  tmp:=[];
  WF:=CoxeterCoset(HF);
  H:=Hecke(HF);
  for o in ReflectionType(WF) do
    t:=o.orbit[1];
    param:=H.parameter{t.indices};
    sqrtparam:=SublistUnbnd(H.rootParameter,t.indices);
    if OrderPerm(o.twist)=1 
    then tbl:=ReflTypeOps.HeckeCharTable(t,param,sqrtparam);
    else tbl:=ReflTypeOps.HeckeCharTable(o,param,sqrtparam);
    fi;
    if Length(o.orbit)>1 then
# descent of scalars: essentially the table is that of 
# the group corresponding to o.orbit[1] with o.twist acting on it.
     tbl.name:="";tbl.identifier:=tbl.name;
     tbl.size:=tbl.size^Length(o.orbit);
     tbl.classes:=List(tbl.centralizers,x->tbl.size/x);
    fi;
    Add(tmp,tbl);
  od;
  res:=tmp[1];
  for tbl in tmp{[2..Length(tmp)]} do
    res:=CharTableDirectProduct(res,tbl);
  od;
  res.classes:=List(res.centralizers,x->res.size/x);
  res.irredinfo:=List(CharParams(WF),x->rec(charparam:=x,
                                            charname:=CharName(WF,x)));
  Inherit(res,ChevieClassInfo(WF));
  res.name:=Concatenation("H(",ReflectionName(WF),")");
  res.identifier:=res.name;
  return res;
end;

# make a hecke coset from Coxeter coset WF and Hecke algebra H or parameters
CoxeterCosetOpsHecke:=function(arg)local W,i,WF,H,res;
  WF:=arg[1];W:=Group(WF);
  if Length(arg)=2 and IsRec(arg[2]) and IsBound(arg[2].reflectionGroup) then
    H:=arg[2];
  else
    arg[1]:=W;H:=ApplyFunc(Hecke,arg);
  fi;
  if W.semisimpleRank>0 then
  for i in Cycles(WF.phi,W.rootInclusion{[1..W.semisimpleRank]}) do
    if Length(Set(H.parameter{W.rootRestriction{i}}))>1 then
      Error("Hecke algebra parameters should be equal for", 
            " phi-conjugate reflections");
    fi;
  od;
  fi;
  res:=rec(hecke:=H,coxeterCoset:=WF,isGroup:=true,operations:=HeckeCosetOps);

  return CHEVIE.GetCached(WF,"HeckeCosets",res,function(x,res)return
    x.hecke.parameter=res.hecke.parameter and 
       x.hecke.rootParameter=res.hecke.rootParameter;end);
end;

HeckeCosetOps.Basis:=function(H,basis)
  if not IsBound(H.operations.(basis)) then 
      Error("basis ",basis," unknown");fi;
  return function(arg)local res;
   res:=Hecke(H).operations.(basis).MakeBasisElt(Hecke(H),basis,arg);
   res.coset:=H;
   return res;
   end;
end;

HeckeCosetOps.Group:=H->H.coxeterCoset;

HeckeCosetOps.HeckeSubAlgebra:=function(HF,subW)local WF,W;
  WF:=CoxeterCoset(HF);
  if IsList(subW) then subW:=CoxeterSubCoset(WF,subW);fi;
  return Hecke(subW,HeckeSubAlgebra(Hecke(HF),Group(subW)));
end;

#############################################################################
##
#F  HeckeClassPolynomials(<h>)  . . . . . . the class polynomials
##
##  The program below conjecturally returns the class polynomials of the
##  element <h> of the Hecke coset <H> with  respect  to  representatives
##  of minimal length in  the F-conjugacy classes of the Coxeter coset of
##  <H>. <h>  is  an element of <H> given in any of the standard bases.
##  The representatives taken are 'WordsClassRepresentatives(W)'

HeckeCosetOps.HeckeClassPolynomials:=function(arg)
  local H,HF,reps,h,T,q,res,min,l,max,inds,i,o,orb,W,WF,F;
  HF:=arg[1].coset;
  H:=Hecke(HF);
  T:=Basis(H,"T");h:=T(arg[1]);
  WF:=HF.coxeterCoset;
  W:=Group(WF);
  F:=Frobenius(WF);
  reps:=WordsClassRepresentatives(WF);

# if there exists w conjugate by cyclic shift to elm and an elementary
# reflection s such that l(swF(s))=l(elm)-2 then orb returns
# q*T_swF(s)+(q-1)T_sw  else an error is signaled.

  orb:=function(elm)local orbit,w,y,s,yy,ind,ILD;
    ILD:=W.operations.IsLeftDescending;
    orbit:=[elm];
    for w in orbit do
      for s in [1..W.semisimpleRank] do
        q:=H.parameter[s];
	y:=W.generators[s]*w;yy:=y*F(W.generators[s]);
	if ILD(W,w,s) then
	  if ILD(W,y^-1,s^WF.phi) then return T([yy,y],[-q[1]*q[2],q[1]+q[2]]);
	  elif not yy in orbit then Add(orbit,yy);
	  fi;
	elif ILD(W,y^-1,s^WF.phi) and not yy in orbit then Add(orbit,yy);
	fi; 
      od;
    od;
    Error("Geck-Kim-Pfeiffer theory for",CoxeterWord(W,elm));
  end;

  min:=[1..Length(reps)]*0*H.parameter[1][1]^0;
  while Length(h.elm)>0 do
   l:=List(h.elm,x->CoxeterLength(W,x));
   max:=Maximum(l);
   inds:=Filtered([1..Length(h.elm)],x->max=l[x]);
   max:=T(h.elm{inds},h.coeff{inds});
   h:=h-max;
   for i in [1..Length(max.elm)] do
     o:=PositionClass(WF,max.elm[i]*WF.phi);
     if Length(reps[o])=CoxeterLength(W,max.elm[i]) then 
       min[o]:=min[o]+max.coeff[i];
     else
       o:=max.coeff[i]*orb(max.elm[i]);
       Append(h.elm,o.elm);Append(h.coeff,o.coeff);
     fi;
   od;
   Normalize(h);
  od;
  return min;
end;
