#############################################################################
##
#A  refltype.g  CHEVIE library    Jean Michel  2-2000
##
#Y  Copyright (C) 1999 Lehrstuhl D f\"ur Mathematik, RWTH Aachen,
#Y  and University Paris VII.
##
##  This file contains operations for reflexion groups which are classified
##  and irreducible (for a Spets, irreducible means only one orbit under the
##  twist -- this includes descent of scalars).
##  It fetches the answers from the CHEVIE tables.
##  
##  A group is classified when its field .type has been filled.
##  (this attribute is obtained by the function 'ReflectionType').
##  This field is a list of records describing each irreducible component
##  (or an orbit of such components under the twist).
##  The argument of the ReflTypeOps routines is one such record.

ReflTypeOps:=OperationsRecord("ReflTypeOps");

ReflTypeOps.DegreeDescent:=function(t)
  if IsBound(t.orbit) then return Length(t.orbit); else return 1; fi;
end;

#ReflTypeOps.CharParams:=t->CHEVIE.Data("CharParams",t);
ReflTypeOps.ReflectionDegrees:=t->CHEVIE.Data("ReflectionDegrees",t);

ReflTypeOps.CartanMat:=function(t)
  if IsBound(t.cartan) then return t.cartan;
  elif IsBound(t.cartanType) then 
       return CHEVIE.Data("CartanMat",t,t.cartanType);
  else return CHEVIE.Data("CartanMat",t);
  fi;
end;

ReflTypeOps.ReflectionCoDegrees:=t->CHEVIE.Data("ReflectionCoDegrees",t);
ReflTypeOps.GeneratingRoots:=t->ShallowCopy(CHEVIE.Data("GeneratingRoots",t));
ReflTypeOps.GeneratingCoRoots:=t->ShallowCopy(CHEVIE.Data("GeneratingCoRoots",t));
ReflTypeOps.Invariants:=t->CHEVIE.Data("Invariants",t);
ReflTypeOps.ParabolicInfo:=t->CHEVIE.Data("ParabolicInfo",t);

ReflTypeOps.PoincarePolynomial:=function(t,param)
  return CHEVIE.Data("PoincarePolynomial",t,param);
end;

ReflTypeOps.HeckeCharTable:=function(arg)local tbl;
  tbl:=ApplyFunc(CHEVIE.Data,Concatenation(["HeckeCharTable"],arg));
  if tbl=false then return tbl;fi;
  if not IsBound(tbl.name) then tbl.name:=tbl.identifier; fi;
  if not IsBound(tbl.order) then tbl.order:=tbl.size; fi;
  return tbl;
end;

ReflTypeOps.HeckeRepresentation:=function(arg)
  return ApplyFunc(CHEVIE.Data,Concatenation(["HeckeRepresentation"],arg));
end;

ReflTypeOps.CharName:=function(arg)
  return ApplyFunc(CHEVIE.Load("CharName",arg[1]).fct,arg{[2..Length(arg)]});
end;

ReflTypeOps.FakeDegree:=function(t,p,q)
  if IsBound(t.scalar) then 
       return CHEVIE.Data("FakeDegree",t,p,Product(t.scalar,s->q*s^-1));
  else return CHEVIE.Data("FakeDegree",t,p,q^ReflTypeOps.DegreeDescent(t));
  fi;
end;

ReflTypeOps.EigenvaluesGeneratingReflections:=function(t)
  if t.series<>"ST" then return List(CHEVIE.Data("CartanMat",t),x->1/2);
  else return CHEVIE.Data("EigenvaluesGeneratingReflections",t);
  fi;
end;

ReflTypeOps.ChevieClassInfo:=function(t)local r;
  r:=CHEVIE.Data("ClassInfo",t);
  if IsBound(t.orbit) and Length(t.orbit)>1 then
    if IsBound(r.classes) then r.classes:=r.classes*
       Product(ReflectionDegrees(t.orbit[1]))^(Length(t.orbit)-1);fi;
    # for nontrivial F the orders are  not easy to determine (and cannot
    # be determined  from the  ReflectionType for  nonsemisimple Spets),
    Unbind(r.orders);
  fi;
  return r;
end;

#############################################################################
##
#F  ReflTypeOps.CharTable( <t> ) . . . . . . . . . . . 'CharTable', using
#F  the classification
##
ReflTypeOps.CharTable:=function(t)local tbl;
  tbl:=CHEVIE.Data("CharTable",t);
  if tbl=false then return false;fi;
  if not IsBound(tbl.name) then tbl.name:=tbl.identifier; fi;
  if not IsBound(tbl.order) then tbl.order:=tbl.size; fi;
  if IsBound(t.orbit) then
    # for nontrivial F the orders are  not easy to determine (and cannot
    # be determined  from the  ReflectionType for  nonsemisimple Spets),
    # and powermaps do not make sense:
    if t.twist<>() then
      Unbind(tbl.orders);
      Unbind(tbl.powermap);
    fi;
    if Length(t.orbit)>1 then
      tbl.size:=tbl.size^Length(t.orbit);
      tbl.order:=tbl.size;
      tbl.classes:=List(tbl.centralizers,a->tbl.size/a);
    fi;
 #  if IsBound(t.scalar) then  # scalar-twisted Spetses
 #    # is the arbitrary choice below reasonable?
 #    tbl.irreducibles:=tbl.irreducibles*Product(t.scalar);
 #  fi;
  fi;
  return tbl;
end;

ReflTypeOps.Representation:=function(arg)
  return ApplyFunc(CHEVIE.Data,Concatenation(["Representation"],arg));
end;

ReflTypeOps.NrConjugacyClasses:=function(arg)
  return ApplyFunc(CHEVIE.Data,Concatenation(["NrConjugacyClasses"],arg));
end;

ReflTypeOps.PrintDiagram:=function(t)local s,t;
  if IsBound(t.orbit) then
     if Length(t.orbit)>1 then
       Print("phi permutes the next ",Length(t.orbit)," components\n");
     fi;
     if t.twist<>() then
       Print("phi");
       if Length(t.orbit)>1 then Print("^",Length(t.orbit));fi;
       Print(" acts as ",t.twist," on the component below\n");
     fi;
     PrintDiagram(t.orbit);
  else s:=ReflTypeOps.ReflectionName(t,rec());
    if IsBound(t.cartanType) then 
         CHEVIE.Data("PrintDiagram",t,t.indices,s,t.cartanType);
    else CHEVIE.Data("PrintDiagram",t,t.indices,s);
    fi;
  fi;
end;

ReflTypeOps.BraidRelations:=function(t)local m,p,r;
  if IsBound(t.orbit) then Error("not for Spets");
  elif t.series<>"ST" then
   m:=CoxeterMatrixFromCartanMat(CHEVIE.Data("CartanMat",t));
   p:=function(i,j,b)return List([1..b],k->i*(k mod 2)+j*((1-k)mod 2));end;
   r:=Concatenation(List([1..Length(m)],i->List([1..i-1],
     j->[p(i,j,m[i][j]),p(j,i,m[i][j])])));
  else r:=CHEVIE.Data("BraidRelations",t);
  fi;
  if IsBound(t.indices) then return List(r,x->List(x,y->t.indices{y}));
  else return r;
  fi;
end;

#############################################################################
##
#F  ReflectionName(<type>) . . . . . . . . . . . . name of a ReflectionType
##  
##  Gives a string which describes the isomorphism type of W (like "A2B3" for
##  the direct product of a root system of type A2 by one of type B3, or like
##  "I2(7)"). 
##

ReflTypeOps.ReflectionName:=function(t,option)local res;
  res:="";
  if IsBound(t.orbit) then  # Coxeter cosets
    if OrderPerm(t.twist)<>1 then Append(res,String(OrderPerm(t.twist)));
      if t.twist=(1,4,2) then Append(res,"'");fi;
    fi;
    if Length(t.orbit)=1 then Append(res,ReflTypeOps.ReflectionName(t.orbit[1],option));
    else PrintToString(res,"(",ReflectionName(t.orbit,option),")");
    fi;
    if IsBound(t.scalar) and ForAny(t.scalar,x->x<>1) then
      PrintToString(res,"[",Join(List(t.scalar,x->Format(x,option))),"]");
    fi;
  else 
    if IsBound(t.tilde) then 
      if IsBound(option.TeX) then Append(res,"\\tilde");
      else Append(res,"~");fi;
    fi;
    if IsBound(t.cartanType) then 
         Append(res,CHEVIE.Data("ReflectionName",t,t.cartanType,option));
    else Append(res,CHEVIE.Data("ReflectionName",t,option));
    fi;
  fi;
  return String(res);
end;

ReflTypeOps.String:=ReflTypeOps.ReflectionName;

# Let t.scalar=[s_1,..,s_r],  where r=Length(t.orbit)  and let p  be the
# PhiFactor of t.twist  associated to the i-th  invariant of t.orbit[1],
# and let d_i be the degree of that invariant.
# If G0 is the Spets described by t.orbit[1],
# And G1:=Ennola(Product(t.scalar),G0)
# Then G is isomorphic to the descent of scalars of G1.
# According to spets 1.5, the Phifactors of Ennola(zeta,G0)
# are \zeta^{d-i} times those of G0;
# And those of an a-descent of scalars are \zeta_a^j\zeta_i^{1/a}
ReflTypeOps.PhiFactors:=function(t)local f,d,a;
  if OrderPerm(t.twist)>1 then f:=CHEVIE.Data("PhiFactors",t);
    if f=false then return false;fi;
  else f:=[1..t.orbit[1].rank]*0+1;fi;
  if IsBound(t.scalar) then 
    d:=Product(t.scalar);
    d:=List(ReflectionDegrees(t.orbit[1]),deg->d^deg);
    f:=List([1..Length(f)],i->f[i]*d[i]);
  fi;
  a:=Length(t.orbit);
  f:=List(f,x->GetRoot(x,a));
  return Concatenation(List([0..a-1],i->f*E(a)^i));
end;

ReflTypeOps.ChevieCharInfo:=function(t)local res,f,get;
  get:=function(f,F)local tmp;
    if not IsBound(res.(f)) then 
      tmp:=CHEVIE.Data(F,t);if tmp<>false then res.(f):=tmp;fi;
    fi;
  end;
  res:=CHEVIE.Data("CharInfo",t);
  if res=false then 
    res:=rec(charparams:=CHEVIE.Data("CharParams",t));
    InfoChevie("# old style CharInfo for ",
       ReflTypeOps.ReflectionName(t,rec()),"\n");
  fi;
  if not IsBound(res.charnames) then
    f:=CHEVIE.Load("CharName",t).fct;
    res.charnames:=List(res.charparams,x->f(x,rec()));
  fi;
  get("positionRefl","PositionRefl");
  get("positionId","PositionId");
  if (IsBound(t.orbit) and t.orbit[1].series<>"ST") or 
     (not IsBound(t.orbit) and t.series<>"ST") then
    get("positionSgn","PositionSgn");
    get("b","LowestPowerFakeDegrees");
    get("B","HighestPowerFakeDegrees");
    get("a","LowestPowerGenericDegrees");
    get("A","HighestPowerGenericDegrees");
  fi;
  for f in ["a","A","b","B"] do
  if IsBound(res.(f)) then res.(f):=res.(f)*ReflTypeOps.DegreeDescent(t);fi;
  od;
  return res;
end;
