#############################################################################
##
#A  tbl/compat3.g                 CHEVIE library                 Frank Lbeck
##
#Y  Copyright (C) 2001  The CHEVIE Team
##
##  This file  contains some functions  for GAP 3  which allow to  use the
##  remaining data files to  be shared by the GAP 3 and  GAP 4 versions of
##  CHEVIE.
##
# allowing `Revision' information in shared data files
if not IsBound(CHEVIE.Revision) then
  CHEVIE.Revision := Ignore;
fi;

CHEVIE.Revision("tbl/compat3.g",
   "$Id: compat3.g,v 1.8 2003/12/04 11:26:45 chevie Exp $");

## just another name in GAP 4
CallFuncList := ApplyFunc;

#############################################################################
##
##  CHEVIE.compat  . . . . . . . . . . . . record to collect functions which
##  are different for GAP 3 and GAP 4
##
CHEVIE.compat := rec();

CHEVIE.compat.InfoChevie := function(arg)
  # looks circumstantially, but works after resetting InfoChevie
  ApplyFunc(InfoChevie, arg);
end;

##  new name in GAP 4
CHEVIE.compat.CharTable := CharTable;

##  very different for GAP 4
CHEVIE.compat.AdjustHeckeCharTable := HeckeAlgebraOps.AdjustCharTable;

##  need `!.' in GAP 4
CHEVIE.compat.ChangeIdentifier := function(obj, name)
  obj.identifier := name;
end;

##  since function is not in GAP 4
CHEVIE.compat.Inherit := Inherit;

##  trivial here
CHEVIE.compat.MakeCharacterTable := function(arg)
  local tbl;
  tbl := arg[1];
  if Length(arg) = 1 then
    tbl.operations := CharTableOps;
  else
    tbl.operations := ShallowCopy(CharTableOps);
    tbl.operations.StringEntry := function(x)
      if x=0*x then return ".";else return Format(x);fi;end;
  fi;
  return tbl;
end;

##  efficient version in GAP 4, but here
NrPartitionTuples := function(n, k)
  return Length(PartitionTuples(n, k));
end;

##  The remaining entries are for character tables in classical types. Their
##  creation is very incompatible between GAP3/4.
CHEVIE.compat.CharTableA := function(rank)
  local tbl,cl,f;
  tbl:=CharTable("Symmetric",rank+1);
  if not IsBound(tbl.size) then
    tbl.size:=Factorial(rank+1);
  fi;
  tbl.identifier:=String(Concatenation("W(A",String(rank),")"));
  tbl.cartan:=CartanMat("A",rank);
  cl:=CHEVIE.RawData("ClassInfo","A")(rank);
  for f in RecFields(cl) do tbl.(f):=cl.(f);od;
  tbl.irredinfo:=List(CHEVIE.RawData("CharParams","A")(rank),
    x->rec(charparam:=x, charname:=CHEVIE.RawData("CharName","A")(x,rec())));
  return tbl;
end;
CHEVIE.compat.HeckeCharTableA := function(n, param,sqrtparam)
  local tbl,cl,f,q;
  q:=-param[1][1]/param[1][2];
  tbl:= CharTableSpecialized(CHEVIE.RawData("Hk","A"), [n+1, q]);
  tbl.cartan:= CartanMat("A", n);
  tbl.parameter:= List([1..n], x-> q);
  cl:=CHEVIE.RawData("ClassInfo","A")(n);
  for f in RecFields(cl) do tbl.(f):=cl.(f);od;
  tbl.irredinfo:=List(CHEVIE.RawData("CharParams","A")(n),x->
             rec(charparam:=x,charname:=CHEVIE.RawData("CharName","A")(x)));
  HeckeAlgebraOps.AdjustCharTable(tbl,param);
  return tbl;
end;
##  case ^2A
CHEVIE.compat.CharTable2A := function(r)
  local i, tbl;
  tbl := CHEVIE.RawData("CharTable","A")(r);
  tbl.identifier := SPrint("W(^2A",r,")");
  for i in [1..Length(tbl.irreducibles)] do
    # [Lusztig's Book, 4.19 line 9]
    # Preferred extension: \sigma acts on \tilde E by (-1)^a_E.w_0
    tbl.irreducibles[i]:=(-1)^CHEVIE.RawData("Charb","A")
        (tbl.irredinfo[i].charparam)*tbl.irreducibles[i];
  od;
  Inherit(tbl,CHEVIE.RawData("ClassInfo","2A")(r));
  return tbl;
end;
CHEVIE.compat.HeckeCharTable2A := function(r,param,rootparam)
  local q, v, W, phi, WF, qE, H, tbl, tmp, cl, i;
  q:=-param[1][1]/param[1][2];
  if not IsBound(rootparam[1]) then v:=GetRoot(q,2,"CharTable(Hecke(2A))");
  else v:=rootparam[1];
  fi;
  W:=CoxeterGroup("A",r);
  if r=1 then phi:=();
  else phi:=Product([1..QuoInt(r,2)],i->(i,r+1-i));
  fi;
  WF:=CoxeterCoset(W,phi);
# If q_E is the square root which deforms to 1 of the eigenvalue of T_{w_0}
# on E which deforms to 1, then we have:
#  E~(T_w\phi)=\overline(E(T_{w^-1w_0}))q_E (trivial extension)
#  E~(T_w\phi)=(-1)^a_E\overline(E(T_{w^-1w_0}))q_E (preferred extension)
# where \overline means q->q^-1
  qE:=HeckeCentralMonomials(Hecke(W,v));
  H:=Hecke(W,v^-2);
  tbl:=ShallowCopy(CharTable(H));
  tmp:=CHEVIE.RawData("ClassInfo","2A")(r);
  tbl.classtext:=tmp.classtext;
  tbl.classnames:=tmp.classnames;
  tbl.classparam:=tmp.classparams;
  tbl.identifier:=SPrint("H(^2A",r,")");
  cl:=List(tbl.classtext,x->EltWord (W,x)*LongestCoxeterElement(W));
  tbl.irreducibles:=TransposedMat(List(cl,x->HeckeCharValues(Basis(H,"T")(x))));
  for i in [1..Length(tbl.irreducibles)] do
    tbl.irreducibles[i]:=(-1)^CHEVIE.RawData("Charb","A")
      (tbl.irredinfo[i].charparam[1])*qE[i]*tbl.irreducibles[i];
  od;
  HeckeAlgebraOps.AdjustCharTable(tbl,param);
  return tbl;
end;
##  for tname in ["B", "Bsym", "C"]
CHEVIE.compat.CharTableB := function(tname)
  return function(rank)
    local tbl,cl,f;
    tbl:=CharTable("WeylB",rank);
    if not IsBound(tbl.size) then
       tbl.size:=2^rank*Factorial(rank);
    fi;
    tbl.identifier:=SPrint("W(",tname,rank,")");
    tbl.cartan:=CartanMat(tname,rank);
    cl:=CHEVIE.RawData("ClassInfo","B")(rank);
    for f in RecFields(cl) do tbl.(f):=cl.(f);od;
    tbl.irredinfo:=List(CHEVIE.RawData("CharParams","B")(rank),
      x->rec(charparam:=x, charname:=CHEVIE.RawData("CharName","B")(x)));
    return tbl;
  end;
end;
CHEVIE.compat.HeckeCharTableB := function(tname)
  return function(n, param,sqrtparam)
    local q, tbl;
    q:=List(param,x->-x[1]/x[2]);
    tbl:= CharTableSpecialized(CHEVIE.RawData("Hk","B"), [n, q[1], q[2]]);
    tbl.identifier := SPrint("H(", tname, n, ")");
    tbl.cartan:= CartanMat(tname, n);
    tbl.parameter:= q;
    tbl.classtext:= CHEVIE.RawData("ClassInfo",tname)(n).classtext;
    tbl.classparams:= List(tbl.classparam, x-> x[2]);
    tbl.classnames:= List(tbl.classparams, PartitionTupleToString);
    tbl.irredinfo:=List(CHEVIE.RawData("CharParams",tname)(n),x->
             rec(charparam:=x,charname:=CHEVIE.RawData("CharName","B")(x)));
    HeckeAlgebraOps.AdjustCharTable(tbl,param);
    return tbl;
  end;
end;
CHEVIE.compat.CharTableD := function(l)
  local hi, erg, i, j, lst, lst2, p, sy, syc, tmp, cli, pow,  S, Sval;

  hi:=CharTable("WeylB",l);
  if not IsBound(hi.size) then
    hi.size:=2^l*Factorial(l);
  fi;
  cli:=CHEVIE.RawData("ClassInfo","D")(l);
  if l mod 2 = 0 then
    S:=CharTable("Symmetric",l/2);
    Sval:=function(aa,pp)
      return S.irreducibles[Position(S.classparam,[1,aa[1]])]
                           [Position(S.classparam,[1,pp[1]/2])];
    end;
  fi;

  # classes in subgroup:
  lst:=[];
  for i in [1..Length(hi.classparam)] do
    if Length(hi.classparam[i][2][2]) mod 2=0 then
      # degenerate classes:
      if hi.classparam[i][2][2] = [] and ForAll(hi.classparam[i][2][1]/2,IsInt)
         then
        Append(lst,[i,i]);
      else
        Add(lst,i);
      fi;
    fi;
  od;

  # similarly for characters:
  lst2:=[];
  for i in [1..Length(hi.irredinfo)] do
    sy:=hi.irredinfo[i].charparam[2];
    if sy[1]<=sy[2] then
      # degenerate characters:
      if sy[1]=sy[2] then Append(lst2,[i,i]);
      else Add(lst2,i);
      fi;
    fi;
  od;

  # for power maps:
  pow:=function(i,p,lst)
    local r;
    r:=Position(lst,p[lst[i]]);
    if '-' in cli.classparams[i] and '+' in cli.classparams[r] then
        r:=r+1;
    fi;
    return r;
  end;

  erg:=rec(identifier:=String(Concatenation("W(D",String(l),")")));
  erg.size:=hi.size/2;
  erg.cartan:=CartanMat("D",l);
  erg.centralizers:=hi.centralizers{lst}/2;
  erg.orders:=hi.orders{lst};
  erg.classparam:=cli.classparams;
  erg.classtext:=cli.classtext;
  erg.classnames:=cli.classnames;
  erg.classes:=hi.classes{lst};
  erg.powermap:=[];
  for i in [1..Length(hi.powermap)] do
    if IsBound(hi.powermap[i]) then
      erg.powermap[i]:=List([1..Length(lst)],j->pow(j,hi.powermap[i],lst));
    fi;
  od;
  erg.text:="extracted from generic character table of type WeylB";
  erg.operations:=CharTableOps;
  erg.irredinfo:=List(List(hi.irredinfo{lst2},ShallowCopy),
                         a->rec(charparam:=a.charparam[2],
                         charname:=PartitionTupleToString(a.charparam[2])));
  erg.irreducibles:=hi.irreducibles{lst2}{lst};

  # corrections for degenerate classes and characters:
  for i in [1..Length(lst)] do
    if '+' in erg.classparam[i] or '-' in erg.classparam[i] then
      erg.centralizers[i]:=2*erg.centralizers[i];
      erg.classes[i]:=erg.classes[i]/2;
    fi;
  od;

  i:=1;
  while i<=Length(lst2) do
    sy:=erg.irredinfo[i].charparam;
    if sy[1]=sy[2] then
      erg.irreducibles[i]:=erg.irreducibles[i]/2;
      erg.irreducibles[i+1]:=erg.irreducibles[i+1]/2;
      erg.irredinfo[i].charparam[2]:='+';
      erg.irredinfo[i+1].charparam[2]:='-';
      erg.irredinfo[i].charname:=
                 CHEVIE.RawData("CharName","D")(erg.irredinfo[i].charparam);
      erg.irredinfo[i+1].charname:=
               CHEVIE.RawData("CharName","D")(erg.irredinfo[i+1].charparam);
      for j in [1..Length(lst)] do
        syc:=erg.classparam[j];
        if '+' in syc then
          tmp:=2^(Length(syc[1])-1)*Sval(sy,syc);
          erg.irreducibles[i][j]:=erg.irreducibles[i][j]+tmp;
          erg.irreducibles[i+1][j+1]:=erg.irreducibles[i+1][j+1]+tmp;
          erg.irreducibles[i][j+1]:=erg.irreducibles[i][j+1]-tmp;
          erg.irreducibles[i+1][j]:=erg.irreducibles[i+1][j]-tmp;
        fi;
      od;
      i:=i+2;
    else
      i:=i+1;
    fi;
  od;

  return erg;
end;

CHEVIE.compat.HeckeCharTableD := function(n, param,sqrtparam)
   local r, tbl,u;
   u:=-param[1][1]/param[1][2];
   tbl:= CharTableSpecialized(CHEVIE.RawData("Hk","D"), [n, u]);
   tbl.cartan:= CartanMat("D", n);
   tbl.parameter:= List([1..n], x-> u);
   tbl.classtext:= CHEVIE.RawData("ClassInfo","D")(n).classtext;

   for r in tbl.irredinfo do
      r.charparam:= r.charparam[2];
      r.charname:= CHEVIE.RawData("CharName","D")(r.charparam);
   od;
   tbl.classparam:= List(tbl.classparam, x-> x[2]);
   tbl.classnames:= List(tbl.classparam, CHEVIE.RawData("CharName","D"));
   HeckeAlgebraOps.AdjustCharTable(tbl,param);
   return tbl;
end;

CHEVIE.compat.CharTable2D := function(l)
  local hi, tbl,lst, lst2;
  hi:=CHEVIE.RawData("CharTable","B")(l);
  lst2:=[1..Length(hi.classparams)];
  lst:=Filtered(lst2, i->Length(hi.classparams[i][2]) mod 2=1);
  lst2:=Filtered(lst2,i->CHEVIE.RawData("testchar","2D")
                     (hi.irredinfo[i].charparam));
  tbl:=rec(identifier:=SPrint("W(^2D",l,")"),
    size:=hi.size/2,
    centralizers:=hi.centralizers{lst}/2,
    orders:=hi.orders{lst},
    classes:=hi.classes{lst},
    text:="extracted from generic character table of type B",
    operations:=CharTableOps,
    irredinfo:=List(hi.irredinfo{lst2},a->rec(charparam:=a.charparam,
		   charname:=CHEVIE.RawData("CharName","2D")(a.charparam))),
    irreducibles:=hi.irreducibles{lst2}{lst});
   Inherit(tbl,CHEVIE.RawData("ClassInfo","2D")(l));
   return tbl;
end;

CHEVIE.compat.HeckeCharTable2D := function(l,param,rootparam)
  local hi,cli,lst,tbl,lst2,q;
  q:=-param[1][1]/param[1][2];
  q:=Concatenation([[q^0,-1]],List([2..l],i->[q,-1]));
  hi:=CHEVIE.RawData("HeckeCharTable","B")(l,q,[]);
  lst2:=[1..Length(hi.classparams)];
  lst:=Filtered(lst2,i->Length(hi.classparams[i][2]) mod 2=1);
  tbl:=rec(identifier:=SPrint("H(^2D",l,")"),
	   size:=hi.size/2,
	   orders:=hi.orders{lst},
	   centralizers:=hi.centralizers{lst}/2,
       classes:=hi.classes{lst},
	   text:="extracted from generic character table of HeckeB",
	   operations:=CharTableOps);
  Inherit(tbl,CHEVIE.RawData("ClassInfo","2D")(l));
  lst2:=Filtered(lst2,
	i->CHEVIE.RawData("testchar","2D")(hi.irredinfo[i].charparam));
  tbl.irredinfo:=List(hi.irredinfo{lst2},a->rec(charparam:=a.charparam,
      charname:=CHEVIE.RawData("CharName","2D")(a.charparam)));
  tbl.irreducibles:=TransposedMat(List(tbl.classtext,
    x->HeckeCharValues(Basis(Hecke(CoxeterGroup("B",l),q),"T")
      (Concatenation([1],Replace(x,[1],[1,2,1]))),hi.irreducibles{lst2})));
  HeckeAlgebraOps.AdjustCharTable(tbl,param);
  return tbl;
end;
