#############################################################################
##
#A  tbl/coxi.g                 CHEVIE library                   Meinolf Geck
##
#Y  Copyright (C) 1992 - 2001  The CHEVIE Team
##
##  This file contains the character table of the Hecke algebras of type I.
##
##  The  characters of the groups I_2(m) are parametrized by [d,b] where
##  d  is their degree and b is  their b-function, i.e. the valuation of
##  the  fake degrees. This identifies uniquely the characters when m is
##  odd.  When m is  even, there are  two characters with [d,b]=[1,m/2].
##  The  one  which  maps  the  generators  [1,2]  to  [1,-1] is denoted
##  [1,m/2,"'"]  and  the  one  which  maps  [1,2]  to [-1,1] is denoted
##  [1,m/2,"''"].
##
CHEVIE.Revision("tbl/coxi.g",
   "$Id: coxi.g,v 1.3 2003/12/04 11:26:45 chevie Exp $");

CHEVIE.AddData("CartanMat","I",function(arg)local bond,type,m;
  m:=[[2,0],[0,2]];
  bond:=arg[1];
  if bond=2 then return m;fi;
  if Length(arg)=2 then type:=arg[2];
  elif bond mod 2=0 then type:=1;
  else type:=E(2*bond)+E(2*bond)^-1;
  fi;
  m[1][2]:=-type;m[2][1]:=(2+E(bond)+E(bond)^-1)/m[1][2];return m;
end);

CHEVIE.AddData("PrintDiagram","I",function(bond,indices,title,type)
  Print(title," ");
  if type=E(2*bond)+E(2*bond)^-1 then 
    Print(indices[1]," -",String(bond),"- ",indices[2],"\n");
  elif type=1 then  Print(indices[1]," >",String(bond),"> ",indices[2],"\n");
  else Print(indices[1]," ?",String(bond),"? ",indices[2],"\n");
  fi;
end);

CHEVIE.AddData("ReflectionName","I",function(arg)local bond,type,TeX;
  bond:=arg[1];TeX:=IsBound(arg[Length(arg)].TeX);
  if Length(arg)=3 then type:=arg[2];
  elif bond mod 2=0 then type:=1;
  else type:=E(2*bond)+E(2*bond)^-1;
  fi;
  if type=1 then 
    if TeX then return SPrint("I_2(",bond,")");
    else return SPrint("I2(",bond,")");fi;
  elif type=E(2*bond)+E(2*bond)^-1 then 
    if bond mod 2=1 then 
      if TeX then return SPrint("I_2(",bond,")");
      else return SPrint("I2(",bond,")");fi;
    else 
      if TeX then return SPrint("I_{\\hbox{sym}2}(",bond,")");
      else return SPrint("Isym2(",bond,")");fi;
    fi;
  elif TeX then
       return SPrint("I_?(",Format(type^2/(2+E(bond)+E(bond)^-1),rec(TeX:=1)),
                     ")(",bond,")");
  else return SPrint("I?(",type^2/(2+E(bond)+E(bond)^-1),")(",bond,")");
  fi;
end);

CHEVIE.AddData("GeneratingRoots", "I", function(m)local a, b, r;
  a := E(2*m)^(m-1);
  b := ComplexConjugate(a);
  if m mod 2 = 0 then r := ER(m/2); else r := 1; fi;
  return [[1, 0], [r*(a+b)/2, r*(a-b)/2/E(4)]];
end);

CHEVIE.AddData("EigenvaluesGeneratingReflections","I",m->[-1,-1]);

CHEVIE.AddData("Size", "I", function(arg) return 2*arg[1]; end);

CHEVIE.AddData("ReflectionDegrees","I",m->[2,m]);

CHEVIE.AddData("ReflectionCoDegrees","I",m->[0,m-2]);

CHEVIE.AddData("NrConjugacyClasses","I", m->2+QuoInt(m-1,2)+((m+1)mod 2)*2);

#how to make a .charname from a .charparam
CHEVIE.AddData("CharName","I",function(x,option)local s;
  if IsList(x[1]) then return PartitionTupleToString(x);
  else
    if IsBound(option.TeX) then s:="phi";else s:="\\phi_";fi;
    s:=SPrint(s,"{",x[1],",",x[2],"}");
    if Length(x)=3 then Append(s,x[3]);fi;
    return String(s);
  fi;
end);

CHEVIE.AddData("CharInfo","I",function(m)local res,applyf;
  res:=rec(positionId:=1);
  res.charparams:=[[1,0]];
  if m mod 2=0 then res.positionSgn:=4;res.positionRefl:=5; 
     Append(res.charparams,[[1,m/2,"'"],[1,m/2,"''"]]);
  else res.positionSgn:=2;res.positionRefl:=3;fi;
  Add(res.charparams,[1,m]);
  Append(res.charparams,List([1..QuoInt(m-1,2)],i->[2,i]));
  res.b:=List(res.charparams,x->x[2]);
  res.B:=List(res.charparams,function(phi)
    if phi[1]=1 then return phi[2]; else return m-phi[2]; fi;
  end);
  res.a:=List(res.charparams,function(phi)
     if phi[1]<>1 or phi[2]=m/2 then return 1;else return phi[2]; fi;
  end);
  res.A:=List(res.charparams,function(phi)
     if phi[1]=1 or phi[2]=m/2 then return m-1;else return phi[2];fi;
  end);
  return res;
end);

CHEVIE.AddData("WordsClassRepresentatives", "I", function(m) local r, x, i;
  if IsInt(m/2) then r:=[[],[1],[2]]; else r:=[[],[1]]; fi;
  x:=[1,2];
  for i in [1..QuoInt(m,2)] do Add(r, ShallowCopy(x)); Append(x,[1,2]); od;
  return r;
end);

#############################################################################
##
#F  CHEVIE.RawData("ClassInfo","I")( <m> ) . .. conjugacy classes for type I.
##
CHEVIE.AddData("ClassInfo","I",function(m)local r, clnp;
  r := CHEVIE.RawData("WordsClassRepresentatives", "I")(m);
  clnp := List(r,IntListToString);
  return rec(classtext:=r, classnames:=clnp, classparams:=clnp);
end);

CHEVIE.AddData("ClassName", "I", IntListToString);

##########################################################################
##
#F  HeckeCharTable( <m>, param, rootparam)  . . . . . . . . .
##  . . . . . . . .  character table of the Hecke algebra of dihedral type
##
##
CHEVIE.AddData("HeckeCharTable","I",function(m,param,rootparam)
  local u, v, squv, cl, cli, r, e, t, ct, k, g1, i, g2, gen, perm, tbl, j;
  u:=-param[1][1]/param[1][2]; v:=-param[2][1]/param[1][2];
  if m mod 2=0 then
   if IsBound(rootparam[1]) and IsBound(rootparam[2]) then
        squv:=rootparam[1]*rootparam[2];
   else squv:=GetRoot(u*v,2,"CharTable(Hecke(I2(",m,")))");
   fi;
  else squv:=u;
  fi;
  cl:=[1];
  cli := CHEVIE.RawData("ClassInfo","I")(m);
  r:=cli.classtext;
  e:=E(m);
  if IsInt(m/2) then
    Append(cl,[m/2,m/2]);
    for i in [1..m/2] do
      Add(cl,2);
    od;
    cl[Length(cl)]:=1;
    t:=[];
    t[1]:=[[[u]],[[v]]];
    t[2]:=[[[u]],[[-u^0]]];
    t[3]:=[[[-v^0]],[[v]]];
    t[4]:=[[[-v^0]],[[-v^0]]];
    ct:=List(t,i->CharRepresentationWords(i,r));
    for j in [1..m/2-1] do
      ct[j+4]:=[];
      for i in [1..Length(r)] do
        k:=Length(r[i])/2;
        if r[i]=[] then ct[j+4][i]:=2*v^0;
        elif r[i]=[1] then ct[j+4][i]:=u-1;
        elif r[i]=[2] then ct[j+4][i]:=v-1;
        else ct[j+4][i]:=squv^k*(e^(k*j)+e^(-k*j));
        fi;
      od;
    od;
   # for j in [1..m/2-1] do
   #   Add(t,[[[-u^0,u^0],[0*u,u]],
   #          [[v,0*v],[u+v+squv*(e^j+e^-j),-v^0]]]);
   # od;
  else
    Add(cl,m);
    for i in [1..(m-1)/2] do Add(cl,2); od;
    t:=[];
    t[1]:=[[[u]],[[v]]];
    t[2]:=[[[-v^0]],[[-v^0]]];
    ct:=List(t,i->CharRepresentationWords(i,r));
    for j in [1..(m-1)/2] do
      ct[j+2]:=[];
      for i in [1..Length(r)] do
        k:=Length(r[i])/2;
        if r[i]=[] then ct[j+2][i]:=2*v^0;
        elif r[i]=[1] then ct[j+2][i]:=u-1;
        elif r[i]=[2] then ct[j+2][i]:=v-1;
        else ct[j+2][i]:=squv^k*(e^(k*j)+e^(-k*j));
        fi;
      od;
    od;
   #for j in [1..(m-1)/2] do
   #   Add(t,[[[-u^0,u^0],[0*u,u]],
   #          [[v,0*v],[u+v+squv*(e^j+e^-j),-v^0]]]);
   # od;
  fi;
##    if m=2 then
##      W:=CoxeterGroup([[2,0],[0,2]]);
##    elif m mod 2=0 then
##      W:=CoxeterGroup([[2,-1],[-2-E(m)-E(m)^-1,2]]);
##    else
##      W:=CoxeterGroup([[2,-E(2*m)-E(2*m)^-1],[-E(2*m)-E(2*m)^-1,2]]);
##    fi;
  g1:=();i:=2; while 2*i<=m+1 do g1:=g1*(i,m-i+2); i:=i+1;od;
  g2:=();i:=1; while 2*i<=m do g2:=g2*(i,m-i+1); i:=i+1;od;
  gen := [g1,g2];
  perm := function(l) if Length(l)=0 then return ();else return
                      Product(gen{l});fi;end;
  tbl:=rec(identifier:=SPrint("H(I2(",m,"))"),
    cartan:=CartanMat("I",2,m), size:=2*m,
    irredinfo:=List(CHEVIE.RawData("CharInfo","I")(m).charparams,x->rec(
       charparam:=x,charname:=CHEVIE.RawData("CharName","I")(x,rec()))),
    parameter:=[u,v],
    classes:=cl, centralizers:=List(cl,i->2*m/i),
    orders:=List(r,i->OrderPerm(perm(i))),
    powermap:=[], irreducibles:=ct*v^0);
  Inherit(tbl,cli);
  tbl := CHEVIE.compat.MakeCharacterTable(tbl);
  CHEVIE.compat.AdjustHeckeCharTable(tbl,param);
  return tbl;
end);

CHEVIE.AddData("Representation","I",function(m,i)
  return CHEVIE.RawData("HeckeRepresentation","I")(m,
  [[1,-1],[1,-1]],[1,1],i);end);

CHEVIE.AddData("HeckeRepresentation","I",function(m,param,rootparam,i)
 local u,v,squv;
  u:=-param[1][1]/param[1][2]; v:=-param[2][1]/param[1][2];
  if m mod 2=0 then
   if IsBound(rootparam[1]) and IsBound(rootparam[2]) then
        squv:=rootparam[1]*rootparam[2];
   else squv:=GetRoot(u*v,2,"Representation(Hecke(I2(",m,")),[",i,"])");
   fi;
  else squv:=u;
  fi;
  if i=1 then return [[[u]],[[v]]];fi;
  if IsInt(m/2) then i:=i-2;fi;
  if i=0 then return [[[u]],[[-u^0]]];
  elif i=1 then return [[[-v^0]],[[v]]];
  elif i=2 then return [[[-v^0]],[[-v^0]]];
  else return [[[-u^0,u^0],[0*u,u]],[[v,0*v],
                  [u+v+squv*(E(m)^(i-2)+E(m)^(2-i)),-v^0]]];
  fi;
end);

CHEVIE.AddData("Frobenius","I",function(m,sqrtu,j)
  return [[0,1/sqrtu/(E(2*m)^j+E(2*m)^-j)],
                           [sqrtu*(E(2*m)^j+E(2*m)^-j),0]]*sqrtu^0;
end);

CHEVIE.AddData("PoincarePolynomial","I",function(m,param)local u,v;
  u:=-param[1][1]/param[1][2];v:=-param[2][1]/param[2][2];
  if IsInt(m/2) then return Sum([1..m/2],i->(u*v)^(i-1))*(u+1)*(v+1);
  else return Sum([1..m],i->u^(i-1))*(u+1);
  fi;
end);

#############################################################################
##
#F  SchurElement( <m>, <phi>, <u> [, <v> [,<squv>]] )   . . . . . . .
#F   . . . . . . . . . . . . . . . . . . Schur element for dihedral type
##
##  'CHEVIE.RawData("SchurElement","I")'  returns  the  constant arising
##  from  the  Schur  relations  for  the  irreducible  characters  with
##  charparam phi of the Hecke algebra of type I with parameters u, v.
##
##  The  constant corresponding to phi  is defined to be  P/D where P is
##  the Poincare polynomial and D is the generic degree of phi.
##
CHEVIE.AddData("SchurElement","I",function(m,phi,param,rootparam)
  local u,v,squv,i,e,b;
  u:=-param[1][1]/param[1][2];v:=-param[2][1]/param[2][2];b:=phi[2];
  if phi[1]=1 then
    if b=0 then
      if IsInt(m/2) then return Sum([1..m/2],i->(u*v)^(i-1))*(u+1)*(v+1);
      else return Sum([1..m],i->u^(i-1))*(u+1);
      fi;
    elif b=m/2 then
      if phi[3]="'" then
	   return v^(-m/2)*Sum([1..m/2],i->u^(i-1)*v^(m/2-i))*(u+1)*(v+1);
      else return u^(-m/2)*Sum([1..m/2],i->u^(i-1)*v^(m/2-i))*(u+1)*(v+1);
      fi;
    elif IsInt(m/2) then
           return (u*v)^(-m/2)*Sum([1..m/2],i->(u*v)^(i-1))*(u+1)*(v+1);
    else return u^(-m)*Sum([1..m],i->u^(i-1))*(u+1);
    fi;
  else e:=E(m);
    if IsBound(rootparam[1]) and IsBound(rootparam[2]) then
         squv:=rootparam[1]*rootparam[2];
    else squv:=GetRoot(u*v,2,"SchurElement(Hecke(I2(",m,"),",phi,"))");
    fi;
    return (squv-e^b)*(squv-e^(-b))*(u+(e^b+e^(-b))*squv+v)*
                                        (-m)/u/v/(e^b-e^(-b))^2;
  fi;
end);

CHEVIE.AddData("FakeDegree","I",function(m,phi,q)
  if phi[1]=1 then return q^phi[2];else return q^phi[2]+q^(m-phi[2]);fi;
end);

CHEVIE.AddData("CharTable","I",function(m)local res;
  res:=CHEVIE.RawData("HeckeCharTable","I")(m,[[1,-1],[1,-1]],[1,1]);
  res.identifier:=SPrint("W(I2(",m,"))");
  return res;
end);

CHEVIE.AddData("UnipotentCharacters","I",
 function(e)local si,sj,nc,uc,c,i,j,ueigen;
  if e mod 2=0 then
    if e=4 then
      uc:=CHEVIE.RawData("UnipotentCharacters","B")(2);
      uc.harishChandra[1].charNumbers:=[4,5,1,3,2];
      return uc;
    elif e=6 then
      uc:=CHEVIE.RawData("UnipotentCharacters","G2")();
      uc.harishChandra[1].series:="G";
      uc.harishChandra[1].cartanType:=ER(3);
      uc.harishChandra[1].relativeType.series:="G";
      uc.harishChandra[1].relativeType.cartanType:=-ER(3);
      uc.families[2].charNumbers:=[4];
      uc.families[3].charNumbers{[3,4]}:=[2,3];
      return uc;
    else 
      Print("***** WARNING: UnipotentCharacters(I2(",e,")) not implemented\n");
      return false;
    fi;
  fi;
# For I2(e) with odd e there are 3 families: Is, St and a big one.
# in the big one the cuspidal chars are S(k,l) where 0<k<l<e-k and
# G.Malle associates to it a symbol (s_0,...,s_{e-1}) where
#   s_i=[0] for i\ne 0,k,l,k+l, s_0=s_{k+l}=[0,1] and s_k=s_l=[]
# The principal series symbols are S(0,l) where 0<l<e/2 with associated symbol
#   s_i=[0] for i\ne 0,l and s_0=s_l=[1].
# The b is min(k+l,e-k-l) and the eigenvalue of F is E(e)^(-k*l).
# Fourier is as given below.
  uc:=rec();
  nc:=Concatenation(List([1..(e-1)/2],k->List([k+1..e-k-1],l->[k,l])));
  uc.harishChandra:=Concatenation(
   [ rec(relativeType:=rec(series:="I",indices:=[1,2],rank:=2,bond:=e),
         parameterExponents:=[1,1],
	 levi:=[],
	 eigenvalue:=1,
	 cuspidalName:="",
	 TeXCuspidalName:="",
	 charNumbers:=[1..(e+3)/2]
	 )],
   List([1..(e-1)^2/4-(e-1)/2],x->
     rec(relativeType:=rec(series:="A",indices:=[],rank:=0),
         parameterExponents:=[],
         levi:=[1,2],
         eigenvalue:=E(e)^(-nc[x][1]*nc[x][2]),
         cuspidalName:=SPrint("I2(",e,")[",nc[x][1],",",nc[x][2],"]"),
         TeXCuspidalName:=SPrint("I_2(",e,")[",nc[x][1],",",nc[x][2],"]"),
         charNumbers:=[x+(e+3)/2])));
  nc:=Concatenation(List([1..QuoInt(e,2)],l->[0,l]),nc);
  uc.charSymbols:=Concatenation([false,false],List(nc,function(s)local S,k,l;
    S:=List([1..e],i->[0]);k:=s[1];l:=s[2];
    if k<>0 then S[1]:=[0,1];S[1+((k+l)mod e)]:=[0,1];S[k+1]:=[];S[l+1]:=[];
    else S[1]:=[1];S[l+1]:=[1];
    fi;
    return S;
    end));
  uc.charSymbols[1]:=List([1..e],x->[0]);uc.charSymbols[1][e]:=[2];
  uc.charSymbols[2]:=List([1..e],x->[0,1]);uc.charSymbols[2][e]:=[1,2];
  c:=a->E(e)^a+E(e)^(-a);
  uc.families:=[Family("C1",[1]),
		rec(fourierMat:=List(nc,i->List(nc,j->
# (-1)^Number([i[1],j[1]],x->x=0)*  This sign is in [Malle, example 6.29]
                                          (c(i*Reversed(j))-c(i*j))/e)),
		    eigenvalues:=List(nc,s->E(e)^(-s[1]*s[2])),
                    charNumbers:=2+[1..(e-1)^2/4],
	            charLabels:=List([1..(e-1)^2/4],String),
	            almostCharLabels:=List([1..(e-1)^2/4],String),
		    special:=1),
                Family("C1",[2])];
  if e=5 then uc.families[2].perm:=(3,4);  fi; ## inserted 27.2.2003: Gunter
  uc.a:=Concatenation([0,e],List(nc,x->1));
  uc.A:=Concatenation([0,e],List(nc,x->e-1));
  return uc;
end);
