#############################################################################
##
#A  tbl/weylbc.g                CHEVIE library                  Gtz Pfeiffer
##
#Y  Copyright (C) 1994 - 2001  The CHEVIE Team
##
##  This file contains special functions for Coxeter groups and Iwahori-Hecke
##  algebras  of type  B.
##
CHEVIE.Revision("tbl/weylbc.g",
   "$Id: weylbc.g,v 1.4 2003/12/04 11:26:45 chevie Exp $");

CHEVIE.AddData("CartanMat","B",function (arg) local  n,type,a;
  n:=arg[1];
  if Length(arg)=2 then type:=arg[2];else type:=2;fi;
  a:=CHEVIE.RawData("CartanMat","A")(n);
  a[1][2]:=-type;a[2][1]:=2/a[1][2];
  return a;
end);

CHEVIE.AddData("PrintDiagram","B",function(r,indices,title,type)local i;
  Print(title," ");
  if type=1 then Print(indices[1]," >=> ",indices[2]);
  elif type=2 then Print(indices[1]," <=< ",indices[2]);
  elif type=ER(2) then Print(indices[1]," = ",indices[2]);
  else Print(indices[1]," ?=? ",indices[2]);
  fi;
  for i in [3..r] do Print(" - ",indices[i]);od;
  Print("\n");
end);

CHEVIE.AddData("ReflectionName","B",function(arg)local i,r,type,option;
  r:=arg[1];option:=arg[Length(arg)];
  if Length(arg)=3 then type:=arg[2];else type:=2;fi;
  if type=2 then 
    if IsBound(option.TeX) then return SPrint("B_",TeXBracket(r));
    else return SPrint("B",r);fi;
  elif type=1 then 
    if IsBound(option.TeX) then return SPrint("C_",TeXBracket(r));
    else return SPrint("C",r); fi;
  elif type=ER(2) then 
    if IsBound(option.TeX) then return SPrint("B^{\\hbox{sym}}_",TeXBracket(r));
    else return SPrint("Bsym",r); fi;
  elif IsBound(option.TeX) then 
         return SPrint("B^?_",TeXBracket(r),"(",Format(type,option),")");
  else return SPrint("B?",r,"(",Format(type),")");
  fi;
end);

CHEVIE.AddData("GeneratingRoots", "B", function(l)local rts, i;
  rts := List([1..l],i->0*[1..l]);
  for i in [1..l-1] do rts[i]{[i,i+1]} := [1,-1]; od;
  rts[l][l] := 1;
  # CHEVIE ordering
  return rts{[l,l-1..1]};
end);

CHEVIE.AddData("GeneratingRoots", "C", function(l)local rts, i;
  rts := List([1..l],i->0*[1..l]);
  for i in [1..l-1] do rts[i]{[i,i+1]} := [1,-1]; od;
  rts[l][l] := 2;
  # CHEVIE ordering
  return rts{[l,l-1..1]};
end);

CHEVIE.AddData("ReflectionDegrees","B", n->2*[1..n]);

CHEVIE.AddData("ReflectionCoDegrees","B", n->2*[0..n-1]);

CHEVIE.AddData("Size", "B",function(arg)return 2^arg[1]*Factorial(arg[1]);end);

#############################################################################
##
#F  ClassParams( <n> ) . . . . . . . . . . . . . class parameters for type B
##
CHEVIE.AddData("ClassParams", "B", function(arg)
  return PartitionTuples(arg[1], 2);
end);

CHEVIE.AddData("NrConjugacyClasses", "B", function(arg)
  return NrPartitionTuples(arg[1], 2);
end);

#how to make classname from classparam
CHEVIE.AddData("ClassName", "B",  PartitionTupleToString);

#############################################################################
##
#F  WordClass( <pi> )  . . . . . . representative of conjugacy class 
##  parametrized by pi (of minimal length)
##
CHEVIE.AddData("WordClass", "B", function(pi) local  w,  i,  l;
    w:= [];  i:= 1;

    #  handle signed cycles in reversed order.
    for l in Reversed(pi[2]) do
      Append(w, [i, i-1 .. 2]); Append(w, [1 .. i+l-1]);
      i:= i+l;
    od;

    #  the unsigned cycles.
    for l in pi[1] do
      Append(w, [i+1 .. i+l-1]);
      i:= i+l;
    od;

   return w;
end);

#############################################################################
##
#F  ClassInfo( <n> ) . . . . . . . . conjugacy classes for type B
##
##  ClassInfo returns a record with three components:
##    classtext:   representatives of minimal length in  the  conjugacy
##                 classes, as words in generators in standard order
##    classparams:  double partitions, parameterizing the classes
##    classnames:  strings for double partitions
##
##  The  ordering  corresponds  to the  order of  the columns of the ordinary
##  character table of the Coxeter group of type $B_n$, as returned by the GAP
##  function 'CharTable("WeylB", <n>)'.
##
CHEVIE.AddData("ClassInfo", "B", function(n) local res;
  res := rec(classparams := CHEVIE.RawData("ClassParams", "B")(n));
  res.classnames := List(res.classparams, CHEVIE.RawData("ClassName", "B"));
  res.classtext := List(res.classparams,CHEVIE.RawData("WordClass", "B"));
  return res;
end);

#############################################################################
##
#V  ClassParamB( <W>, <w> )  . . . . . . . .  class parameter of w in type B.
##
##  given an element w of a Coxeter group W of type B, returns the classparam
##
# faster version of 'ClassParamB':
# parameter n = semisimple rank of W
CHEVIE.AddData("ClassParameter", "B", function(n, w)
  local x, i, res, mark, cyc, j;

  x:=();
  for i in w do
    if i=1 then x:=x*(1,n+1); else x:=x*(i-1,i)(i-1+n,i+n);fi;
  od;

  res:=[[],[]];
  mark:=[1..n];
  for i in [1..n] do
    if mark[i]<>0 then
      cyc:=CyclePermInt(x,i);
      if i+n in cyc then Add(res[2],Length(cyc)/2);
      else Add(res[1],Length(cyc));
      fi;
      for j in cyc do
        if j>n then mark[j-n]:=0;
        else mark[j]:=0;
        fi;
      od;
    fi;
  od;

  Sort(res[1]); Sort(res[2]);
  return [Reversed(res[1]),Reversed(res[2])];
end);

#############################################################################
##
#F  CharParams( <n> ) . . . . . . . . . . . . . . . characters for type B
##
CHEVIE.AddData("CharParams", "B", n->PartitionTuples(n,2));

#how to make charname from charparam
CHEVIE.AddData("CharName", "B",
  function(arg) return PartitionTupleToString(arg[1]); end);

CHEVIE.AddData("Charb","B", function(p) local pp, m, res;
  pp:=SymbolPartitionTuple(p,1); m:=Length(pp[2]);
  res:=pp[1]*[m,m-1..0];
  if pp[2]<>[] then res:=res+pp[2]*[m-1,m-2..0]; fi;
  return 2*res+Sum(pp[2])-m*(m-1)*(4*m+1)/6;
end);

CHEVIE.AddData("CharInfo","B",function(n)local res;
  res:=rec(charparams:=CHEVIE.RawData("CharParams","B")(n),
    positionSgn:=n+1);
  res.positionId:=Length(res.charparams)-1;
  res.positionRefl:=res.positionId-2;
  res.a:=List(res.charparams,
    p->LowestPowerGenericDegreeSymbol(SymbolPartitionTuple(p,1)));
  res.A:=List(res.charparams,
    p->HighestPowerGenericDegreeSymbol(SymbolPartitionTuple(p,1)));
  res.b:=List(res.charparams,CHEVIE.RawData("Charb","B"));
  res.B:=res.a+res.A-res.b;
  return res;
end);

##  essentially the library call to table "WeylB", l, but different
##  in GAP 3 and GAP 4
CHEVIE.IndirectAddData("CharTable", ["B"], CHEVIE.compat.CharTableB);

#############################################################################
##
#V  HeckeCharTable( <n>, <para> ). . .  character table of $H(B_n)$.
##
##  HeckeCharTable returns the character  table of the Hecke
##  algebra  $H(B_n)$ with parameters <para>.
##
CHEVIE.tmp := ShallowCopy(CharTableWeylB);

CHEVIE.tmp.identifier := "HeckeB";

CHEVIE.tmp.specializedname:=nq->SPrint("H(B",nq[1],")");

CHEVIE.tmp.order:=nq-> 2^nq[1]*Factorial(nq[1]);
CHEVIE.tmp.size:=nq-> 2^nq[1]*Factorial(nq[1]);

CHEVIE.tmp.domain:= function(nq)
   return IsList(nq) and Length(nq) = 3 and IsInt(nq[1]) and nq[1] > 0;
end;

CHEVIE.tmp.text:= "generic character table of Hecke algebras of type B";

CHEVIE.tmp.classparam:=[nq -> PartitionTuples(nq[1],2)];
CHEVIE.tmp.charparam:=[nq -> PartitionTuples(nq[1],2)];

CHEVIE.tmp.irreducibles := [[function(nq, gamma, pi)
   local n, q, Q, k, val, t, nn, alpha, dif, BHk;

   n:= nq[1];  q:= nq[3];  Q:= nq[2]; #  for the sake of clearness.

   #  termination condition.
   if n = 0 then
      return q^0;
   fi;

   val:= 0*q; #  initialize character value.

   BHk:=CHEVIE.RawData("Hk","B").irreducibles[1][1];

   #  positive cycles first.
   if pi[1] <> [] then

      k:= pi[1][1]; #  get length of the longest cycle.

      #  loop over double paritions of n-k.
      for alpha in PartitionTuples(n-k, 2) do
         dif:= [];
         dif[1]:= DifferencePartitions(gamma[1], alpha[1]);
         dif[2]:= DifferencePartitions(gamma[2], alpha[2]);
         if dif[1] <> false and dif[2] <> false then
            dif:= rec(cc:= dif[1].cc + dif[2].cc, ll:= dif[1].ll + dif[2].ll);
            val:= val + (q-1)^(dif.cc-1) * (-1)^dif.ll * q^(k-dif.ll-dif.cc)
                * BHk([n-k, Q, q], alpha,[pi[1]{[2..Length(pi[1])]}, pi[2]]);
         fi;
      od;

   else # pi[2] = []

      #  get length of the longest cycle.
      k:= pi[2][1];

      #  loop over k-hooks in gamma[1].
      nn:= Sum(gamma[1]);
      if nn >= k then
         for alpha in Partitions(nn - k) do
            dif:= DifferencePartitions(gamma[1], alpha);
            if dif <> false and dif.cc = 1 then
               val:= val + Q * (-1)^dif.ll * q^(n+dif.d)
                 * BHk([n-k, Q, q], [alpha, gamma[2]],
                                    [pi[1], pi[2]{[2..Length(pi[2])]}]);
            fi;
         od;
      fi;

      #  loop over hooks in gamma[2].
      nn:= Sum(gamma[2]);
      if nn >= k then
         for alpha in Partitions(nn - k) do
            dif:= DifferencePartitions(gamma[2], alpha);
            if dif <> false and dif.cc = 1 then
               val:= val + (-1)^(dif.ll+1) * q^(n+dif.d)
                 * BHk([n-k, Q, q], [gamma[1], alpha],
                                    [pi[1], pi[2]{[2..Length(pi[2])]}]);
            fi;
         od;
      fi;

   fi;

   #  return the result.
   return val;

end]];

##  Info... statements commented out to make it GAP4 compatible
CHEVIE.tmp.matrix:= function(nq)

   local scheme, beta, pm, i, m, k, t, n, x, y, np, col, res, charCol, hooks,
     DoublePartitions;


#############################################################################
##
#F  DoublePartitions( <n> ) . . . . . . . . . . . . . .  pairs of partitions.
##
##  JM: seems to be used just for the Sortex at the end. Should be suppressed
##      by using a different enumeration...
##
DoublePartitions:= function(n) local m, k, pm, t, s, res;

   if n = 0 then return [[[], []]]; fi;

   pm:= List([1..n], x->[]);

   #  second position.
   for m in [1..n] do

      #  add the m-cycle.
      Add(pm[m], [[], [m]]);
      for k in [m+1..n] do
         for t in pm[k-m] do
            s:= [[], [m]]; Append(s[2], t[2]); Add(pm[k], s);
         od;
      od;
   od;

   #  first position.
   for m in [1..QuoInt(n,2)] do

      #  add the m-cycle.
      Add(pm[m], [[m], []]);
      for k in [m+1..n-m] do
         for t in pm[k-m] do
            s:= [[m], t[2]]; Append(s[1], t[1]); Add(pm[k], s);
         od;
      od;
   od;

   #  collect.
   res:= [];
   for k in [1..n-1] do
      for t in pm[n-k] do
         s:= [[k], t[2]]; Append(s[1], t[1]); Add(res, s);
      od;
   od;
   Add(res, [[n], []]);
   Append(res, pm[n]);

   return res;

end;
   n:= nq[1]; x:= nq[3]; y:= nq[2];

   pm:= [];
   scheme:= [];

   #  how to encode all hooks.
   hooks:= function(beta, m)
      local i, j, k, hk, pr, cbs, prs, leg, hks, lb, ll, lg, lh, gamma, new;

      hks:= List([1..m], x->[]);
      prs:= [];
      lb:= [Length(beta[1]), Length(beta[2])];

      #  find all hooks.
      for i in [1, 2] do
         prs[i]:= [];
         for j in beta[i] do
            leg:= 0;
            for k in Reversed([0..j-1]) do
               if  k in beta[i] then
                   leg:= leg + 1;
               else
                   Add(prs[i],
                        rec(from:= j, to:= k, leg:= leg, pow:= m+k-lb[i]));
               fi;
            od;
         od;
      od;

      #  construct combinations.
      cbs:= List(prs[1], x-> [[x], []]);
      Append(cbs, List(prs[2], x-> [[], [x]]));
      for hk in cbs do

         #  extend.
         for pr in prs[1] do
            if hk[2] = [] and pr.to > hk[1][Length(hk[1])].from then
               new:= List(hk, ShallowCopy);
               Add(new[1], pr);
               Add(cbs, new);
            fi;
         od;
         for pr in prs[2] do
            if hk[2] = [] or pr.to > hk[2][Length(hk[2])].from then
               new:= List(hk, ShallowCopy);
               Add(new[2], pr);
               Add(cbs, new);
            fi;
         od;


         #  encode.
         ll:= Sum(hk[1], x-> x.from - x.to) + Sum(hk[2], x-> x.from - x.to);
         lg:= Sum(hk[1], x-> x.leg) + Sum(hk[2], x-> x.leg);
         lh:= Length(hk[1]) + Length(hk[2]);
         new:= rec(wgt:= [(-1)^lg * x^(ll-lg-lh) * (x-1)^(lh-1), 0],
                   adr:= 1);
         if lh = 1 then
            if IsBound(hk[1][1]) then
               new.wgt[2]:= (-1)^lg * y * x^hk[1][1].pow;
            else
               new.wgt[2]:= (-1)^(lg+1) * x^hk[2][1].pow;
            fi;
         fi;

         #  recalculate address.
         if ll < m then
            gamma:= [];
            for i in [1, 2] do
                gamma[i]:= Difference(beta[i], List(hk[i], x-> x.from));
                UniteSet(gamma[i], List(hk[i], x-> x.to));
                if 0 in gamma[i] then
                    j:= 0;
                    while j < Length(gamma[i]) and gamma[i][j+1] = j do
                       j:= j+1;
                    od;
                    gamma[i]:= gamma[i]{[j+1..Length(gamma[i])]}-j;
                fi;
            od;
            new.adr:= Position(pm[m-ll], gamma);
         fi;

         #  insert.
         Add(hks[ll], new);
      od;
      return hks;
   end;

   #  collect hook encodings.
##     InfoCharTable2("#I  Scheme: \c");
   for i in [1..n] do
##        InfoCharTable2(i, " \c");
      pm[i]:= List(PartitionTuples(i, 2), p-> List(p, BetaSet));
      scheme[i]:= [];
      for beta in pm[i] do
         Add(scheme[i], hooks(beta, i));
      od;
   od;
##     InfoCharTable2("done.\n");

   #  how to construct a new column.
   charCol:= function(n, t, k, p) local col, pi, hk, val;
      col:= [];
      for pi in scheme[n] do
         val:= 0*y;
         for hk in pi[k] do
            val:= val + hk.wgt[p] * t[hk.adr];
         od;
         Add(col, val);
      od;
      return col;
   end;

   #  construct the columns.
##     InfoCharTable2("#I  Cycles: \c");
   pm:= List([1..n], x->[]);

   #  second position.
   for m in [1..n] do

      #  add the m-cycle.
##        InfoCharTable2(m, " \c");
      Add(pm[m], charCol(m, [1], m, 2));
      for k in [m+1..n] do
         for t in pm[k-m] do
            Add(pm[k], charCol(k, t, m, 2));
         od;
      od;
   od;

   #  first position.
   for m in [1..QuoInt(n,2)] do

      #  add the m-cycle.
##        InfoCharTable2(m, " \c");
      Add(pm[m], charCol(m, [1], m, 1));
      for k in [m+1..n-m] do
         for t in pm[k-m] do
            Add(pm[k], charCol(k, t, m, 1));
         od;
      od;
   od;
##     InfoCharTable2("done.\n");

   #  collect.
##     InfoCharTable2("#I  Tables: \c");
   res:= [];
   for k in [1..n-1] do
##        InfoCharTable2(k, " \c");
      for t in pm[n-k] do
         Add(res, charCol(n, t, k, 1));
      od;
   od;
   Add(res, charCol(n, [1], n, 1));
   Append(res, pm[n]);
##     InfoCharTable2("done.\n");

   res:=Permuted(res,Sortex(DoublePartitions(n))/Sortex(PartitionTuples(n,2)));

   return TransposedMat(res);

end;

CHEVIE.AddData("Hk","B",CHEVIE.tmp);
Unbind(CHEVIE.tmp);

CHEVIE.IndirectAddData("HeckeCharTable", ["B"],
                                       CHEVIE.compat.HeckeCharTableB);

#############################################################################
##
#F  PoincarePolynomial( <n>, <para> )  Poincare polynomial for type B.
##
##  PoincarePolynomial returns the Poincare polynomial of the
##  Coxeter group $W$ of type $B_n$, ie.  the sum of $q^l(w)$ over all
##  elements $w$ of $W$.
##
CHEVIE.AddData("PoincarePolynomial","B",
function(n, para)
   local q1,q2;
   q1:=-para[1][1]/para[1][2]; q2:=-para[2][1]/para[2][2];
   return Product([0..n-1], i-> (q2^i*q1 + 1) * Sum([0..i], k-> q2^k));
end);


#############################################################################
##
#F  SchurElement( <n> , <char>, <param>, <sqrtparam> )  . . . . Schur element
#F  for type B.
##
##  SchurElement returns the Schur element  for the character labeled <char>
##  in table of the Hecke algebra  of type $B_n$ with parameters <param> and
##  square roots  <sqrtparam>. Here  <param> can be  a single  parameter (in
##  which case  all parameters are  assumed to be equal  to this one),  or a
##  pair of parameters  (in which case the first parameter  is associated to
##  the first generator, and the second to all the remaining generators). It
##  is also possible to pass the complete list of parameters as an argument.
##
##  [Reference: Carter II, 13.5, p. 447.]
##
CHEVIE.AddData("SchurElement","B",
function(n,pi, param,sqrtparam)
   local uuu, s, res, m, i, j, u, v;
   uuu:= k-> Sum([0..k-1], e-> u^e); # quantum integers.
   u:= -param[2][1]/param[2][2];  v:= -param[1][1]/param[1][2];

   # transform charparam pi (double partition) into a symbol
   s:=SymbolPartitionTuple(pi,1); m:= Length(s[2]);

   res:= u^((2*m+1)*m*(m-2)/3) * v^(m*(m-1)/2);

   for i in s[1] do
      for j in [0..i-1] do
         if j in s[1] then res:=res*u^-j;
	              else res:=res*uuu(i-j);
	 fi;
	 if j in s[2] then res:=res*u^-j;
	              else res:=res*(u^(i-j-1)*v+1);
	 fi;
      od;
   od;

   for i in s[2] do
      for j in [0..i-1] do
	 if j in s[2] then res:=res*u^-j;
	              else res:=res*uuu(i-j);
	 fi;
         if j in s[1] then res:=res*u^(1-j)*v^-1;
	              else res:=res*(u^(i-j+1)* v^-1+1);
         fi;
      od;
      if i in s[1] then
         res:= res/u^(i-1);
	 m:=m-1;
      fi;
   od;

   return res*(u+v)^m;
end);

CHEVIE.AddData("HeckeRepresentation","B",function(arg)
  return CHEVIE.RawData("HeckeRepresentation","imp")(2,1,arg[1],arg[2],arg[4]);
end);

CHEVIE.AddData("Representation","B",function(n,i)
  return CHEVIE.RawData("Representation","imp")(2,1,n,i);
end);

#############################################################################
#F  FakeDegree( n, <c>,q ) Fake Degree of char. with charparam <c>
##
CHEVIE.AddData("FakeDegree","B",function(n,c,q)
  return Value(CycPolFakeDegreeSymbol(SymbolPartitionTuple(c,1)),q);
end);

## some cleanup for "B" <-> "C" ? XXX
CHEVIE.AddData("UnipotentCharacters","B",function(arg)
  local uc,symbols,r,d,s,rank;
  rank:=arg[1];
  uc:=rec(harishChandra:=[],charSymbols:=[]);
  for d in 1+2*[0..QuoInt(-1+RootInt(1+4*rank,2),2)] do
    s:=(d^2-1)/4;
    s:=rec(relativeType:=rec(series:="B",indices:=[1+s..rank],rank:=rank-s),
	   levi:=[1..s],
	   eigenvalue:=(-1)^QuoInt(d+1,4),
	   parameterExponents:=Concatenation([d],[2+s..rank]*0+1),
	   cuspidalName:=SPrint("B",s),
	   TeXCuspidalName:=SPrint("B_",TeXBracket(s)));
    Add(uc.harishChandra,s);
    symbols:=Symbols(rank,d);
    s.charNumbers:=[1..Length(symbols)]+Length(uc.charSymbols);
    FixRelativeType(s);
    Append(uc.charSymbols,symbols);
  od;
  uc.harishChandra[1].cuspidalName:="";
  uc.harishChandra[1].TeXCuspidalName:="";
  uc.a:=List(uc.charSymbols,LowestPowerGenericDegreeSymbol);
  uc.A:=List(uc.charSymbols,HighestPowerGenericDegreeSymbol);
  uc.families:=FamiliesClassical(uc.charSymbols);
  if Length(arg)=2 and arg[2]=1 then
    uc.harishChandra[1].relativeType.cartanType:=1;
  fi;
  return uc;
end);
