if not IsBound(UnipotentCharactersOps) then ReadChv("unip/uc");fi;
UnipCharOps:=OperationsRecord("UnipCharOps");

CHEVIE.PrintUniChars:=rec(long:=1);

UnipCharOps.Format:=function(r,option)local m,c,i,s,n,res;
  res:="";
  s:=UnipotentCharacters(r.group).charNames;
  m:=Maximum(List(s,Length))+3;
  for i in [1..Length(r.v)] do
    n:=ConcatenationString("<",s[i],">");
    c:=Format(r.v[i]);
    if IsBound(option.short) then
      if c<>"0" then
	if c="1" then Append(res,"+");
	elif c="-1" then Append(res,"-");
	else  if '+' in c{[2..Length(c)]} or '-' in c{[2..Length(c)]} then
		  c:=Concatenation("(",c,")");
	      fi;
	      if not c[1] in "+-" then Append(res,"+");fi;
	      Append(res,c);
	fi;
	Append(res,n);
      fi;
    else PrintToString(res,"\n",n,String("",m-Length(n)),c);
    fi;
  od;
  if Length(res)=0 then res:="0";fi;
  if res[1]='+' then res:=res{[2..Length(res)]};fi;
  if IsBound(r.name) then 
    res:=SPrint("DLvar[",ReflectionName(r.group),",",r.name,"]=",res);
  else
    res:=SPrint("[",ReflectionName(r.group),"]=",res);
  fi;
  return String(res);
end;

UnipCharOps.String:=r->Format(r,CHEVIE.PrintUniChars);

UnipCharOps.Print:=function(r)Print(String(r));end;

UnipCharOps.\-:=function(a,b)local res;
  res:=ShallowCopy(a);res.v:=a.v-b.v;Unbind(res.name);return res;
end;

UnipCharOps.\+:=function(a,b)local res;
  res:=ShallowCopy(a);res.v:=a.v+b.v;Unbind(res.name);return res;
end;

UnipCharOps.\*:=function(a,b)local res;
  if not IsPolynomial(a) and not IsCyc(a) then 
     Error("a unipotent function can be multiplied only by a polynomial");
  fi;
  res:=ShallowCopy(b);res.v:=a*res.v;Unbind(res.name);return res;
end;

UnipCharOps.Degree:=function(a)
   return a.v*UnipotentDegrees(a.group,X(Cyclotomics));
end;

UnipChar:=function(W,v)local r;
  if IsInt(v) then r:=[1..Length(UnipotentCharacters(W).charNames)]*0;
                   r[v]:=1;v:=r;
	      fi;
  return rec(v:=v,group:=W,operations:=UnipCharOps);
end;

UnipCharOps.ScalarProduct:=function(dummy,a,b)return a.v*b.v;end;
 
# induce unip char u to WF
LusztigInduce:=function(WF,u)local t;
  t:=LusztigInductionTable(u.group,WF);
  if t=false then return false; else return UnipChar(WF,t.scalar*u.v);fi;
end;

# Restrict unip char u to HF
LusztigRestrict:=function(HF,u)
  return UnipChar(HF,u.v*LusztigInductionTable(HF,u.group).scalar);
end;

# induce unip char u to WF
HCInduce:=function(WF,u)
  return UnipChar(WF,HCInductionTable(u.group,WF).scalar*u.v);
end;

# Restrict unip char u to HF
HCRestrict:=function(HF,u)
  return UnipChar(HF,u.v*HCInductionTable(HF,u.group).scalar);
end;

RwTable:=function(W)local uc;
  uc:=UnipotentCharacters(W);
  if not IsBound(uc.rwTable) then
    uc.rwTable:=ComplexConjugate(TransposedMat(CharTable(W).irreducibles))*
      uc.operations.Fourier(uc){uc.almostHarishChandra[1].charNumbers};
  fi;
  return uc.rwTable;
end;
  
NMTable:=function(H)local WF,t,uc;
  if IsBound(H.coxeterCoset) then WF:=CoxeterCoset(H);
  else WF:=Group(H);
  fi;
  t:=CharTable(H).irreducibles;
  t:=t*(1+0*Sum(t,Sum));
  uc:=UnipotentCharacters(WF);
  return ComplexConjugate(TransposedMat(t))*
    uc.operations.Fourier(uc){uc.almostHarishChandra[1].charNumbers};
end;

# usage Rw(braid in a Coxeter group)
#    or Rw(W or WF,<cox. word>)
#    or Rw(W or WF,<perm>)
#    or Rw(W or WF,integer i) (for ith class)
Rw:=function(arg)local ct,uc,W,WF,w,d,i;
  if Length(arg)=1 then # assume arg is a braid
    w:=PermBraid(arg[1]);
    WF:=Group(arg[1]);
  else WF:=arg[1];w:=arg[2];
  fi;
  if not IsSpets(WF) then
    if IsCoxeterGroup(WF) then WF:=CoxeterCoset(WF); else WF:=Spets(WF); fi;
  fi;
  W:=Group(WF);
  if not IsBound(WF.Rw) then
    WF.Rw:=[];
  fi;
  if IsList(w) then w:=EltWord(WF,w);fi;
  if IsPerm(w) then w:=PositionClass(WF,w);fi;
  if not IsBound(WF.Rw[w]) then
    ct:=CharTable(WF);
    uc:=UnipotentCharacters(WF);
    WF.Rw[w]:=ComplexConjugate(ct.irreducibles{[1..Length(ct.irreducibles)]}[w])
      *uc.operations.Fourier(uc){uc.almostHarishChandra[1].charNumbers};
  fi;
  return UnipChar(WF,WF.Rw[w]);
end;

# usage NM(hecke element)
NM:=function(h)local WF,H,W,ct,u,uc,i;
  if IsBound(h.coset) 
  then H:=h.coset; u:=Hecke(H).unit;WF:=CoxeterCoset(H);W:=Group(WF);
  else H:=Hecke(h);u:=H.unit;W:=Group(H);WF:=CoxeterCoset(W);
  fi;
  uc:=UnipotentCharacters(WF);
  h:=Basis(H,"T")(h);
  if not IsBound(H.Rw) then
    ct:=CharTable(H);
    H.Rw:=TransposedMat(120*uc.operations.Fourier(uc)
     {uc.almostHarishChandra[1].charNumbers})*ct.irreducibles/(120*u);
    H.Rw:=TransposedMat(H.Rw);
  fi;
  return UnipChar(WF,HeckeClassPolynomials(h)*H.Rw);
end;
