# JM's library of useful functions

# dump record
dr:=function(r)local p,f;
  f:=RecFields(r);
  Print("rec(");
  for p in f do 
    Print(p,":=");
    if IsFunc(r.(p)) then Print("<<function>>"); 
    else Print(r.(p)); 
    fi;
    if p<>f[Length(f)] then Print(",\n");fi;
  od;
  Print(")\n");
end;

# another way to dump record
dp:=function(r)local p;
  Print("rec(");
  for p in RecFields(r) do
    Print(p,":=",Format(r.(p)),"\n");
  od;
  Print(")\n");
end;

# positions of all occurences of o in l
Positions:=function(l,o)return Filtered([1..Length(l)],j->l[j]=o);end;

TeXs:=function(arg)local name,s;
  s:=arg[1];
  if Length(arg)=1 then name:="tmp.tex";
  else name:=Concatenation(arg[2],".tex");fi;
  PrintTo(name,"\\tabskip=0.1cm\n");
  AppendTo(name,"\\input amssym.def\n\\input amssym.tex\n");
  AppendTo(name,"\\def\\BZ{{\\Bbb Z}}\n");
  Cut(s,78,"_{}+-=&",name);
  AppendTo(name,"\\bye\n");
  Exec(Concatenation("tex ",name));
  Exec(Concatenation("xdvi -expert -s 5 ",Replace(name,".tex",".dvi"),"&"));
end;

TeX:=function(arg)local name,s,o;
  o:=rec(TeX:=1);
  if Length(arg)>1 and IsRec(arg[2]) then
    Inherit(o,arg[2]);arg:=Concatenation([arg[1]],arg{[3..Length(arg)]});
  fi;
  arg[1]:=Format(arg[1],o);
  ApplyFunc(TeXs,arg);
end;

LaTeXs:=function(arg)local name,s;
  s:=arg[1];
  if Length(arg)=1 then name:="tmp.tex";
  else name:=Concatenation(arg[2],".tex");fi;
  PrintTo(name,"\\documentclass{article}\n");
  AppendTo(name,"\\usepackage{amssymb}\\newcommand{\\BZ}{{\\mathbb Z}}\n");
  AppendTo(name,"\\usepackage{array}\\newcolumntype{R}{>{$}r<{$}}\n");
  AppendTo(name,"\\usepackage{supertabular}\n");
  AppendTo(name,"\\begin{document}\n");
  Cut(s,78,"_}+-=&",name);
  AppendTo(name,"\\end{document}\n");
  Exec(Concatenation("latex ",name));
  Exec(Concatenation("xdvi -expert -s 5 ",Replace(name,".tex",".dvi"),"&"));
end;

LaTeX:=function(arg)local name,s,o;
  o:=rec(TeX:=1,LaTeX:=1);
  if Length(arg)>1 and IsRec(arg[2]) then
    Inherit(o,arg[2]);arg:=Concatenation([arg[1]],arg{[3..Length(arg)]});
  fi;
  arg[1]:=Format(arg[1],o);
  ApplyFunc(LaTeXs,arg);
end;

#IsZero:=x->x=0*x;

# order of the center of the complex reflection group W
OrderCenter:=function(W)local d;
  d:=Filtered(ReflectionDegrees(W),x->x<>1);
  if Length(d)=0 then return 1;
  elif Length(d)=1 then return d[1];
  else return ApplyFunc(Gcd,d);
  fi;
end;

# elements of Cartesian([l1,l2,...,ln]) with all coordinates distinct
DistinctCartesian:=function(l)
  if Length(l)=1 then return List(l[1],x->[x]);fi;
  return Concatenation(List(DistinctCartesian(l{[2..Length(l)]}),
    r->List(Difference(l[1],r),e->Concatenation([e],r))));
end;

# next best thing to be able to invert m
CoFactors:=function(m)local v;
  if Length(m)=1 then return [[1]];fi;
  v:=[1..Length(m)];
  return TransposedMat(List(v,i->List(v,j->(-1)^(i+j)*
    DeterminantMat(m{Filtered(v,x->x<>i)}{Filtered(v,x->x<>j)}))));
end;

myDet:=function(m)local n,v,i,j,k;
  n:=Length(m);
  if Length(m)=1 then return m[1][1];
  elif Length(m)=2 then return m[1][1]*m[2][2]-m[1][2]*m[2][1];
  fi;
  for j in [1..n] do
    i:=PositionProperty(m{[1..n]}[j],x->Length(x.elm)=1);
    if i<>false then Print(Length(m),":",[j,i],"\n");fi;
    if i<>false then 
      m:=ShallowCopy(m);
      v:=m[i][j];
      for k in Filtered([1..n],j->j<>i) do
	m[k]:=m[k]-m[k][j]*v^-1*m[i];
      od;
      return (-1)^(i+j)*m[i][j]*myDet(
        m{Concatenation([1..i-1],[i+1..n])}{Concatenation([1..j-1],[j+1..n])});
    fi;
  od;
  return Sum([1..n],function(i)
    if Length(m)>=5 then Print(Length(m),":",i," \c");fi;
    if m[i][1]=0*m[i][1] then return 0;
    else return (-1)^(i-1)*m[i][1]*
      myDet(m{Concatenation([1..i-1],[i+1..n])}{[2..n]});
    fi;end);
end;

# next best thing to be able to invert m
myCoF:=function(m)local v,j,res,i;
  if Length(m)=1 then return [[1]];fi;
  v:=[1..Length(m)];
  res:=m*0;
  for i in v do for j in v do Print("i=",i," j=",j,"\n");
    res[i][j]:=(-1)^(i+j)*myDet(m{Filtered(v,x->x<>i)}{Filtered(v,x->x<>j)});
  od;od;
  return TransposedMat(res);
end;

# decompose a character in term of irreducibles
DecomposeChar:=function(W,char)local ct;
  ct:=CharTable(W);
  return List(ct.irreducibles,x->ScalarProduct(ct,x,char));
end;

# permutation des chars. de W donne par \chi->\chi\otimes\epsilon
SgnPerm:=function(W)local ct,sgn;
  ct:=CharTable(W);
  sgn:=PositionSgn(W);
  return List(ct.irreducibles,x->Position(ct.irreducibles,
               List([1..Length(x)],i->x[i]*ct.irreducibles[sgn][i])));
end;

# compare two objects recursively
# additional arguments specify names to give in diagnostic messages
# cmpobj(a,b[,<name of a>,<name of b> [,suffix,history]][, recursion level])
#
# return status: 0=unknown 1=equal 2=difer
cmpobj:=function(arg)
  local a,b,d,fa,fb,na,nb,l,pr,differ,compared,res,suff,but,history;
  a:=arg[1];b:=arg[2];arg:=arg{[3..Length(arg)]};
  pr:=function(arg)if l>=0 then ApplyFunc(Print,List(arg,Format));fi;end;
  if Length(arg)>1 then na:=arg[1];nb:=arg[2];arg:=arg{[3..Length(arg)]};
  else na:="a";nb:="b";suff:="";fi;
  if Length(arg)>1 then suff:=arg[1];history:=arg[2];arg:=arg{[3..Length(arg)]};
  else suff:="";history:=[[],[]];
  fi;
  if Length(arg)>0 then l:=arg[1];else l:=20;fi;
  if l<-4 then return 0;fi;
  but:=function(a,b) return SPrint(" for ",a," but not ",b,"\n");end;
# if a=b then return;fi;
  fa:=PositionProperty(history[1],x->IsIdentical(x,a));
  fb:=PositionProperty(history[2],x->IsIdentical(x,b));
  if fa=fb and fa<>false then pr(suff," already seen\n");return 1;fi;
  history:=[Concatenation(history[1],[a]),Concatenation(history[2],[b])];
  if IsRec(a) then
    if not IsRec(b) then pr(suff," is a record",but(na,nb));return 2;fi;
    differ:=0;compared:=true;
    fa:=RecFields(a);fb:=RecFields(b);
    d:=Difference(fa,fb);
    if d<>[] then pr(d," are fields",suff,but(na,nb));differ:=differ+1;fi;
    d:=Difference(fb,fa);
    if d<>[] then pr(d," are fields",suff,but(nb,na));differ:=differ+1;fi;
    for d in Intersection(fa,fb) do 
      if d in ["operations","parent", "group","domain", "baseRing",
        "classInvariants","stabChain","groundRing", "relativeGroups",
	"transversal","Spetss", "SubCosets","ReflectionSubgroups"]
      then ;# pr("will not compare ",na,".",d," and ",nb,".",d,"\n");
      else
	res:=cmpobj(a.(d),b.(d),na,nb,SPrint(suff,".",d),history,l-1);
	if res=0 then pr(suff,".",d," not compared\n");compared:=false;
	elif res=2 then 
	  if l=0 then Print(suff,".",d," differs\n");fi;differ:=differ+1;
	fi;
      fi;
    od;
    if differ>0 then return 2;elif compared then return 1;else return 0;fi;
  elif IsString(a) then
    if not IsString(b) then pr(suff," is a string",but(na,nb));return 2;
    elif a<>b then pr(suff,"=",a," in ",na," but =",b," in ",nb,"\n");return 2;
    else return 1;
    fi;
  elif IsList(a) then
    if not IsList(b) then pr(suff," is a list",but(na,nb));return 2;fi;
    if Length(a)<>Length(b) then pr(suff," is of length ",Length(a),
		     " in ",na," but ",Length(b)," in ",nb,"\n");return 2;fi;
    differ:=0;compared:=0;
    for d in  [1..Length(a)] do 
      if IsBound(a[d]) then
	if IsBound(b[d]) then 
	  res:=cmpobj(a[d],b[d],na,nb,SPrint(suff,"[",d,"]"),history,l-1);
	  if res=0 then pr(suff,"[",d,"] not compared\n");compared:=compared+1;
            if compared>5 then 
	      pr("...\n");if differ>0 then return 2;else return 0;fi;
	    fi;
	  elif res=2 then 
	    if l=0 then Print(suff,"[",d,"] differs\n");fi;
	    differ:=differ+1;
	    if differ>5 then pr("...\n");return 2;fi;
	  fi;
	else pr(suff,"[",d,"] is bound",but(na,nb));differ:=differ+1;
	fi;
      elif IsBound(b[d]) then
        pr(suff,"[",d,"] is bound",but(nb,na));differ:=differ+1;
      fi;
    od;
    if differ>0 then return 2;elif compared=0 then return 1;else return 0;fi;
  elif IsFunc(a) then 
    if not IsFunc(b) then pr(suff," is a function",but(na,nb));return 2;
    elif a<>b then pr("Function ", suff," differs\n");return 2;
    else return 1;
    fi;
  elif a<>b then pr(suff,"=",a," in ",na," but =",b," in ",nb,"\n");return 2;
  else return 1;
  fi;
end;

# compare two chevie tables
cmptables:=function(arg)local f,ff,msg,t1,t2,p;
  t1:=Copy(arg[1]);t2:=Copy(arg[2]);
  if Length(arg)<3 or arg[3]=1 then
    f:=Filtered([1..Length(t1.scalar)],i->t1.scalar[i]<>t2.scalar[i]);
  else f:=Filtered([1..Length(t1.scalar)],i->ForAny(t1.scalar[i],x->x<>0*x)
    or ForAny(t2.scalar[i],x->x<>0*x));
  fi;
  msg:=SPrint("[",Length(f),"/",Length(t1.scalar),",");
  if Length(f)=0 then Print("Tables agree!\n");return;fi;
  p:=PermListList(t1.scalar{f},t2.scalar{f});
  if p<>false then
    Print("Permuted lines:\n",Join(List(TransposedMat(
     [t1.gNames(t1,rec()){f},Permuted(t2.gNames(t2,rec()){f},p)]),
      x->Join(x,"->")),"\n"),"\n");
    return;
  fi;
  p:=PermListList(TransposedMat(t1.scalar),TransposedMat(t2.scalar));
  if p<>false then
    Print("Permuted columns:\n",Join(List(TransposedMat(
     [t1.uNames(t1,rec()),Permuted(t2.uNames(t2,rec()),p)]),
      x->Join(x,"->")),"\n"),"\n");
    return;
  fi;
  f:=[1..Length(f)];
  if Length(arg)<3 or arg[3]=2 then
    ff:=Filtered([1..Length(t1.scalar[1])],i->t1.scalar{f}[i]<>t2.scalar{f}[i]);
  else
    ff:=[1..Length(t1.scalar[1])];
  fi;
  PrintToString(msg,Length(ff),"/",Length(t1.scalar[1]),"] of ");
  Print(msg);Display(t1,rec(charGroup:=f,charSubgroup:=ff));
  Print(msg);Display(t2,rec(charGroup:=f,charSubgroup:=ff));
end;

2A:=n->CoxeterCoset(CoxeterGroup("A",n),Product([1..QuoInt(n,2)],i->(i,n+1-i)));
2D:=n->CoxeterCoset(CoxeterGroup("D",n),(1,2));
2E6:=CoxeterCoset(CoxeterGroup("E",6),(1,6)(3,5));
3D4:=CoxeterCoset(CoxeterGroup("D",4),(1,2,4));
2B2:=CoxeterCoset(CoxeterGroup("Bsym",2),(1,2));
2G2:=CoxeterCoset(CoxeterGroup("Gsym",2),(1,2));
2F4:=CoxeterCoset(CoxeterGroup("Fsym",4),(1,4)(2,3));
2I:=n->CoxeterCoset(CoxeterGroup("Isym",2,n),(1,2));
3G333:=Spets(ComplexReflectionGroup(3,3,3),(3,46,35)(6,42,39)(9,54,18)
 (12,45,31)(13,27,20)(14,15,32)(16,17,36)(19,26,41)(23,53,30)(28,29,51)
 (33,34,40)(37,38,43));
3pG333:=Spets(ComplexReflectionGroup(3,3,3),(3,35,46)(6,39,42)(9,18,54)
 (12,31,45)(13,20,27)(14,32,15)(16,36,17)(19,41,26)(23,30,53)(28,51,29)
 (33,40,34)(37,43,38));
4G333:=Spets(ComplexReflectionGroup(3,3,3),(2,12,22,38)(3,35,14,32)(5,26,8,33)
 (6,39,13,20)(7,42,24,27)(9,51,29,54)(10,15,21,46)(16,50,30,44)(17,53,23,36)
 (18,47,28,49)(19,41,40,34)(31,45,37,43));

ProperStableParabolics:=function(WF)local W,l;
  if IsSpets(WF) then W:=Group(WF);else W:=WF;fi;
  l:=Combinations(W.rootInclusion{W.generatingReflections});
  if IsSpets(WF) then l:=Filtered(l,x->OnSets(x,WF.phi)=x);fi;
  SortParallel(List(l,Length),l);
  return l{[1..Length(l)-1]};
end;

ProperParabolics:=function(WF)local W,l;
  if IsSpets(WF) then W:=Group(WF);else W:=WF;fi;
  l:=Combinations(W.rootInclusion{W.generatingReflections});
  SortParallel(List(l,Length),l);
  return l{[1..Length(l)-1]};
end;

# fonction de mesure incrementale du temps
CHEVIE.time:=0;

Stime:=function()local oldtime;
  if not IsBound(CHEVIE.time) then CHEVIE.time:=0;fi;
  oldtime:=CHEVIE.time;
  CHEVIE.time:=Runtime();
  oldtime:=CHEVIE.time-oldtime;
  return SPrint(QuoInt(oldtime,1000),".",QuoInt(oldtime,100),"sec");
end;

braidmatrix:=function(l)
  pmat(List(l,i->List(l,j->BraidRelation(i,j))));
end;
  
testreflrep:=function(W)local m,r,G,p,p1;
  for m in [1..Length(W.matgens)] do AsReflection(W.matgens[m],W.roots[m]);od;
  p:=ChevieCharInfo(W).positionRefl;
  Print("position reflection character=",p,"\n");
  m:=Representations(W,p);
  p1:=Position(CharTable(W).irreducibles,CharRepresentationWords(m,ChevieClassInfo(W).classtext));
  if p1<>p then Error("does not agree with reflection rep.\n");fi;
  Print("matrices:\n");
  for p in m do pmat(p);Print("-------------------\n");od;
  r:=List(m,AsReflection);
  G:=PermRootGroup(List(r,x->x[1]),List(r,x->x[2]));
  braidmatrix(G.matgens);
  Print("ReflectionType=",ReflectionType(G),"\n");
  return G;
end;

mval:=function(l,v) 
  if IsList(l) then return List(l,x->mval(x,v));
  elif IsUnknown(l) then return l;
  else return Value(l,v);fi;
end;
