# utility functions for the rest of the file
CHECK:=rec(error:=false,log:=false,title:="no title",relativeSeries:=true);
Diagnose:=function(arg)
  CHECK.diagnosed:=true;
  if CHECK.log then
    ApplyFunc(AppendTo,Concatenation([CHECK.logfile,"[",CHECK.title,"]"],arg));
  elif CHECK.error then
    ApplyFunc(Error,Concatenation(["[",CHECK.title,"]"],arg)); 
  fi;
  Print(ApplyFunc(SPrint,Concatenation(["****** WARNING!: [",CHECK.title,"]"],
          arg)));
end;

CHEVIE.Diagnose:=Diagnose;

Ok:=function(arg)
  if CHECK.diagnosed then return;fi;
  ApplyFunc(Print,Concatenation(["[",CHECK.title,"] "],arg,[" OK!\n"]));
end;

Title:=function(arg) 
  CHECK.diagnosed:=false;
  CHECK.title:=ApplyFunc(SPrint,arg);
end;

InitCheck:=function(W)local msg;
  msg:=SPrint("======== diagnostics for ",ReflectionName(W)," =========\n");
  Stime();
  if CHECK.log then AppendTo(CHECK.logfile,msg);fi;Print(msg);
end;

EndCheck:=function()local msg;
  msg:=SPrint(Stime(),"\n");
  if CHECK.log then AppendTo(CHECK.logfile,msg);fi;Print(msg);
end;

TestGroups:=[[CoxeterGroup,"A",1],[2A,2],[CoxeterGroup,"A",2],
  [CoxeterGroup,"I",2,5],[CoxeterGroup,"B",2],[CoxeterGroup,"G",2],
  [CoxeterGroup,"A",3],[2B2],[2G2],[2A,3],[ComplexReflectionGroup,4],[3D4],
  [2A,4],[2F4],[ComplexReflectionGroup,3,1,2],[ComplexReflectionGroup,3,3,3],
  [ComplexReflectionGroup,4,4,3],
  [CoxeterGroup,"B",3],[CoxeterGroup,"C",3],
  [CoxeterGroup,"A",4],[CoxeterGroup,"H",3],[2D,4],[ComplexReflectionGroup,6],
  [2A,5],[ComplexReflectionGroup,8],[ComplexReflectionGroup,24],
  [CoxeterGroup,"D",4],[CoxeterGroup,"A",5],
  [ComplexReflectionGroup,3,3,4],
  [CoxeterGroup,"B",4],[CoxeterGroup,"C",4],[CoxeterGroup,"F",4],
  [ComplexReflectionGroup,25],[CoxeterGroup,"D",5],[2D,5],[2A,6],
  [CoxeterGroup,"A",6], [CoxeterGroup,"B",5], [CoxeterGroup,"C",5],
  [ComplexReflectionGroup,26], [ComplexReflectionGroup,29],
  [CoxeterGroup,"E",6],[ComplexReflectionGroup,14], [CoxeterGroup,"H",4],
  [2D,6], [CoxeterGroup,"D",6], [ComplexReflectionGroup,33],
  [2E6], [2A,7],[CoxeterGroup,"A",7],
  [ComplexReflectionGroup,27],[CoxeterGroup,"E",7],[CoxeterGroup,"D",7],
  [CoxeterGroup,"B",6],[CoxeterGroup,"C",6],
  [ComplexReflectionGroup,32],[CoxeterGroup,"B",7],[CoxeterGroup,"C",7],
  [ComplexReflectionGroup,34],
  [CoxeterGroup,"E",8],[2D,7]];

TestGroup:=function(i)local t;
  t:=TestGroups[i];
  if Length(t)>1 then return(ApplyFunc(t[1],t{[2..Length(t)]}));
  else return(t[1]);
  fi;
end;
############################################################################

Tensor:=function(arg)local W,t,ct;
# decompose tensor product arg{[2..Length(arg)]} (given as indices in chars)
  W:=arg[1];ct:=CharTable(W).irreducibles;
  t:=TransposedMat(ct{arg{[2..Length(arg)]}});
  t:=List(t,Product);
  return Decomposition(ct,[t],"nonnegative")[1]; 
end;

IsClassical:=function(W)local n;n:=ReflectionType(W)[1];
  return (IsBound(n.series) and (n.series in ["A","B","C","D","G","I",] or
    (n.series="ST" and IsBound(n.p)))) or (IsBound(n.orbit) and 
    n.orbit[1].series in ["A","B","C","D","I",] and OrderPerm(n.twist)<>3);
end;

# cmpvec(a,b [,fun])
# compares vectors a and b calling func on discrepancy
cmpvec:=function(arg)local a,b,na,j,t,pa,pb;
  a:=arg[1];b:=arg[2];
  if a=b then return;fi;
  if -a=b then Diagnose("global sign error\n");fi;
  if Length(arg)>=3 then na:=arg[3];else na:=j->SPrint("a[",j,"]");fi;
  pa:=SortingPerm(a);pb:=SortingPerm(b);
  if Permuted(a,pa)=Permuted(b,pb) then
    Diagnose("order error: should be",List(a,function(x)local v;
         v:=Filtered([1..Length(b)],i->b[i]=x);
	 if Length(v)=1 then return v[1];else return v;fi;end),"\n");
  elif Permuted(a,pa)=-Permuted(b,pb) then
    Diagnose("order and global sign error: should be",List(a,function(x)local v;
         v:=Filtered([1..Length(b)],i->b[i]=-x);
	 if Length(v)=1 then return v[1];else return v;fi;end),"\n");
  else 
  for j in [1..Length(a)] do
    if a[j]<>b[j] then
      if a[j]=-b[j] then Diagnose(na(j)," sign error\n");
      else   t:=Filtered([1..Length(b)],k->b[k]=a[j]);
        if Length(t)>0 then Diagnose(na(j)," found at ",t,"\n");
        else t:=Filtered([1..Length(b)],k->b[k]=-a[j]);
          if Length(t)>0 then Diagnose(na(j)," found at -",t,"\n");
          else Diagnose(na(j)," not found\n");
          fi;
        fi;
      fi;
    fi;
  od;
  fi;
end;

cmpvcyc:=function(a,b,na,nb)local q;
 if a=b then return;fi;
 if b=0*b then 
   if a<>0*a then Diagnose(na,"=",a," but ",nb,"=",b,"\n");fi;
   return;
 fi;
 q:=a/b;
 if Length(q.vcyc)=0 then
      Diagnose(na,"=",q,"*",nb," where ",nb,"=",b,"\n");
 else Diagnose(na,"=",a," but ",nb,"=",b,"\n");
 fi;
end;

# Computes an Hecke algebra with parameters depending on type:
# type= 0: [a,b,c] 
#       1: [1,b,c] 
#       2: [1,q,q^2] (if several Hplanes take Lcm)
#       3: [q,E3,E3^2] (Spets)
#       4: [1,E3,E3^2]
GenericHecke:=function(W,type)local W,r,o,vars,v,i,j,q;
 vars:="xyztuvwabcdefghijklmnopqrs";
 r:=Set(W.orbitRepresentative{W.generatingReflections});
 o:=List(r,i->OrderPerm(W.(i)));
 v:=1;
 r:=[];
 for i in [1..Length(o)] do
   r[i]:=[];
   for j in [0..o[i]-1] do
     if type=0 then Add(r[i],Mvp([vars[v]])); v:=v+1;
     elif type=1 then
       if j=0 then Add(r[i],Mvp("q")^0);
       else Add(r[i],Mvp([vars[v]])); v:=v+1;
       fi;
     elif type=2 then Add(r[i],Mvp("q")^(j*Lcm(o)/o[i]));
     elif type=3 then
       if j=0 then Add(r[i],Mvp("q"));
       else Add(r[i],E(o[i])^j);
       fi;
     elif type=4 then Add(r[i],E(o[i])^j);
     fi;
   od;
 od;
 o:=List(W.generatingReflections,i->r[W.orbitRepresentative[i]]);
 return Hecke(W,o);
end;

###############################################################################
#  Various tests of integrity of the data                                     #
###############################################################################

# test if the powermap of CharTable(W) is OK
CheckPowerMap:=function(W)local cl,ct,p,i;
  cl:=List(ConjugacyClasses(W),Representative);
  ct:=CharTable(W);
  p:=ct.powermap;
  Title("CheckPowerMap");
  for i in [1..Length(p)] do
   if IsBound(p[i]) and List(cl,x->PositionClass(W,x^i))<>p[i] then
     Diagnose(i,"-th map is ",p[i],"\n",
                "should be:",List(cl,x->PositionClass(W,x^i)),"\n");
   fi;
  od;
  Ok();
end;

# test if positionRefl is correct
CheckPositionRefl:=function(W)local ct,ci,p;
  ct:=CharTable(W);
  p:=Position(ct.irreducibles,ReflectionCharacter(W));
  Title("CheckPositionRefl");
  ci:=ChevieCharInfo(W);
  if not IsBound(ci.positionRefl) then
    Diagnose("not bound but should be ",p,":",CharNames(W)[p],"\n");
  elif p<>ci.positionRefl then 
    Diagnose("is ",ci.positionRefl," should be ",p,":",CharNames(W)[p],"\n");
  else Ok(" refl=",CharParams(W)[ci.positionRefl]);
  fi;
end;

# test if positionSgn is correct
CheckPositionSgn:=function(W)local ct,ci,p;
  if ForAny(W.EigenvaluesGeneratingReflections,x->x<>1/2) then return;fi;
  ct:=CharTable(W);
  p:=Position(ct.irreducibles, 
        List(ChevieClassInfo(W).classtext,c->(-1)^Length(c)));
  Title("CheckPositionSgn");
  if p=false then return;fi;
  ci:=ChevieCharInfo(W);
  if not IsBound(ci.positionSgn) then
    Diagnose("not bound but should be ",p,":",CharNames(W)[p],"\n");
  elif p<>ci.positionSgn then 
    Diagnose("is ",ci.positionSgn," should be ",p,":",CharNames(W)[p],"\n");
  else Ok(" sgn=",CharParams(W)[ci.positionSgn]);
  fi;
end;

# test if fakedegrees agree with the group-theoretic ones
CheckFakeDegrees:=function(W)local fd,ffd,q;
  q:=X(Cyclotomics);
  fd:=FakeDegrees(W,q);
  if IsSpets(W) then ffd:=SpetsOps.FakeDegrees(W,q);
  else ffd:=PermRootOps.FakeDegrees(W,q);
  fi;
  Title("CheckFakeDegrees");
  if fd<>ffd then Diagnose(" should be:",fd,"\n but are:",ffd,"\n",
                List(ffd,x->Position(fd,x)),"\n");
  fi;
  if IsSpets(W) then Ok("agree with SpetsOps.FakeDegrees");
  else  Ok("agree with PermRootOps.FakeDegrees");
  fi;
end;

# For a Group generated by true reflections,
# check that unip.deg(rho)=q^N unip.deg(rho tensor sign)(q^-1)
CheckSgnUd:=function(W)local q,ud,p,i,N;
  if IsSpets(W) then return;fi; # should we?
  if PositionSgn(W)=false then return;fi; # for sure
  q:=X(Cyclotomics);q.name:="q";
  ud:=UnipotentDegrees(W,q);
  p:=SgnPerm(W);
  Title("CheckSgnUd:gendeg(chi)=q^N gendeg(chi.eps)(q^-1)");
  N:=Sum(ReflectionDegrees(W)-1);
  if not ForAll([1..Length(p)],i->q^N=ud[i]/Value(ud[p[i]],q^-1)) then
    Diagnose("\n");
  fi;
  Ok();
end;

# CheckDegreesFromHC(W,series # [,guess])
# Checks that unipotent degrees from that series agree with those from
# relative Hecke algebra
CheckDegreesFromHC:=function(arg) local W,i,guess,uw,hw,L,index,cusp,t,R,q,l,ud,
  o,aA,f,reldeg,sch,den,sgnpos;
  W:=arg[1];i:=arg[2];if Length(arg)=3 then guess:=arg[3];else guess:=false;fi;
  uw:=UnipotentCharacters(W); hw:=uw.harishChandra[i];
  L:=ReflectionSubgroup(W,hw.levi);#ReflectionType(L);
  q:=X(Cyclotomics);
  index:=CycPol(GenericOrder(W,q)/GenericOrder(L,q));index.valuation:=0;
  cusp:=CycPolUnipotentDegrees(L)[FindCuspidalInLevi(hw,L,W)];
  Title("CheckDegreesFromHC_",ReflectionName(L),"(cusp=",hw.cuspidalName,":",
     cusp,")");
  if Length(hw.relativeType)=0 then R:=CoxeterGroup();
  else t:=hw.relativeType[1];t.operations:=ReflTypeOps;R:=ReflectionGroup(t);
  fi;
# check parameters of relative algebra by the formula
# u_{s,j}=\zeta_s^j  q^(([a+A]_\rho_{det_s^j}-[a+A]_\rho_{\Id})/e_s#class(s)
# where \rho_\chi is the unip. char corresp. to \chi in relative group
  f:=l->Maximum(l)-Concatenation([0],l);
  aA:=List(hw.charNumbers,x->uw.a[x]+uw.A[x]);
  Print("a+A-Inf(a+A)=",FormatGAP(aA-Minimum(aA)));
  for o in HyperplaneOrbits(R) do Print(" es=",o.e_s," esNs=",o.e_s*o.N_s);od;
  Print(" [G:L]=",index,"\n");
  Print("Parameter: computed:",FormatGAP(
    List(HyperplaneOrbits(R), o->f(List(o.det_s,i->(aA[i]-aA[
      ChevieCharInfo(R).positionId])/o.e_s/o.N_s)))));
  Print(" stored:",FormatGAP(hw.parameterExponents),"\n");
  l:=hw.parameterExponents;
  if Length(l)=0 then den:=1;
  else den:=Lcm(List(Flat(l),Denominator));
  fi;
  if IsList(l) then
    l:=List(l,function(p)if IsList(p) then return List([1..Length(p)],
	     function(i)local e;e:=E(Length(p))^(i-1);
	       if p[i]<0 then return -e*q^(den*-p[i]); 
	       else return e*q^(den*p[i]);fi;end);
	     elif p<0 then return -q^(den*-p); else return q^(den*p);
	     fi;
	   end);
  else l:=q^l;
  fi;
  ud:=CycPolUnipotentDegrees(W);
  reldeg:=List(ud{hw.charNumbers},x->Value(cusp*index/x,q^den));
  R:=Hecke(R,l);Print("  Relative=",R,"\n");
  sch:=SchurElements(R);
  if ForAny(sch,x->x=false) then
    Diagnose("No Schur elements implemented!\n");
    return false;
  fi;
  sch:=sch*q^0;
  if sch<>reldeg then
  Print("Comparing generic degrees of rel.algebra to actual degrees\n");
  cmpvec(sch,reldeg,j->SPrint(j,"=ud[",hw.charNumbers[j],"]"));
  fi;
  Ok();
  if guess then  return reldeg;
  else  return List(sch,x->Value(cusp*index,q^den)/x);
  fi;
end;

CheckAllDegreesFromHC:=function(W)local h,i;
 h:=UnipotentCharacters(W).harishChandra;
 for i in [1..Length(h)] do
   if h[i].relativeType<>[] then CheckDegreesFromHC(W,i);fi;
 od;
end;

# FindRepresentation(H,r [,true])
# find which char corresponds; false if none; check all char values if 3rd arg
FindRepresentation:=function(arg)local H,gr,check,W,i,ct,pos,H,O,l,iszero,r,t;
  H:=arg[1];gr:=arg[2]; iszero:=x->x=0*x;
  check:=Length(arg)>2;
  if IsHeckeAlgebra(H) then W:=Group(H);O:=H;
  else O:=H;W:=O;H:=Hecke(H);fi;
  t:=ChevieClassInfo(W).classtext;
  ct:=CharTable(O).irreducibles;
  l:=[1..Length(t)];SortParallel(l,List(t,Length));t:=t{l};pos:=[1..Length(l)];
  for i in [1..Length(t)] do
    r:=CharRepresentationWords(gr,[t[i]])[1];
    if IsHeckeAlgebra(O) then r:=r*H.unit;fi;
    pos:=Filtered(pos,j->iszero(ct[j][l[i]]-r));
    if not check and Length(pos)=1 then return pos[1];
    elif Length(pos)=0 then return false;
    fi;
  od;
  if Length(pos)=1 then return pos[1];
  elif Length(pos)=0 then return false;
  else Error("two characters are equal");
  fi;
end;

# check that the representations of the group or hecke algebra W
# satisfy the braid and quadratic relations, and agree with chartable
# CheckRepresentations(H [,l])
# check representations in list l or of dim. l (default: all)
CheckRepresentations:=function(arg)local W,i,gr,r,ct,pos,H,O,l;
  W:=arg[1];
  if IsHeckeAlgebra(W) then H:=W;W:=Group(H);O:=H;
  else H:=Hecke(W);O:=W;fi;
  Title("CheckRepresentations(",O,")");
  ct:=CharTable(O).irreducibles;
  if Length(arg)=1 then l:=[1..NrConjugacyClasses(W)];
  elif IsInt(arg[2]) then l:=Filtered([1..NrConjugacyClasses(W)],
    i->ct[i][1]=arg[2]);
  else l:=arg[2];
  fi;
  for i in l do
    Print("Representation #",i);
    gr:=Representations(O,i);
    if gr=false then Print("=false\n");
    else
      Print(" dim:",Length(gr[1]),"...\c");
      CheckHeckeDefiningRelations(H,gr);
      pos:=FindRepresentation(O,gr,true);
      if pos=i then Print("\n");
      elif pos<>false then Diagnose("character found at ",pos,"\n");
      else Diagnose("character does not match\n");
	pos:=TransposedMat([[1..Length(ct[i])],ct[i],
	  CharRepresentationWords(gr,ChevieClassInfo(W).classtext)]);
	pos:=Filtered(pos,x->x[2]<>x[3]);
	pos:=List(pos,x->List(x,FormatGAP));
	Print(StringTable(Concatenation([["no","is","should be"]],pos)));
      fi;
    fi;
  od;
  Ok();
end;

# arg[1]=some Hecke algebra  arg[2]=dimension  of chars to explore
CheckG4_22Hecke:=function(arg)local H,X,Y,Z,rows,n,t,v,u,g,name,ix,para;
  H:=arg[1];
  t:=ReflectionType(H.reflectionGroup);para:=H.parameter;
  if Length(t)>1 or not IsBound(t[1].ST) or not t[1].ST in [4..22] then
    Error("H should be a Hecke algebra of a group in G4..G22");
  fi;
  n:=t[1].ST;X:=[1,-1];Y:=[1,E(3),E(3)^2];
  if n in [4..7] then
    u:=Concatenation(X,Y,Y);g:=7;name:="G4_7";
    if   n=4 then Z:=para[1];
    elif n=5 then Y:=para[1];Z:=para[2];
    elif n=6 then X:=para[1];Z:=para[2];
    elif n=7 then X:=para[1];Y:=para[2];Z:=para[3];
    fi;
  elif n in [8..15] then
    Z:=[1,E(4),-1,E(4)^3];u:=Concatenation(X,Y,Z);g:=11;name:="G8_15";
    if   n=8 then Z:=para[1];
    elif n=9 then X:=para[1];Z:=para[2];
    elif n=10 then Y:=para[1];Z:=para[2];
    elif n=11 then X:=para[1];Y:=para[2];Z:=para[3];
    elif n=12 then X:=para[1];
    elif n=13 then
      if para[1]<>[1,-1] or para[2]<>[1,-1] then
	Error("HeckeCharTable of G13 not implemented");fi;
      X:=para[1];
      Z:=[GetRoot(para[2][1],2),GetRoot(para[2][2],2)];Z:=Concatenation(Z,-Z);
    elif n=14 then X:=para[1];Y:=para[2];
    elif n=15 then X:=para[1];Y:=para[2];
      Z:=[GetRoot(para[3][1],2),GetRoot(para[3][2],2)];Z:=Concatenation(Z,-Z);
    fi;
  elif n in [16..22] then
    Z:=[1,E(5),E(5)^2,E(5)^3,E(5)^4];
    u:=Concatenation(X,Y,Z);g:=19;name:="G16_22";
    if   n=16 then Z:=para[1];
    elif n=17 then X:=para[1];Z:=para[2];
    elif n=18 then Y:=para[1];Z:=para[2];
    elif n=19 then X:=para[1];Y:=para[2];Z:=para[3];
    elif n=20 then Y:=para[1];
    elif n=21 then X:=para[1];Y:=para[2];
    elif n=22 then X:=para[1];
    fi;
  fi;
  # assume non-cyc params are monomials in one Mvp
  t:=TransposedMat([Concatenation(X,Y,Z),u]);
  t:=List(t,function(x)local x,v,p;
    v:=x[2];x:=x[1];
    if not IsMvp(x) or Length(x.elm)<>1 or Length(x.elm[1].elm)<>1 or 
      AsRootOfUnity(x.coeff[1])=false then return false;fi;
    p:=(AsRootOfUnity(v/x.coeff[1])+[0..x.elm[1].coeff[1]-1])/x.elm[1].coeff[1];
    return [x.elm[1].elm[1],List(p,x->E(Denominator(x))^Numerator(x))];
  end);
  t:=Filtered(Set(t),x->x<>false);
  v:=Set(List(t,x->x[1]));
  t:=List(v,x->[x,Intersection(List(Filtered(t,y->y[1]=x),y->y[2]))[1]]);
  Print("specialization=",Join(List(t,x->SPrint(x[1],"->",x[2]))),"\n");
  rows:=List(CharTable(Hecke(ComplexReflectionGroup(g),[X,Y,Z])).irreducibles,
   x->x{ChevieClassInfo(H.reflectionGroup).indexclasses});
  u:=List(rows,x->List(x,function(y)if IsCyc(y) then return y;
                 else return ScalMvp(Value(y,Concatenation(t)));fi;end));
  t:=CHEVIE.(name).CharTable(n).irreducibles;
  Print("how chars of G",g," for these params specialize to chartable\n");
  v:=List(u,x->Position(u,x));
  v:=List(Set(v),y->[y,Filtered([1..Length(v)],i->v[i]=y)]);
  v:=Set(List(v,x->[Position(t,u[x[1]]),x[2]]));
  u:=Maximum(List(v,x->Length(x[2])));
  for ix in v do Append(ix[2],List([1..u-Length(ix[2])],i->""));od;
  ix:=CHEVIE.RawData("CharInfo",name)(n).indexchars;
  v:=List(v,function(x)local ok;ok:="no";
    if x[1]<>false then if ix[x[1]] in x[2] then ok:="";fi;
       return Concatenation([x[1],ix[x[1]],ok],x[2]);
     else return Concatenation([false,false,ok],x[2]);fi;end);
  if Length(arg)=2 then v:=Filtered(v,x->t[x[1]][1]=arg[2]);fi;
  u:=Concatenation([Concatenation(["char no.","chosen","ok"],List([1..u],x->""))],v);
  Print(StringTable(TransposedMat(u)));
  t:=Representations(H);
  u:=List(t,r->FindRepresentation(H,r));
  if u<>[1..Length(u)] then 
    u:=TransposedMat([[1..Length(u)],u]);
    u:=Filtered(u,x->x[1]<>x[2]);
    u:=Concatenation([["rep no.","found at"]],u);
    Print("representations do not match:\n");
    Print(StringTable(TransposedMat(u)));
  fi;
  return v;
end;

# CheckHCDegrees(WF [,series])
# check unipotent degrees in WF agree with those obtained by
# the generic degrees of the relative Hecke algebra
CheckHCDegrees:=function(arg)local WF,sers,uc,ud,uh,cn,j,q,hw,L,Index,cusp,t,R;
  WF:=arg[1];
  if not IsSpets(WF) then WF:=Spets(WF);fi;
  uc:=UnipotentCharacters(WF);
  q:=X(Cyclotomics);
  ud:=UnipotentDegrees(WF,q);
  if Length(arg)=2 then sers:=uc.harishChandra{arg[2]};
  else sers:=uc.harishChandra;
  fi;
  for hw in sers do
    Title("CheckHCDegrees(",Position(uc.harishChandra,hw),")");
    L:=SubSpets(WF,hw.levi);
    Index:=GenericOrder(WF,q)/GenericOrder(L,q);Index.valuation:=0;
    cusp:=UnipotentDegrees(L,q)[FindCuspidalInLevi(hw,L,WF)];
    Print("L=",ReflectionName(L)," cusp:",String(CycPol(cusp))," Index=",
      String(CycPol(Index)),"\n");
    for t in hw.relativeType do t.operations:=ReflTypeOps;od;
    R:=ApplyFunc(ReflectionGroup,hw.relativeType);
    R:=Hecke(R,List(hw.parameterExponents,
     function(i) 
       if IsList(i) then
         return List([1..Length(i)],j->E(Length(i))^(j-1)*q^i[j]);
       elif i<0 then return -q^-i; else return q^i;fi;end));
    Print("Relative Hecke=",R,
          Concatenation(List(hw.relativeType,x->x.indices)),"\n");
    uh:=List(SchurElements(R),x->cusp*(Index)/x);
    cn:=hw.charNumbers;
    cmpvec(ud{cn},uh,j->
         SPrint("ud[",cn[j],"]=",uc.charNames[cn[j]],"=",CycPol(ud[cn[j]])));
    Ok();
  od;
end;

# check unipotent degrees of W are consistent with HC induction from
# those of L_J
CheckGendegsHCInduce:=function(W,J)local L,index,q,pred,ind,nh,inddeg,p,tbl;
  q:=X(Cyclotomics);
  if IsSpets(W) then L:=SubSpets(W,J);
  else L:=ReflectionSubgroup(W,J);
  fi;
  nh:=UnipotentCharacters(L).charNames;
  index:=GenericOrder(W,q)/GenericOrder(L,q);index.valuation:=0;
  if IsSpets(W) then index:=index*GenericSign(L)/GenericSign(W);fi;
  pred:=UnipotentDegrees(L,q)*index;
  tbl:=HCInductionTable(L,W);
  ind:=List(TransposedMat(tbl.scalar),x->UnipChar(W,x));
  inddeg:=List(ind,Degree);
  Title("CheckGendegsHCInduce(",ReflectionName(L),")Relgroups:",
    Join(List(tbl.pieces,x->ReflectionName(x.g))));
  cmpvec(pred,inddeg,j->SPrint("|G:L|*",nh[j],"(#",j,")"));
  Ok();
end;

# check CharTable(W) is correct
CheckCharTable:=function(W)local ct,W1,ct1,f1,f,p,i,cl;
  ct:=CharTable(W);
  Unbind(W.charTable);
  Title("CheckCharTable");
  cl:=List(ConjugacyClasses(W),Representative);
  if IsSpets(W) then ct1:=SpetsOps.CharTable(W);
    cl:=List(ct.irreducibles,x->PositionProperty(ct1.irreducibles,
      y->ProportionalityCoefficient(y,x)<>false));
    if ForAll(cl,x->x<>false) then 
      f:=List([1..Length(cl)],i->ProportionalityCoefficient(
	ct1.irreducibles[cl[i]],ct.irreducibles[i]));
      if ForAny(f,x->x<>1) then
	#Diagnose("coefficients:",f,"\n");
	for i in [1..Length(cl)] do 
	  ct1.irreducibles[cl[i]]:=ct1.irreducibles[cl[i]]/f[i];
	od;
      fi;
    fi;
  else
    W1:=Group(W.generators,());
    ct1:=CharTableSplitClasses(CharTable(W1),List(cl,x->PositionClass(W1,x)));
    cl:=List(ct.irreducibles,x->Position(ct1.irreducibles,x));
  fi;
  if ForAny(cl,x->x=false) or Length(Set(cl))<>Length(cl) then
    Diagnose("irreducibles");
  else  ct1.irreducibles:=ct1.irreducibles{cl};
    if IsBound(ct1.irredinfo) then ct1.irredinfo:=ct1.irredinfo{cl};fi;
    Ok();
  fi;
  cmpobj(ct,ct1,"CharTable","GroupOps.CharTable");
end;

# Check Schur elements satisfy Schur relations (H)
CheckSchurRelations:=function(H)local W,i,q,news,p,ct,pp;
  W:=Group(H);
  news:=SchurElements(H); pp:=news[PositionId(W)]; news:=List(news,x->pp/x);
  ct:=CharTable(H);
  if ct=false then return;fi;
  i:=news*CharTable(H).irreducibles;
  Title("CheckSchurRelations");
  if not ForAll([2..Length(news)],x->i[x]=i[x]*0) then 
    Diagnose("generic*CharTable(H) not ok!!\n");
    i:=news*0;i[1]:=pp;
    p:=SolutionMat(CharTable(H).irreducibles,i);
    for i in [1..Length(news)] do
      if news[i] <> p[i] then
        Print("computed schur[",i,"]<>schur[",i,"]!!\n");
      fi;
    od;
    Error();
  fi;
  Ok("generic*CharTable(H)");
end;

# check Lusztig induction from split levi J agrees with HC induction
CheckSplitLusztig:=function(WF,J)local HF,a,b,c;
  if not IsSpets(WF) then WF:=Spets(WF);fi;
  HF:=SubSpets(WF,J);
  a:=UnipotentCharacters(WF);
  Title("CheckSplitLusztig from ",ReflectionName(HF));
  b:=LusztigInductionTable(HF,WF);
  c:=HCInductionTable(HF,WF);
  if c.scalar<>b.scalar then cmptables(c,b);Diagnose("disagree!\n");fi;
  return [a,b,c];
end;

CheckMackeyWithTori:=function(tbl)local LF,WF,t,f,rhs,c,i;
  WF:=tbl.g;LF:=tbl.u;
  Title("CheckMackeyWithTori of ",tbl.head(tbl,rec()));
  f:=FusionConjugacyClasses(LF,WF);
  c:=Zip(CharTable(WF).centralizers{f},CharTable(LF).centralizers,
         function(a,b)return a/b;end);
  t:=RwTable(LF);
  rhs:=List([1..Length(ConjugacyClasses(WF))],function(i)local j,res;
    res:=t[1]*0;
    for j in [1..Length(f)] do
      if f[j]=i then res:=res+t[j]*c[j];fi;
    od;
    return res;end);
  tbl:=RwTable(WF)*tbl.scalar;
  if tbl<>rhs then 
    tbl:=rec(scalar:=tbl,
        uNames:=function(t,option)
          return CharNames(UnipotentCharacters(LF),option);end,
        gNames:=function(t,option)return ChevieClassInfo(WF).classnames;end,
        head:=function(t,option)return "<RTwG,RLG chi>";end,
        what:="Mackey ",
        u:=LF,g:=WF,
        operations:=InductionTableOps);
    rhs:=rec(scalar:=rhs,
        u:=LF,g:=WF,
        uNames:=function(t,option)
          return CharNames(UnipotentCharacters(LF),option);end,
        gNames:=function(t,option)return ChevieClassInfo(WF).classnames;end,
        head:=function(t,option)return "Sum_w'~w<chi,RTwL>";end,
        what:="Mackey ",
        operations:=InductionTableOps);
    cmptables(tbl,rhs,2);
    Diagnose("fail\n");
  fi;
end;

CheckTransitivity:=function(tbl)local RLGRwL,RwG;
  RLGRwL:=RwTable(tbl.u)*TransposedMat(tbl.scalar);
  RwG:=RwTable(tbl.g){FusionConjugacyClasses(tbl.u,tbl.g)};
  Title("CheckTransitivityFromTori of ",tbl.head(tbl,rec()));
  cmpvec(RLGRwL,RwG,j->SPrint("RLGRwL[",j,"]"));
end;

CheckLusztig:=function(L,WF)local t,uh,hd,ud,index,pred,ind,j,q;
  Title("Check R^{",ReflectionName(WF),"}_{",ReflectionName(L),"}");
  t:=LusztigInductionTable(L,WF);
  if t=false then Diagnose("Failed\n");return;fi;
  t.head:=function(t,option)return SPrint("R_",
	   ReflectionName(t.u,option),"^",ReflectionName(t.g,option));end;
  if L.phi=WF.phi then CheckSplitLusztig(WF,
    Group(L).rootInclusion{Group(L).generatingReflections});fi;
  uh:=UnipotentCharacters(L);
  if uh=false then return;fi;
  q:=X(Cyclotomics);
  uh:=uh.charNames; hd:=UnipotentDegrees(L,q); ud:=UnipotentDegrees(WF,q);
  index:=GenericOrder(WF,q)/GenericOrder(L,q);index.valuation:=0;
  index:=index*GenericSign(L)/GenericSign(WF);
  pred:=hd*index;
  ind:=List(TransposedMat(t.scalar),x->UnipChar(WF,x));
  Title("CheckGendegsInduce with ",t.head(t,rec()));
  for j in [1..Length(hd)] do
    if pred[j]<>Degree(ind[j]) then
      Diagnose("  RLG(",uh[j],")=",Format(ind[j]),"\n  ",
        CycPol(Degree(ind[j]))," instead of ",CycPol(pred[j]),
        " quotient ", CycPol(Degree(ind[j]))/CycPol(pred[j]),"\n");
    fi;
  od;
  CheckMackeyWithTori(t);
  CheckTransitivity(t);
  Title("Check R^{",ReflectionName(WF),"}_{",ReflectionName(L),"}");
  if L.phi=WF.phi then CHECK.title:=Replace(CHECK.title,"Check","Check Split");
  fi;
  Ok();
end;

# check Lusztig induction from all levis
CheckAllLusztig:=function(arg)local WF,l,J,L;
  WF:=arg[1];if not IsSpets(WF) then WF:=Spets(WF);fi;
  if Length(arg)=2 then l:=arg[2];
  else l:=ProperParabolics(WF);
  fi;
  for J in l do 
    Title("Twistings(",Format(WF),",",FormatGAP(J),")");
    for L in Twistings(WF,J) do CheckLusztig(L,WF);od;
  od;
end;

# test that a family satisfies Lusztig's "exotic fourier transform" properties.
# CheckFamily(family record [,try harder]) or CheckFamily(W,famno [,try harder])
CheckFamily:=function(arg)local W,f,inds,T,P,S,Sbar,Id,real,Ep,i,j,v,check,
  TrS,d,hard,ps,no,uc,fd,special,s,p,cospecial;
  if IsGroup(arg[1]) then 
    W:=arg[1];no:=arg[2];uc:=UnipotentCharacters(W);f:=uc.families[no];
    arg:=arg{[3..Length(arg)]};
  else f:=arg[1]; arg:=arg{[2..Length(arg)]};
  fi;
  hard:=Length(arg)>0;
  if Length(f.eigenvalues)=1 then return;fi; # nothing interesting to test
  T:=DiagonalMat(f.eigenvalues);
  S:=f.fourierMat;
  if IsBound(f.deltas) then S:=S*DiagonalMat(f.deltas);fi;
  Sbar:=ComplexConjugate(S);Id:=IdentityMat(Length(S));TrS:=TransposedMat(S);
  v:="CheckFamily";
  if IsBound(W) then PrintToString(v," ",ReflectionName(W),".",no);fi;
  PrintToString(v,"#",Length(f.fourierMat));
  if IsBound(f.name) then PrintToString(v,"(",f.name,")");fi;
  if IsBound(f.perm) then 
    PrintToString(v,"P");
    Ep:=PermListList(ComplexConjugate(f.eigenvalues),f.eigenvalues);
    if Ep<>false then PrintToString(v,"E");
    fi;
  fi;
  Title(v);
  if hard then Print("[",v,"] ...\n| ");fi;
  check:=function(a,b,msg)
    if a<>b then Print("\n");Diagnose("failed:",msg,"\n| ");
    elif hard then Print(" ",msg,"\c"); 
    fi;
  end;
  ps:=LsListList(ComplexConjugate(S),S);
  if ps=false then Diagnose("S# is not ps(S)\n| ");fi;
  inds:=[1..Length(f.eigenvalues)];
  if IsBound(W) then
    d:=CycPolUnipotentDegrees(W){f.charNumbers};
    fd:=uc.operations.FakeDegrees(uc,X(Rationals)){f.charNumbers};
    special:=PositionProperty(inds,i->fd[i].valuation=d[i].valuation);
    cospecial:=PositionProperty(inds,i->Degree(fd[i])=Degree(d[i]));
    check(f.special,special,"special=min.valuation");
    if special<>f.special then 
      Diagnose(" special is ",f.special, " should be ",special,"\n| ");
    fi;
    if special<>cospecial then
      if not IsBound(f.cospecial) then
        Diagnose(" cospecial is not bound; should be ",cospecial,"\n| ");
      elif cospecial<>f.cospecial then 
        Diagnose(" cospecial is ",f.cospecial, " should be ",cospecial,"\n| ");
      fi;
    fi;
    if hard then
      Print( " special=",special," cospecial=",cospecial);
    fi;
    if IsBound(f.perm) then
      check(ComplexConjugate(d),Permuted(d,f.perm),"ud#=ud^perm");
      if Ep<>false then
        check(Permuted(f.eigenvalues,Ep/f.perm),f.eigenvalues,"eig#=eig^perm");
      fi;
    fi;
    if hard then
      Print("\n|  S->S#:",DisplayLs(ps),"\n| ");
      if not IsBound(f.perm) then 
        if ForAll(ps,x->x>0) then
          Diagnose("family seems real but .perm not bound\n| ");
        else
          check(ComplexConjugate(d),LsPermuted(d,ps),"ud#=ud^last");
        fi;
      else # check ps is a sub-permutation of perm
        if not ForAll(inds,x->ps[x]=x or ps[x]=x^f.perm) then
          Diagnose("*** S->S# is not a sub-perm of .perm\n| ");
        fi;
      fi;
    fi;
  fi;
  check(S*TransposedMat(Sbar),Id,"S*tS#=1");
  if IsBound(f.perm) then
    if hard and Ep=false then Diagnose("ud=ud# but eig<>eig#\n| ");fi;
    P:=List(inds,x->inds*0);for i in inds do P[i][i^f.perm]:=1;od;
    check(S*P,P*S,"SP=PS");
  fi;
  if not IsBound(f.sh) then 
    check(S,TrS,"S=tS");
    check(T*S^2,S^2*T,"TSS=SST");
    if not IsBound(f.perm) or not IsBound(f.lusztig) then
         check((T*S)^3,Id,"(TS)^3=1");
    else check(S*T*S,T*P*S*T,"STS=TPST");fi;
  else check((T*TrS*DiagonalMat(f.sh)^-1*S)^2,S^0,"(T*tS*Sh-1*T)^2=1");
  fi;
  check((S*TrS)^2,Id,"(S*tS)^2=1");
# Print("\n| ");
# check(T,S*T*S,"T=STS");
  if not IsBound(f.special) then 
    Diagnose("special not bound\n| ");
    f.special:=1;
  fi;
# real:=S=Sbar and S=TrS;
  real:=IsBound(f.perm) or (S=Sbar and S=TrS);
# if real then Print(" evalf(S[f.special]): ",
#     List(S[f.special]/S[f.special][f.special],evalf));
# fi;
  if hard then Print("\n| ");fi;
  if Length(S)<=21 or (hard and Length(S)<=50) then
  s:=ShallowCopy(Sbar);
  for i in inds do s{inds}[i]:=s{inds}[i]/s[f.special][i];od;
  for i in inds do 
    if Length(S)>8 or hard then Print(" ",i,"\c");fi;
    for j in inds do 
      v:=List(inds,k->Sum(inds,l->S[i][l]*S[j][l]*s[k][l]));
      if not ForAll(v,IsInt) then
        Diagnose("not integral:",i,",",j,":",Format(v),"\n| ");
      fi;
      p:=ForAll(v,x->x>=0);
      if real and not p then 
        if ForAll(v,x->x<=0) then
          Diagnose("family is real but Sums[",i,",",j,"]<=0\n| ");
        else
          Diagnose("family is real but Sums[",i,",",j,"]=",Format(v),"\n| ");
        fi;
      fi;
      if hard then if p then Print("+\c");else Print(".\c");fi;fi;
    od;
  od;
  fi;
  if hard and Length(S)<50 then Print("\n");fi;
# CHECK.diagnosed:=true;
  Ok();
end;

sums:=function(m,special)local ind;
  m:=TransposedMat(m);
  ind:=[1..Length(m)];
  return List(ind,i->List(ind,j->List(ind,k->Sum(m{ind}{[i,j,k,special]},
    x->Product(x)/x[4]^2))));
end;

goodlines:=function(f)local m,ind,special,ss;
  m:=f.fourierMat;ind:=[1..Length(m)];
  if IsBound(f.delta1) then m:=m^DiagonalMat(f.delta1);fi;
  special:=Filtered(ind,x->Product(m[x])<>0);
  m:=TransposedMat(m);
  if Length(m)>15 then Print("lines to check:",special,"\n");fi;
  ss:=Filtered(special,ss->
    ForAll(ind,function(i)
      if Length(m)>15 then Print("[ss=",ss,",i=",i,"]\c");fi;
      return ForAll(ind,j->ForAll(ind,function(k)local
        res;res:=Sum(m{ind}{[i,j,k,ss]},x->Product(x)/x[4]^2);return IsInt(res) and res>=0;end));end));
  if ss<>special then Print("******* of non-0 lines:",special," only ",ss," are positive\n");fi;
  return ss;
end;

allfams:=function()
 return Concatenation(List([1..Length(TestGroups)],function(i)local W,n,f,ud,i;
   W:=TestGroup(i);n:=ReflectionName(W);f:=UnipotentCharacters(W).families;
   if IsCoxeterGroup(W) or IsCoxeterCoset(W) then return [];fi;
   ud:=CycPolUnipotentDegrees(W);
   for i in [1..Length(f)] do
     f[i].label:=SPrint(n,".",i); f[i].ud:=ud{f[i].charNumbers};
     f[i].operations:=rec(Print:=function(f)
       if not IsBound(f.name) then f.name:="?";fi;
       Print(f.label,"#",Length(f.charNumbers),"=",f.name);end);
   od;
   return Filtered(f,x->Length(x.charNumbers)>1);
   end));
end;
   
# Test ShF/F U has same dim. as U for split groups
# according to our formulae we have to test
# Frob*Fourierinverse*Frob*ud=ud 
# Multiplying by Fourier*Frob^-1 this is Fourier*Frob^-1*ud=Frob*ud
# If f.lusztig is true we consider Lusztig's fourier which is
# complexconjugation*ourfourier and ud are real so we have to test
# Fourier*Frob*ud=Frob*ud
CheckSh:=function(W)local uc,ud,f,e,sh,hprod;
  hprod:=function(x,y)return Zip(x,y,function(a,b)return a*b;end);end;
  uc:=UnipotentCharacters(W);
  ud:=UnipotentDegrees(W,X(Cyclotomics));
  Title("dim Sh F/F U_chi=dim U_chi");
  for f in uc.families do
    e:=hprod(ud{f.charNumbers},f.eigenvalues);
    if IsBound(f.lusztig) and not IsBound(f.deltas) then sh:=f.fourierMat*e;
    else sh:=f.fourierMat*hprod(ud{f.charNumbers},
                                ComplexConjugate(f.eigenvalues));
    fi;
    if sh<>e then Diagnose("Failed for ",Position(uc.families,f),
      "-th family of",ReflectionName(W),"\n");fi;
  od;
  Ok();
end;

CheckFamilies:=function(arg)local i;
  arg:=Concatenation([arg[1],0],arg{[2..Length(arg)]});
  for i in [1..Length(UnipotentCharacters(arg[1]).families)] do 
    arg[2]:=i;
    ApplyFunc(CheckFamily,arg);
  od;
end;
  
CheckFakeDegreesInduce:=function(W,J)local L,t,hd,ud,index,q,pred,found;
  q:=X(Cyclotomics);
  if not IsSpets(W) then W:=Spets(W);fi;
  for L in Twistings(W,J) do
   t:=InductionTable(L,W);
   hd:=FakeDegrees(L,q);
   ud:=FakeDegrees(W,q);
   index:=GenericSign(L)*GenericOrder(W,q)/(GenericOrder(L,q)*GenericSign(W));
	  index.valuation:=0;
   pred:=hd*index;found:=ud*t.scalar;
   Title("CheckFakeDegreesInduce with R^{",ReflectionName(W),"}_{",
     ReflectionName(L),"}");
   if pred<>found then Diagnose("quotient ",
     List([1..Length(hd)],i->CycPol(pred[i])/CycPol(found[i])),"\n");
   fi;
   Ok();
  od;
end;

# check the charparams are consistent with actual data in various ways
CheckCharParams:=function(W) local p,q,l,n,k,occursin,issign,fakdegbegins,
  fakdegsmaller,isconjugate,inducedfromid,nm,value,L,t,f;
  f:=x->Position(nm,x);
  occursin:=function(arg)
   if not 0<>ApplyFunc(Tensor,
        Concatenation([W],List(arg{[2..Length(arg)]},f)))[f(arg[1])]
   then Diagnose(arg[1]," should occur in tensor",arg{[2..Length(arg)]});
   fi;
  end;
  fakdegbegins:=function(a,p)
   if not Degree(FakeDegrees(W,q)[f(a)]-p)<p.valuation then
   Diagnose("FakeDegree(",a,") should begin ",p,"...");
   fi;
  end;
  fakdegsmaller:=function(a,b)
   if not Degree(FakeDegrees(W,q)[f(a)])<Degree(FakeDegrees(W,q)[f(b)]) then
   Diagnose("FakeDeg(",a,") should have a smaller degree than FakeDeg(",b,")");
   fi;
  end;
  issign:=function(a,b)
    if not 1=Tensor(W,PositionSgn(W),f(b))[f(a)] then
       Diagnose(a," should be tensor by sign of ",b);
    fi;
  end;
  isconjugate:=function(a,b)local ct;
    ct:=CharTable(W).irreducibles;
    if not ct[f(a)]=ComplexConjugate(ct[f(b)]) then
       Diagnose(a," should be conjugate to ",b);
    fi;
  end;
  inducedfromid:=function(a,J)local L,t;
    L:=ReflectionSubgroup(W,J);
    t:=InductionTable(L,W);
    if not t.scalar[f(a)][PositionId(L)]<>0 then
       Diagnose(a," should occur in the induced of Id from ",ReflectionName(L));
    fi;
  end;
  value:=function(a,c,v)
    if not CharTable(W).irreducibles[f(a)][c]=v then
      Diagnose(a," should take value ",v," at class ",c);
    fi;
  end;
  n:=ReflectionName(W);
  q:=X(Rationals);
  p:=List(FakeDegrees(W,q),x->[Value(x,1),x.valuation]);
  if IsClassical(W) then 
    if not n in ["G3,3,4","G3,3,5","G4,4,3"] then return;fi;
    Title("CheckCharParams");
    l:=ChevieCharInfo(W).malle;
    nm:=List(l,x->CHEVIE.RawData("CharName","G2")(x,rec()));
  else
    Title("CheckCharParams");
    l:=Filtered(Collected(CharParams(W)),x->x[2]>1);
    if Length(l)>0 then Diagnose("Duplicated:",l,"\n");fi;
    l:=List(CharParams(W),x->x[1]);
    nm:=CharNames(W);
  fi;
  if not  ForAll([1..Length(l)],i->p[i]=l[i]{[1,2]}) then
    Diagnose("disagree with (\\chi(1),b_\\chi)\n");
    cmpobj(p,List(l,x->x{[1,2]}),"[chi(1),b_chi]","param");
    Error();
  fi;
  if n="G3,3,4" then
    Print("details for G3_3_4");
    occursin("phi{6,5}'","phi{4,1}","phi{4,1}");
  elif n="G3,3,5" then
    Print("details for G3_3_5");
    occursin("phi{30,7}''","phi{5,1}","phi{5,1}","phi{5,1}","phi{5,1}");
  elif n="G4,4,3" then
    Print("details for G4_4_3");
    isconjugate("phi{3,2}''","phi{3,1}");
    isconjugate("phi{3,6}''","phi{3,5}");
  elif n="G5" then
    Print("details for G5");
    isconjugate("phi{1,8}''","phi{1,16}");
    isconjugate("phi{1,4}'","phi{1,8}'");
    isconjugate("phi{1,4}''","phi{1,8}'''");
    isconjugate("phi{1,12}'","phi{1,12}''");
    isconjugate("phi{2,5}'","phi{2,7}'");
    isconjugate("phi{2,7}''","phi{2,5}'''");
    isconjugate("phi{2,5}''","phi{2,1}");
    isconjugate("phi{2,3}'","phi{2,3}''");
    p:=Position(ChevieClassInfo(W).classtext,[1]);
    value("phi{1,4}'",p,1);
    value("phi{1,12}'",p,E(3));
    value("phi{2,5}'",p,-1);
    value("phi{2,3}'",p,-E(3));
  elif n="G6" then
    Print("details for G6");
    fakdegsmaller("phi{2,3}'","phi{2,3}''");
    fakdegsmaller("phi{2,5}'","phi{2,5}''");
  elif n="G8" then
    Print("details for G8");
    isconjugate("phi{2,7}'","phi{2,1}");
  elif n="G14" then
    Print("details for G14");
    fakdegsmaller("phi{3,6}'","phi{3,6}''");
  elif n="G25" then
    Print("details for G26");
    isconjugate("phi{3,5}'","phi{3,1}");
    isconjugate("phi{3,13}'","phi{3,17}");
    isconjugate("phi{6,8}'","phi{6,10}");
    isconjugate("phi{6,4}'","phi{6,2}");
  elif n="G26" then
    Print("details for G26");
    fakdegbegins("phi{8,6}'",q^24);
    fakdegbegins("phi{8,9}'",q^27);
    isconjugate("phi{3,8}'","phi{3,4}");
    isconjugate("phi{3,5}'","phi{3,1}");
    isconjugate("phi{3,16}'","phi{3,20}");
    isconjugate("phi{3,13}'","phi{3,17}");
    isconjugate("phi{6,8}'","phi{6,10}");
    isconjugate("phi{6,11}'","phi{6,13}");
    isconjugate("phi{6,4}'","phi{6,2}");
    isconjugate("phi{6,7}'","phi{6,5}");
  elif n="G27" then
    Print("details for G27");
    fakdegsmaller("phi{3,5}'","phi{3,5}''");  
    fakdegsmaller("phi{3,20}'","phi{3,20}''");
    fakdegsmaller("phi{8,9}''","phi{8,9}'");
    inducedfromid("phi{5,15}'",[1,3]);
    L:=ReflectionSubgroup(W,[1,3]);
    t:=InductionTable(L,W);
    if not t.scalar[Position(nm,"phi{5,6}'")][PositionId(L)]=1 then
       Diagnose("phi{5,6}' should occur once the induced of Id from ",
        ReflectionName(L));
    fi;
  elif n="G29" then
    Print("details for G29");
    occursin("phi{15,4}''","phi{4,1}","phi{4,3}");
    issign("phi{15,4}''","phi{15,12}''");
    occursin("phi{6,10}'''","phi{4,1}","phi{4,1}");
    isconjugate("phi{6,10}''''","phi{6,10}'''");
    inducedfromid("phi{6,10}'",[1,2,4]);
  elif n="G31" then
    Print("details for G31");
    occursin("phi{15,8}'","phi{4,1}","phi{20,7}");
    occursin("phi{15,20}'","phi{4,1}","phi{20,7}");
    occursin("phi{45,8}''","phi{4,1}","phi{20,7}");
    fakdegbegins("phi{30,10}'",q^50+q^46);
    isconjugate("phi{20,13}'","phi{20,7}");
    issign("phi{45,12}'","phi{45,8}'");
  elif n="G32" then
    Print("details for G32");
    isconjugate("phi{30,12}''","phi{30,12}''");
    isconjugate("phi{60,15}''","phi{60,15}''");
    isconjugate("phi{20,9}'","phi{20,3}");
    isconjugate("phi{20,29}'","phi{20,13}");
    isconjugate("phi{30,12}'","phi{30,24}");
    isconjugate("phi{30,20}'","phi{30,16}");
    isconjugate("phi{60,11}'","phi{60,7}");
    isconjugate("phi{60,15}'","phi{60,9}");
  elif n="G33" then
    Print("details for G33");
    issign("phi{10,17}'","phi{10,8}'");
    issign("phi{40,5}'","phi{40,14}'");
    fakdegbegins("phi{10,8}'",q^28+q^26);
    fakdegbegins("phi{40,5}'",q^31+q^29+2*q^27);
  elif n="G34" then
    Print("details for G34");
    occursin("phi{20,33}'","phi{6,1}","phi{15,14}");
    occursin("phi{70,9}''","phi{6,1}","phi{15,14}");
    occursin("phi{560,18}'''","phi{6,1}","phi{336,17}");
    occursin("phi{280,12}'","phi{6,1}","phi{336,17}");
    occursin("phi{280,30}''","phi{6,1}","phi{336,17}");
    occursin("phi{540,21}'","phi{6,1}","phi{105,20}");
    issign("phi{70,45}''","phi{70,9}''");
    isconjugate("phi{105,8}'","phi{105,4}");
    isconjugate("phi{840,13}'","phi{840,11}");
    isconjugate("phi{840,23}'","phi{840,19}");
    isconjugate("phi{70,9}'","phi{70,9}'");
    isconjugate("phi{70,45}'","phi{70,45}'");
    isconjugate("phi{560,18}'","phi{560,18}'");
    inducedfromid("phi{120,21}'",[1,2,4,5,6]);
  fi;
  Ok();
end;

# check that parabolicinfo is correct
CheckParabolicInfo:=function(W)local pi,j,d,pinfo,pcomp,r,p;
  if IsClassical(W) or IsBound(W.isCoxeterGroup) then return;fi;
  Title("CheckParabolicInfo");
  pi:=CHEVIE.Data("ParabolicInfo",W.type[1]);
  if pi=false then Diagnose("absent\n");return;fi;
  for j in W.generatingReflections do
   d:=Difference(W.generatingReflections,[j]);
   pcomp:=PermRootOps.ReflectionType(PermRootOps.ReflectionSubgroup(W,d));
   p:=MappingPermListList([1..Length(d)],d);
   for r in pcomp do 
     r.indices:=OnTuples(r.indices,p);
     Unbind(r.cartan);
     Unbind(r.reflectionName);
   od;
   pinfo:=PositionProperty(pi,x->d=Set(Concatenation(List(x,x->x.indices))));
   if pinfo=false then Diagnose(d,": no such parabolic\n");
   else pinfo:=pi[pinfo];
   for r in pinfo do r.operations:=ReflTypeOps;od;
   if Set(pinfo)<>Set(pcomp) then 
     Diagnose("Parabolic ", d,"=",ReflectionName(pcomp)," computed differs\n");
     for r in pcomp do Unbind(r.reflectionName);od;
     cmpobj(pinfo,pcomp,"stored","computed");
   fi;
   fi;
  od;
  Ok();
end;

# check eigenvalues in families agree with those in HC series
CheckEigen:=function(W)local uc,i,e,j,f,n;
  uc:=UnipotentCharacters(W);
  e:=Eigenvalues(uc,[1..Length(uc.charNames)]);
  Title("CheckEigen");
  for i in [1..Length(uc.harishChandra)] do
    f:=uc.harishChandra[i];
    for j in [1..Length(f.charNumbers)] do
      n:=f.charNumbers[j];
      if e[n]<>f.eigenvalue then
        Diagnose("HCfamily ",i,"#",j,":",uc.charNames[n],"=",n,
          " eigen from fam.=",e[n]," but from HC=",f.eigenvalue,"\n");
      fi;
    od;
  od;
  Ok("from unip.char.=from HC series");
end;

# check stored a and A are correct
CheckaA:=function(W)local uc,ud,q;
  Title("CheckaA");
  q:=X(Cyclotomics); uc:=UnipotentCharacters(W); ud:=UnipotentDegrees(W,q);
  if IsBound(uc.a) then cmpvec(uc.a,List(ud,x->x.valuation),j->SPrint("uc.a[",j,"]"));fi;
  if IsBound(uc.A) then cmpvec(uc.A,List(ud,Degree),j->SPrint("uc.A[",j,"]"));fi;
  Ok("agree with unipotent degrees");
end;

CheckHeckeSpecializes:=function(W)local H,i;
  Unbind(W.HeckeAlgebras);
  CHEVIE.checkroots:=true;
  for i in [0..4] do
    H:=GenericHecke(W,i);
    Print("parameters=",H.parameter,"\n");
    CharTable(H);
  od;
end;

CheckUnipotentDegreesClassical:=function(W)local uc,ud,i,v,d,vud,q,s;
  if not IsClassical(W) or ReflectionName(W){[1,2]}="2A" then return;fi;
  uc:=UnipotentCharacters(W);v:=X(Cyclotomics);
  ud:=List(UnipotentDegrees(W,v),CycPol);
  vud:=List(uc.charSymbols,x->CycPolGenericDegreeSymbol(x[1]));
  Title("CheckUnipotentDegreesClassical");
  for i in [1..Length(ud)] do
    cmpvcyc(vud[i],ud[i],Concatenation("GenDegSymbol",
      StringSymbol(uc.charSymbols[i][1])),"ud");
  od;
  Ok();
end;

CheckFakeDegreesClassical:=function(W)local uc,fd,i,v,d,vfd,n,h;
  if not IsClassical(W) or ReflectionName(W){[1,2]}="2A" then return;fi;
  uc:=UnipotentCharacters(W);h:=uc.almostHarishChandra[1].charNumbers;
  v:=X(Cyclotomics);fd:=List(FakeDegrees(W,v),CycPol);n:=ReflectionName(W);
  if n{[1,2]} in ["2D","2I","2B","2G"] then 
       vfd:=List(uc.almostCharSymbols{h},x->CycPolFakeDegreeSymbol(x[1],1));
  else vfd:=List(uc.almostCharSymbols{h},x->CycPolFakeDegreeSymbol(x[1]));
  fi;
  Title("CheckFakeDegreesClassical");
  for i in [1..Length(h)] do
    cmpvcyc(vfd[i],fd[i],Concatenation("FakeDegSymbol",
      StringSymbol(uc.almostCharSymbols[h[i]][1])),"fd");
  od;
  Ok();
end;

# check that the parameter exponents of the relative hecke algebras
# agree with unipotent degrees corresponding to cyclic
# relative groups of sub-series
CheckParameterExponents:=function(arg) # (W[,HC no])
  local h,i,I,L,hh,ud,t,H,s,exp,W,ser,G,m,e;
  W:=arg[1];
  if Length(arg)=2 then ser:=UnipotentCharacters(W).harishChandra{arg[2]};
  else ser:=UnipotentCharacters(W).harishChandra;
  fi;
  for h in ser do
    I:=Concatenation(List(h.relativeType,x->x.indices));
    if IsSpets(W) then I:=List(I,x->Cycle(W.phi,x));
    else I:=List(I,x->[x]);
    fi;
    for i in [1..Length(I)] do
      L:=ReflectionSubgroup(W,Concatenation(h.levi,I[i]));
      t:=ReflectionType(L);
      H:=ReflectionSubgroup(L,h.levi);
      Title("CheckParameterExponent from ",ReflectionName(H),":",I[i]);
      if t=false or (not IsBound(t[1].indices) and not IsBound(t[1].orbit))then
        Diagnose("Levi ",Concatenation(h.levi,I[i])," could not be identified\n");
      else
        if not IsSpets(L) then L:=Spets(L);H:=SubSpets(L,h.levi);fi;
        hh:=FindSeriesInParent(h,H,L,UnipotentCharacters(L).harishChandra).ser;
        if Length(hh.charNumbers)<>2 then 
	  s:=Series(L,H,Position(UnipotentCharacters(H).charNames,h.cuspidalName),1);
	  if Group(H).semisimpleRank=0 then 
	    s.projector:=IdentityMat(Group(H).rank);
	  else 
	   G:=Group(H);
	   m:=MatXPerm(G,Product(G.reflections{G.generatingReflections}));
	   e:=Difference(EigenvaluesMat(m),[1]);
	   s.projector:=Product(e,x->(m-m^0*x)/(1-x));
	  fi;
	  SeriesOps.RelativeGroup(s);
	  if SeriesOps.chars(s)=false or SeriesOps.fill(s)=false then 
	    Error("could not fill ",s);
	  else s.mC:=Permuted(s.mC,SortingPerm(s.assignment));
	    s.charNumbers:=Permuted(s.charNumbers,SortingPerm(s.assignment));
	    if IsInt(hh.parameterExponents[1]) then 
	      exp:=[1..s.e]*0;exp[1]:=hh.parameterExponents[1];
	    else exp:=hh.parameterExponents[1];
	    fi;
	    t:=Filtered([0,s.translation..s.e],function(d)local v;
	      v:=1+List([1..s.e]+d,x->x mod s.e);
	      return s.mC{v}=exp and s.charNumbers{v}=hh.charNumbers;end);
	    if Length(t)=0 then Error("unexpected");fi;
	    Ok("decs=",t);
	  fi;
        else
          ud:=UnipotentDegrees(L,X(Cyclotomics));
          ud:=ud[hh.charNumbers[1]]/ud[hh.charNumbers[2]];
          if Length(ud.coefficients)<>1 or ud.coefficients[1]<>1 then
            Diagnose("not monomial");
          elif h.parameterExponents[i]<>ud.valuation then
            Diagnose(ReflectionName(L),": wrong parameter ",
	      h.parameterExponents[i],
	      " instead of ",ud.valuation,"\n");
          fi;
          Ok("=",ud.valuation);
        fi;
      fi;
    od;
  od;
end;

CheckRelativeCosets:=function(W)local h,r,t,i,tt,hh;
  Title("CheckRelativeCosets");
  if IsSpets(W) then 
    for h in UnipotentCharacters(W).almostHarishChandra do
      r:=RelativeCoset(W,h.levi);
      t:=ReflectionType(r);
      if Length(t)<>Length(h.relativeType) then Error();fi;
      for i in [1..Length(t)] do
	if Length(t[i].orbit)>1 then Error("not irred");fi;
	tt:=t[i].orbit[1];
	if not IsBound(h.relativeType[i].orbit) then
	  if t[i].twist<>() then Error("should be twisted");fi;
	  hh:=h.relativeType[i];
	else hh:=h.relativeType[i].orbit[1];
	fi;
	if hh.series<>tt.series then 
	    Diagnose("series do not match: is ",hh.series,
		  " should be ",tt.series,"\n");fi;
	if hh.indices<>Group(r).relativeIndices{tt.indices} then 
	  Diagnose("indices do not match: are ",hh.indices,
		" should be ",Group(r).relativeIndices{tt.indices},"\n"); fi;
      od;
    od;
  else
    for h in UnipotentCharacters(W).almostHarishChandra do
      r:=RelativeGroup(W,h.levi,
            Concatenation(List(h.relativeType,
            x->Concatenation(List(x.orbit,y->y.indices)))));
      t:=ReflectionType(r);
      if Length(t)<>Length(h.relativeType) then Error();fi;
      for i in [1..Length(t)] do
        tt:=t[i];
        if h.relativeType[i].orbit[1].series<>tt.series then 
          Diagnose("stored series is ",h.relativeType[i].orbit[1].series,
                  " should be ",tt.series,"\n");fi;
	if not IsBound(tt.indices) then
          Diagnose("could not find indices (expected ",
	            h.relativeType[i].orbit[1].indices,")");
	elif Maximum(tt.indices)>Length(r.relativeIndices) then
	  Diagnose("use non-simple roots:",tt.indices," of ",r,
	      "<",Join(r.relativeIndices),">\n");
	elif h.relativeType[i].orbit[1].indices<>r.relativeIndices{tt.indices} 
	then Diagnose("stored indices are ",
	   h.relativeType[i].orbit[1].indices,
          " should be ",r.relativeIndices{tt.indices},"\n"); fi;
      od;
    od;
  fi;
  Ok();
end;

# According to Michel (principal series) and Gunter (other cases)
# the eigenvalue of Ennola_E(z)^k(rho_chi) is  
#  EigenEnnola(W,i,k)/Ennolascalar(rho,E(z)^k)
# where z=|ZW| and rho_chi is i-th unip. char.
CheckEnnolaEigen:=function(W)local z,pred,actual,uw,e,i;
  z:=OrderCenter(W);if z=1 then return;fi;
  Title("CheckEnnolaEigen");
  e:=Ennola(W)[1]; uw:=UnipotentCharacters(W);
  pred:=List([1..Length(uw.a)], i->EigenEnnola(W,i,1)/
     e.scalars[PositionProperty(uw.harishChandra,h->i in h.charNumbers)]);
  actual:=[];
  for i in [1..Length(e.ls)] do
    actual{uw.families[i].charNumbers}:=
      uw.families[i].eigenvalues{List(e.ls[i],AbsInt)};
  od;
  e:=ListBlist(TransposedMat([[1..Length(uw.a)],uw.charNames,pred,actual]),
      List([1..Length(pred)],i->pred[i]<>actual[i]));
  e:=Concatenation([["no.","char","predicted","actual"]],e);
  if Length(e)<>1 then Diagnose("\n",StringTable(e));fi;
  Ok();
end;

# The sum of squares of fakedegrees should equal the sum of
# normsquares of unipotent degrees
CheckSumofSquares:=function(W)local q;
  Title("CheckSumOfSquares");
  q:=X(Cyclotomics);
  if Sum(FakeDegrees(W,q),x->x^2)<>Sum(UnipotentDegrees(W,q),
   x->x*ComplexConjugate(x)) then Diagnose("fails\n");fi;
  Ok();
end;

# CheckBraidRelations(W[,t]) checks W satisfies relations of its type
# (of type t if given)
CheckBraidRelations:=function(arg)local W,r,gens,rel;
  W:=arg[1];
  if Length(arg)=1 then r:=BraidRelations(W);
  else r:=BraidRelations(arg[2]);
  fi;
  Title("CheckBraidRelations");
  gens:=List(W.generatingReflections,i->Reflection(W,i));
  for rel in r do
    rel:=List(rel,x->List(W.rootRestriction{x},
      y->Position(W.generatingReflections,y)));
    CheckBraidRelation(gens,rel,Diagnose);
  od;
  Ok();
end;

# CheckWFromBraidRelations(W)
# check abstract group given by G's braid relations has same size as G
# (thus no relation is missing)
CheckWFromBraidRelations:=function(W)local n,F,r;
  n:=Length(W.generators);
  F:=FreeGroup(n);
  r:=List(BraidRelations(W),x->List(x,y->Product(y,z->F.(z))));
  r:=List(r,x->x[1]/x[2]);
  Append(r,List([1..n],i->F.(i)^OrderPerm(W.(i))));
  return Size(W)=Size(F/r);
end;

#partial list of parabolic representatives for g34
p34:=[ [ 1 ], [ 3, 1 ], [ 1, 2 ], [ 5, 3, 1 ], [ 3, 4, 1 ], [ 3, 1, 2 ], 
  [ 1, 2, 17 ], [ 6, 5, 3, 1 ], [ 5, 3, 1, 2 ], [ 5, 3, 1, 11 ] ];

# omnibus test
Check:=function(arg)local W,uc,s;
  W:=arg[1];
  InitCheck(W);
  if not IsSpets(W) then 
    CheckBraidRelations(W);CheckPowerMap(W); 
    CheckPositionRefl(W); CheckPositionSgn(W);
  fi;
  CheckCharTable(W);
  CheckFakeDegrees(W);
  if not IsSpets(W) then CheckParabolicInfo(W); CheckCharParams(W); fi;
# ApplyFunc(CheckSchurRelations,arg);
  uc:=UnipotentCharacters(W);
  if uc<>false then
    Display(uc);
    CheckUnipotentDegreesClassical(W);
    CheckFakeDegreesClassical(W);
    CheckEigen(W);
    CheckaA(W);
    CheckParameterExponents(W);
    CheckRelativeCosets(W);
    CheckSgnUd(W);
    for s in ProperStableParabolics(W) do CheckGendegsHCInduce(W,s);od;
    if not IsSpets(W) then  CheckAllDegreesFromHC(W);
    else CheckHCDegrees(W);
    fi;
    CheckSeries(W);
    CheckFamilies(W);
    if not IsSpets(W) then CheckSh(W);CheckEnnolaEigen(W);fi;
    CheckAllLusztig(W);
  else for s in ProperParabolics(W) do CheckFakeDegreesInduce(W,s);od;
  fi;
  EndCheck();
end;

RegressionTest:=function(arg)local oldlog,i,t;
  if Length(arg)=0 then arg:=[1..Length(TestGroups)];
  elif IsList(arg[1]) then  arg:=arg[1];
  fi;
  oldlog:=CHECK.log;
  CHECK.log:=true;
  CHECK.logfile:="all.log";
  for i in arg do Print(i,":"); Check(TestGroup(i));od;
  CHECK.log:=oldlog;
end;
