##############################################################################
##
#A  util.g       VKCURVE package         Jean Michel
##
#Y  Copyright (C) 2001 - 2002  University Paris VII, France.
##
##  This file holds various utility functions.
## 
#############################################################################
BigNorm:=x->AbsInt(x.r)+AbsInt(x.i); # alternative to ComplexOps.Norm(x)

SmallNorm:=x->Maximum(AbsInt(x.r),AbsInt(x.i)); # another alternative

# 'vraie' norme d'un nombre complexe
Rho:=function(a)
  a:=ComplexOps.Norm(a);
  if IsRat(a) then a:=evalf(a,-DecimalLog(a)+5);fi;
  return GetRoot(a,2);
end;

# For a non-zero rational x, returns k such that 10^k<|x|<=10^(k+1) 
#
DecimalLog:=function(x)local d;
  if x=0*x then Error("trying to take decimal log of 0");fi;
  d:=QuoInt(Denominator(x),Numerator(x));
  if d=0 then d:=QuoInt(Numerator(x),Denominator(x));
     return Length(String(AbsInt(d)))-1;
  else return -Length(String(AbsInt(d)));fi;
end;

DecimalToRat:=function(a)
  if IsRat(a) then return a;
  else return a.mantissa/a.precision;
  fi;
end;

ComplexToComplexRat:=a->Complex(DecimalToRat(a.r),DecimalToRat(a.i));

ComplexRatToGaussian:=a->a.r+E(4)*a.i;

# returns [minimum of l, position of that minimum in l]
#
MinPos:=function(l)local min,pos,i; pos:=1;min:=l[1];
  for i in [2..Length(l)] do if l[i]<min then min:=l[i];pos:=i;fi;od;
  return [min,pos];
end;

# minimum distance between 2 points of v
# if the minimum m is realized between p1 and p2,
# returns [m,[index of p1 in v, index of p2 in v]]
#
Dispersal:=function(v)local l,p,minpos;
  l:=Concatenation(List([1..Length(v)],i->List([1..i-1],j->[j,i])));
  p:=MinPos(List(l,x->BigNorm(v[x[1]]-v[x[2]])));
  return [p[1],l[p[2]]];
end;

# distance of z to segment [a,b]
DistSeg:=function(z,a,b) local r;
  b:=b-a;z:=z-a;r:=Rho(b);z:=r/b*z;
  if z.r<0 then return Rho(z);
  elif z.r>r then return Rho(z-b);
  elif z.i>0 then return z.i;
  else return -z.i;
  fi;
end;

PrimeCoeffs:=function(p)local F,q;
  F:=DefaultField(p.coeff);
  q:=List(Coefficients(p,"y"),q->Polynomial(F,ScalMvp(Coefficients(q,"x"))));
  q:=Gcd(q);
  if Degree(q)>0 then
    Error("coefficients of ",p," have gcd=",Value(q,Mvp("x")),"\n");
  fi;
end;

# discriminant with respect to y of an Mvp in x,y with rational coeffs
Discy:=function(p)local n,v;
  n:=2*Length(Coefficients(p,"x"))*Length(Coefficients(p,"y"));
  v:=List([1..n],function(i)local q;
    q:=ScalMvp(Coefficients(Value(p,["y",i]),"x"));
    if Length(q)=0 then return 0;
    elif Length(q)=1 then return q[1];
    else return DeterminantMat(ResultantMat(q,Deriv(q)));
    fi;
    end);
  v:=InterpolatedPolynomial(DefaultField(v),[1..n],v);
  return Value(v,Mvp("x"));
end;

# Resultant matrix of polynomials with vector-of-coeffs v, w
ResultantMat:=function(v,w)local m,i;
  v:=Reversed(v);
  w:=Reversed(w);
  m:=List([1..Length(v)+Length(w)-2],x->[1..Length(v)+Length(w)-2]*0*v[1]);
  for i in [1..Length(w)-1] do m[i]{[i..i+Length(v)-1]}:=v;od;
  for i in [1..Length(v)-1] do m[i+Length(w)-1]{[i..i+Length(w)-1]}:=w;od;
  return m;
end;

# valeur en z du vecteur-polynome p par la methode de Horner
Horner:=function(p,z)local c,res; 
  res:=z*0;for c in Reversed(p) do res:=z*res+c;od;
  return res;
end;

Deriv:=p->List([2..Length(p)],i->(i-1)*p[i]);

# Cut(string,width=78,cutplaces=",",file=false)
Cut:=function(arg)local p,s,w,letter,file,wr,pos;
  s:=arg[1];
  if Length(arg)>1 then w:=arg[2]; else w:=78;fi;
  if Length(arg)>2 then letter:=arg[3]; else letter:=",";fi;
  if Length(arg)>3 then file:=arg[4];
    wr:=function(arg)ApplyFunc(AppendTo,Concatenation([file],arg));end;
  else wr:=Print;
  fi;
  pos:=0;
  while Length(s)>pos+w+1 do
    p:=PositionProperty(Reversed([1..w]),x->s[pos+x] in letter);
    if p=false then Error("could not cut ",s{pos+[1..w]});fi;
    wr(s{pos+[1..w+1-p]},"\n");
    pos:=pos+w+1-p;
  od;
  wr(s{[pos+1..Length(s)]},"");
end;
